-- Raven is an addon to monitor auras and cooldowns, providing timer bars, action bar highlights, and helpful notifications.

-- Bars.lua supports mapping auras to bars and grouping bars into multiple moveable frames using LibBars-1.0.
-- It has special case code for tooltips, test bars, turning off auras with right click, and the druid Eclipse proc.
-- There are no exported functions at this time other than those called to initialize and update bars.

local MOD = Raven
local LBF = LibStub("LibButtonFacade", true)
local media = LibStub("LibSharedMedia-3.0")
local tc = { r = 1, g = 1, b = 1, a = 1 }
local fishSpell = GetSpellInfo(7620)
local hidden = false
local detectedBar = { enableBar = true, sorder = 0 }
local settingsTemplate = {} -- settings are initialized from default bar group template
local defaultNotificationIcon = "Interface\\Icons\\Spell_Nature_WispSplode"
local testColors = { "Blue1", "Cyan", "Green1", "Yellow1", "Orange1", "Red1", "Pink", "Purple1", "Brown1", "Gray" }

-- Saved variables don't handle being set to nil properly so need to use alternate value to indicate an option has been turned off
local Off = 0 -- value used to designate an option is turned off
local function IsOff(value) return value == nil or value == Off end -- return true if option is turned off
local function IsOn(value) return value ~= nil and value ~= Off end -- return true if option is turned on

local classNames = { DEATHKNIGHT = "Death Knight", DRUID = "Druid", HUNTER = "Hunter", MAGE = "Mage", PALADIN = "Paladin",
	PRIEST = "Priest", ROGUE = "Rogue", SHAMAN = "Shaman", WARLOCK = "Warlock", WARRIOR = "Warrior" }

MOD.BarGroupTemplate = { -- default bar group settings
	enabled = true, locked = true, merged = false, linkSettings = true, checkCondition = false, barColors = "Spell", bgColors = "Normal",
	combatFader = false, combatAlpha = 1, combatClicks = true, combatTips = true,
	useDefaultDimensions = true, useDefaultFontsAndTextures = true, useDefaultColors = true, sor = "A", reverseSort = false,
	anchor = false, anchorX = 0, anchorY = 0, anchorLastBar = false,
	configuration = 1, growDirection = true, flashExpiring = false, flashTime = 5, fade = false, fadeTime = 5,
	wrap = 0, wrapDirection = false, labelOffset = 0, labelInset = 0, labelWrap = false, labelAlign = "MIDDLE",
	timeOffset = 0, timeInset = 0, iconTextOffset = 0, iconTextInset = 0,
	showSolo = true, showParty = true, showRaid = true, showCombat = true, showOOC = true, showFishing = true, showResting = true, showMounted = true,
	hideIcon = false, hideClock = false, hideBar = false, hideSpark = false, hideValue = false, hideLabel = false, showTooltips = true,
	detectBuffs = false, detectDebuffs = false, detectCooldowns = false, detectBuffsMonitor = "player", detectBuffsCastBy = "Player",
	detectDebuffsMonitor = "player", detectDebuffsCastBy = "Player", detectCooldownsBy = "Player", detectTracking = false,
	detectSpellCooldowns = true, detectTrinketCooldowns = true, detectPotionCooldowns = true, detectOtherCooldowns = false,
	detectSharedJudgements = true, detectSharedFrostTraps = true, detectSharedShocks = true, detectRuneCooldowns = false,
	setDuration = false, uniformDuration = 120, checkDuration = false, minimumDuration = true, filterDuration = 120,
	checkTimeLeft = false, minimumTimeLeft = true, filterTimeLeft = 120, showNoDuration = false, showOnlyNoDuration = false,
	showNoDurationBackground = false, timeFormat = 6, timeSpaces = false, timeCase = false,
	filterBuff = true, filterBuffLink = true, filterDebuff = true, filterDebuffLink = true, filterCooldown = true, filterCooldownLink = true,
}

MOD.BarGroupLayoutTemplate = { -- all the bar group settings involved in layout configuration
	barWidth = 0, barHeight = 0, iconSize = 0, scale = 0, spacingX = 0, spacingY = 0, iconOffsetX = 0, iconOffsetY = 0,
	useDefaultDimensions = 0, configuration = 0, growDirection = 0, wrap = 0, wrapDirection = 0, snapCenter = 0,
	labelOffset = 0, labelInset = 0, labelWrap = 0, labelAlign = 0, timeOffset = 0, timeInset = 0, iconOffset = 0, iconInset = 0,
	hideIcon = 0, hideClock = 0, hideBar = 0, hideSpark = 0, hideValue = 0, hideLabel = 0, showTooltips = 0
}

-- Initialize bar groups from those specified in the profile after, for example, a reloadUI or reset profile
function MOD:InitializeBars()
	local bgs = Nest_GetBarGroups()
	if bgs then -- first remove any bar groups represented in the graphics library
		for _, bg in pairs(bgs) do Nest_DeleteBarGroup(bg) end
	end -- then set up the ones specified in the profile
		
	for _, bg in pairs(MOD.db.profile.BarGroups) do
		if IsOn(bg) then
			for n, k in pairs(MOD.db.global.Defaults) do if bg[n] == nil then bg[n] = k end end -- add default settings for layout-fonts-textures
			for n, k in pairs(MOD.BarGroupTemplate) do if bg[n] == nil then bg[n] = k end end -- add defaults from the bar group template
			for _, bar in pairs(bg.bars) do -- iterate over each bar in the bar group to force cached icon lookup
				if bar.enableBar and (bar.barType == "Notification") and bar.associatedSpell then MOD:GetIcon(bar.associatedSpell) end
			end
			MOD:InitializeBarGroup(bg, 0, 0)
		end
	end
	if LBF then -- restore ButtonFacade skin information from the profile
		local lbf = MOD.db.profile.ButtonFacadeSkin -- table containing saved skin data
		if lbf then Nest_SetButtonFacadeData(lbf.SkinID, lbf.Gloss, lbf.Backdrop, lbf.Colors) end
	end
	MOD:UpdateAllBarGroups() -- this is done last to get all positions updated correctly
end

-- Finalize bar groups prior to logout, stripping out all values that match current defaults
function MOD:FinalizeBars()
	for bn, bg in pairs(MOD.db.profile.BarGroups) do
		if IsOn(bg) then
			for n, k in pairs(MOD.db.global.Defaults) do if bg[n] == k then bg[n] = nil end end -- remove default settings for layout-fonts-textures
			for n, k in pairs(MOD.BarGroupTemplate) do if bg[n] == k then bg[n] = nil end end -- remove defaults from the bar group template
		else
			MOD.db.profile.BarGroups[bn] = nil -- okay to delete these since no default bar groups
		end
	end
	if LBF then -- save ButtonFacade skin information in the profile
		local lbf = MOD.db.profile.ButtonFacadeSkin -- table containing saved skin data
		if not lbf then lbf = {}; MOD.db.profile.ButtonFacadeSkin = lbf end
		lbf.SkinID, lbf.Gloss, lbf.Backdrop, lbf.Colors = Nest_GetButtonFacadeData()
	end
end

-- Remove default values from bar group settings
function MOD:FinalizeSettings()
	for _, settings in pairs(MOD.db.global.Settings) do
		for n, k in pairs(settingsTemplate) do if settings[n] == k then settings[n] = nil end end -- remove values still set to defaults
	end
end

-- Raven is disabled so hide all features
function MOD:HideBars()
	if not hidden then
		for _, bp in pairs(MOD.db.profile.BarGroups) do
			if IsOn(bp) then MOD:ReleaseBarGroup(bp) end
		end
		hidden = true
	end
end

-- Initialize bar group settings by adding default values if necessary
function MOD:InitializeSettings()
	for n, k in pairs(MOD.BarGroupTemplate) do settingsTemplate[n] = k end -- initialize the settings template from bar group defaults
	for n, k in pairs(MOD.db.global.Defaults) do settingsTemplate[n] = k end -- add default settings for layout-fonts-textures
	settingsTemplate.enabled = nil; settingsTemplate.locked = nil; settingsTemplate.merged = nil; settingsTemplate.linkSettings = nil	
	for _, settings in pairs(MOD.db.global.Settings) do
		for n, k in pairs(settingsTemplate) do if settings[n] == nil then settings[n] = k end end -- add missing defaults from settings template
	end
end

-- Fire off test bars for this bar group, remove if any already exist
local function TestBarGroup(bp)
	local bg = Nest_GetBarGroup(bp.name)
	if bg then
		local found = false
		local icon = "Interface\\Icons\\INV_Drink_20"
		for i = 1, 10 do
			local bar = Nest_GetBar(bg, ">>Test<<" .. string.format("%02d", i))
			if bar then found = true; Nest_DeleteBar(bg, bar) end
		end
		if not found then
			for i = 1, 10 do			
				local bar = Nest_CreateBar(bg, ">>Test<<" .. string.format("%02d", i))
				if bar then
					local c = MOD.ColorPalette[testColors[i]]
					Nest_SetColor(bar, c.r, c.g, c.b, 1, c.r, c.g, c.b, 1)
					Nest_SetLabel(bar, "Test Bar " .. i)
					Nest_SetIcon(bar, icon)
					Nest_SetCount(bar, i)
					Nest_StartTimer(bar, i*5, 60, 60)
					Nest_SetAttribute(bar, "updated", true)
				end
			end
		end
	end
end

-- Make sure not to delete any unexpired test bars
local function UpdateTestBars(bg)
	for i = 1, 10 do
		local bar = Nest_GetBar(bg, ">>Test<<" .. string.format("%02d", i))
		if bar then
			local timeLeft = Nest_GetTimes(bar)
			if timeLeft and (timeLeft > 0) then Nest_SetAttribute(bar, "updated", true) end
		end
	end
end

-- Show tooltip when entering a bar group anchor
local function Anchor_OnEnter(anchor, bgName)
	if GetCVar("UberTooltips") == "1" then
		GameTooltip_SetDefaultAnchor(GameTooltip, anchor)
	else
		GameTooltip:SetOwner(anchor, "ANCHOR_BOTTOMRIGHT")
	end
	local bg, bgType, attachment = Nest_GetBarGroup(bgName), "Custom Bar Group", nil
	if bg then
		if Nest_GetBarGroupAttribute(bg, "isAuto") then bgType = "Auto Bar Group" end
		attachment = Nest_GetBarGroupAttribute(bg, "attachment")
	end
	GameTooltip:AddDoubleLine("Raven", bgType)
	if attachment then
		GameTooltip:AddLine('|cffffff00Attached|r to "' .. attachment .. '"')
		GameTooltip:AddLine("|cffffff00Left-click|r to unattach and then move the anchor")
	else
		GameTooltip:AddLine("|cffffff00Left-click|r to move the anchor")
	end
	GameTooltip:AddLine("|cffffff00Right-click|r to lock and hide the anchor")
	GameTooltip:AddLine("|cffffff00Shift-left-click|r to toggle test bars")
	GameTooltip:AddLine("|cffffff00Shift-right-click|r to toggle direction")
	GameTooltip:AddLine("|cffffff00Alt-left-click|r to open/close options menu")
	GameTooltip:AddLine("|cffffff00Alt-right-click|r to change configuration")
	GameTooltip:Show()
end

-- Hide tooltip when leaving a bar group anchor
local function Anchor_OnLeave(anchor, bgName)
	GameTooltip:Hide()
end

-- Callback function for tracking location of the bar group
local function Anchor_Moved(anchor, bgName)
	local bp = MOD.db.profile.BarGroups[bgName]
	if IsOn(bp) then
		local bg = Nest_GetBarGroup(bgName)
		if bg then
			bp.pointX, bp.pointY = Nest_GetAnchorPoint(bg) -- returns fraction of display in x and y
			if bp.anchor then bp.anchor = false end -- no longer anchored to other bar groups			
			if bp.linkSettings then
				local settings = MOD.db.global.Settings[bp.name] -- when updating a bar group with linked settings always overwrite position
				if settings then settings.pointX = bp.pointX; settings.pointY = bp.pointY; settings.anchor = bp.anchor end
			end
			Anchor_OnLeave(anchor) -- turn off tooltip while moving the anchor
			MOD.updateOptions = true -- if options panel is open then update it in case viewing position info
		end
		return
	end
end

-- Callback function for when a bar group anchor is clicked with a modifier key down
-- Shift left click is test bars, right click is "toggle lock and hide",
local function Anchor_Clicked(anchor, bgName, button)
	local shiftLeftClick = (button == "LeftButton") and (IsShiftKeyDown() == 1)
	local shiftRightClick = (button == "RightButton") and (IsShiftKeyDown() == 1)
	local altLeftClick = (button == "LeftButton") and (IsAltKeyDown() == 1)
	local altRightClick = (button == "RightButton") and (IsAltKeyDown() == 1)
	local rightClick = (button == "RightButton")

	local bp = MOD.db.profile.BarGroups[bgName]
	if IsOn(bp) then
		if shiftLeftClick then -- test bars
			TestBarGroup(bp)
		elseif shiftRightClick then -- toggle grow up/down
			bp.growDirection = not bp.growDirection
		elseif altLeftClick then -- toggle options menu
			MOD:OptionsPanel()
		elseif altRightClick then -- toggle configuration
			bp.configuration = bp.configuration + 1
			if bp.configuration == (Nest_MaxBarConfiguration + 1) then bp.configuration = 1 end
			if bp.configuration == (Nest_MaxIconConfiguration + 1) then bp.configuration = Nest_MaxBarConfiguration + 1 end
		elseif rightClick then -- lock and hide
			bp.locked = true
		end
		MOD:UpdateBarGroup(bp)
		MOD.updateOptions = true -- if options panel is open then update it in case viewing configuration info
		return
	end
end

-- Update linked settings. If dir is true then update the shared settings, otherwise update the bar group.
-- Also, if dir is true, create a linked settings table if one doesn't yet exist.
local function UpdateLinkedSettings(bp, dir)
	local settings = MOD.db.global.Settings[bp.name]
	if not settings then
		if not dir then return end
		settings = {}
		MOD.db.global.Settings[bp.name] = settings
	end

	local p, q = settings, bp
	if dir then p = bp; q = settings end

	for n in pairs(settingsTemplate) do q[n] = p[n] end -- copy every setting in the template
	q.pointX = p.pointX; q.pointY = p.pointY -- always copy the location
end

-- Update a linked filter list. If dir is true then update the shared list, otherwise update the bar group's list.
-- Also, if dir is true, create a linked filter list if one doesn't yet exist.
local function UpdateLinkedFilter(bp, dir, filterType)
	local shared = MOD.db.global["Filter" .. filterType]
	local bgname = "filter" ..  filterType .. "List"

	if not shared[bp.name] then
		if not dir or not bp[bgname] or not next(bp[bgname], nil) then return end
		shared[bp.name] = {}
	end
	
	if not bp[bgname] then bp[bgname] = {} end

	local p, q = shared[bp.name], bp[bgname]
	if dir then p = bp[bgname]; q = shared[bp.name] end

	for _, v in pairs(q) do if not p[v] then q[v] = nil end end -- delete any keys in q not in p
	for _, v in pairs(p) do if not q[v] then q[v] = v end end -- copy everything from p to q
end

-- Initialize a bar group from a shared filter list, if any.
-- This function is called whenever filterLink is changed.
function MOD:InitializeFilterList(bp, filterType)
	if bp and bp["filter" .. filterType] and bp["filter" .. filterType .. "Link"] then
		UpdateLinkedFilter(bp, false, filterType) -- use the shared settings, if any, for a linked layout
	end
end

-- Get spell associated with a bar
function MOD:GetAssociatedSpellForBar(bar)
	if bar.barType == "Notification" then
		if bar.associatedSpell and (bar.associatedSpell ~= "") then return bar.associatedSpell end
		if bar.action then
			local sp = MOD:GetConditionSpell(bar.action)
			if sp and (sp ~= "") then return sp end
		end
		return nil
	end
	return bar.action
end

-- Get icon associated with a bar
function MOD:GetIconForBar(b)
	if b then
		if b.barType == "Notification" then
			if not b.notIcon then
				local sp = MOD:GetAssociatedSpellForBar(b)
				if sp then return MOD:GetIcon(sp) end
			end
			return defaultNotificationIcon
		elseif b.action then return MOD:GetIcon(b.action) end
	end
	return nil
end

-- Get color for the spell associated with a bar
-- If no spell is set then, if default is true, return the correct default color
local function GetAssociatedSpellColorForBar(bar, default)
	local c = nil
	if bar.barType == "Notification" then
		if not bar.notColor then
			if bar.associatedSpell and (bar.associatedSpell ~= "") then
				c = MOD:GetColor(bar.associatedSpell, nil)
			elseif bar.action then
				local sp = MOD:GetConditionSpell(bar.action)
				if sp and (sp ~= "") then c = MOD:GetColor(sp, nil) end
			end
		end
		if default and not c then c = MOD.db.global.DefaultNotificationColor end
	else
		if default then c = MOD:GetColor(bar.action, bar.barType) else c = MOD:GetColor(bar.action, nil) end
	end
	return c
end

-- Set color for the spell associated with a bar
local function SetAssociatedSpellColorForBar(bar, c)
	if bar.barType == "Notification" then
		if not bar.notColor then
			if bar.associatedSpell and (bar.associatedSpell ~= "") then
				MOD:SetColor(bar.associatedSpell, c)
			elseif bar.action then
				local sp = MOD:GetConditionSpell(bar.action)
				if sp and (sp ~= "") then MOD:SetColor(sp, c) end
			end
		end
	else
		MOD:SetColor(bar.action, c)
	end
end

-- Return the bar's color after checking if cached or if alternate color is specified; if not found then return a suitable default
-- If edit is true then return an editable copy of the color if it is one of the defaults
function MOD:GetBarColor(bar, edit)
	local c = bar.color
	if c then return c end -- override of cached color is already defined for this bar
	c = GetAssociatedSpellColorForBar(bar, true) -- with default true this must return a valid color
	if edit == true then return CopyColor(c) end
	return c
end

-- Set the bar's color in the cache or alternate color.
-- If using default then set an alternate color and, if linked, save in the color cache.
function MOD:SetBarColor(bar, r, g, b, a)
	local c = bar.color -- check if already using an alternate color
	if c then c.r = r; c.g = g; c.b = b; c.a = a; return end
	c = GetAssociatedSpellColorForBar(bar, false)
	if c then c.r = r; c.g = g; c.b = b; c.a = a; return end
	local tc = { r = r, g = g, b = b, a = a } -- need to create an override for this bar
	bar.color = tc
	if not bar.colorLink then SetAssociatedSpellColorForBar(bar, tc) end
end

-- Either link or decouple the bar's color from the color cache for it's associated action.
function MOD:LinkBarColor(bar)
	if not bar.colorLink then -- delete the current color setting to revert back to either color cache or default for the bar type
		bar.color = nil
	else -- decouple from the color cache
		local c = MOD:GetBarColor(bar, false) -- get current color assigned to the bar
		bar.color = CopyColor(c)
	end
end

-- Get the right color for the bar based on bar group settings
function MOD:GetColorForBar(bg, bar, btype)
	local bt, ba, b, c = bar.barType, bar.action, nil, nil
	if bg.barColors == "Class" then
		if bt == "Buff" then b = MOD.BuffTable[ba] end
		if bt == "Debuff" then b = MOD.DebuffTable[ba] end
		if bt == "Cooldown" then b = MOD.CooldownTable[ba] end
		if b and b.class then
			if CUSTOM_CLASS_COLORS then c = CUSTOM_CLASS_COLORS[b.class] else c = RAID_CLASS_COLORS[b.class] end
			if c then c.a = 1; return c end
		end
	end
	if bg.barColors == "Spell" then return MOD:GetBarColor(bar, false) end

	local cc = not bg.useDefaultColors -- indicates the bar group has overrides for standard colors
	c = cc and bg.buffColor or MOD.db.global.DefaultBuffColor -- use this as default in case unrecognized bar type
	if bt == "Debuff" then
		c = cc and bg.debuffColor or MOD.db.global.DefaultDebuffColor
		if btype then
			if btype == "Poison" then c = cc and bg.poisonColor or MOD.db.global.DefaultPoisonColor end
			if btype == "Curse" then c = cc and bg.curseColor or MOD.db.global.DefaultCurseColor end
			if btype == "Magic" then c = cc and bg.magicColor or MOD.db.global.DefaultMagicColor end
			if btype == "Disease" then c = cc and bg.diseaseColor or MOD.db.global.DefaultDiseaseColor end
		end
	end
	if bt == "Cooldown" then c = cc and bg.cooldownColor or MOD.db.global.DefaultCooldownColor end
	if bt == "Notification" then c = cc and bg.notificationColor or MOD.db.global.DefaultNotificationColor end
	c.a = 1
	return c
end

-- Initialize bar group in LibBars and set default values from those set in profile
function MOD:InitializeBarGroup(bp, offsetX, offsetY)
	local bg = Nest_GetBarGroup(bp.name)
	if not bg then bg = Nest_CreateBarGroup(bp.name) end
	if bp.linkSettings then UpdateLinkedSettings(bp, false) end
	if bp.auto then -- initialize the auto bar group filter lists
		if bp.filterBuff and bp.filterBuffLink then UpdateLinkedFilter(bp, false, "Buff") end -- shared settings for buffs
		if bp.filterDebuff and bp.filterDebuffLink then UpdateLinkedFilter(bp, false, "Debuff") end -- shared settings for debuffs
		if bp.filterCooldown and bp.filterCooldownLink then UpdateLinkedFilter(bp, false, "Cooldown") end -- shared settings for buffs
	end
	if not bp.pointX or not bp.pointY then bp.pointX = 0.5 + (offsetX / 600); bp.pointY = 0.5 + (offsetY / 600) end
	MOD:SetBarGroupPosition(bp)
	Nest_SetBarGroupCallbacks(bg, Anchor_Moved, Anchor_Clicked, Anchor_OnEnter, Anchor_OnLeave)
end

-- Initialize a bar group from linked settings, if any, and always update the bar group location.
-- This function is called whenever linkSettings is changed.
function MOD:InitializeBarGroupSettings(bp)
	if bp and bp.enabled then
		if bp.linkSettings then UpdateLinkedSettings(bp, false) end
		MOD:SetBarGroupPosition(bp)
	end
end

-- Save bar group settings into the linked settings.
function MOD:SaveBarGroupSettings(bp) UpdateLinkedSettings(bp, true) end

-- Validate and update a bar group's display position. If linked, also update the position in the linked settings.
function MOD:SetBarGroupPosition(bp)
	local scale = bp.useDefaultDimensions and MOD.db.global.Defaults.scale or bp.scale
	local bg = Nest_GetBarGroup(bp.name)
	if bg then Nest_SetAnchorPoint(bg, bp.pointX, bp.pointY, scale or 1) end
	if bp.linkSettings then
		local settings = MOD.db.global.Settings[bp.name] -- when updating a bar group with linked settings always overwrite position
		if settings then settings.pointX = bp.pointX; settings.pointY = bp.pointY end
	end
end

-- Returns whether solo, party or raid
local function PartyInfo()
	if GetNumRaidMembers() > 0 then return "raid" end
	if GetNumPartyMembers() > 0 then return "party" end
	return "solo"
end

-- Bar sorting functions, assumes sort order was built into name
local function SortAlphaUp(a, b) return a.name > b.name end
local function SortAlphaDown(a, b) return a.name < b.name end
local function SortTimeLeftUp(a, b) if a.value ~= b.value then return a.value > b.value else return a.name > b.name end end
local function SortTimeLeftDown(a, b) if a.value ~= b.value then return a.value < b.value else return a.name < b.name end end
local function SortDurationUp(a, b) if a.maxValue ~= b.maxValue then return a.maxValue > b.maxValue else return a.name > b.name end end
local function SortDurationDown(a, b) if a.maxValue ~= b.maxValue then return a.maxValue < b.maxValue else return a.name < b.name end end

-- Update a bar group in LibBars with the current values in the profile
function MOD:UpdateBarGroup(bp)
	if bp.enabled then
		if bp.linkSettings then UpdateLinkedSettings(bp, true) end -- update shared settings in a linked bar group
		if bp.auto then -- update auto bar group filter lists
			if bp.filterBuff and bp.filterBuffLink then UpdateLinkedFilter(bp, true, "Buff") end -- shared settings for buffs
			if bp.filterDebuff and bp.filterDebuffLink then UpdateLinkedFilter(bp, true, "Debuff") end -- shared settings for debuffs
			if bp.filterCooldown and bp.filterCooldownLink then UpdateLinkedFilter(bp, true, "Cooldown") end -- shared settings for buffs
		end

		local bg = Nest_GetBarGroup(bp.name)
		if not bg then MOD:InitializeBarGroup(bp); bg = Nest_GetBarGroup(bp.name) end

		local dp, fp = bp, bp -- select which table to use for dimension and font/texture options
		if bp.useDefaultDimensions then dp = MOD.db.global.Defaults end
		if bp.useDefaultFontsAndTextures then fp = MOD.db.global.Defaults end		
		local fgtexture = media:Fetch("statusbar", fp.texture)
		local bgtexture = fgtexture
		if fp.bgtexture then bgtexture = media:Fetch("statusbar", fp.bgtexture) end
		Nest_SetBarGroupLock(bg, bp.locked)
		Nest_SetBarGroupBarLayout(bg, dp.barWidth, dp.barHeight, dp.iconSize, dp.scale, dp.spacingX, dp.spacingY, dp.iconOffsetX, dp.iconOffsetY,
			bp.labelOffset, bp.labelInset, bp.labelWrap, bp.labelAlign, bp.timeOffset, bp.timeInset, bp.iconOffset, bp.iconInset,
			bp.configuration, bp.growDirection, bp.wrap, bp.wrapDirection, bp.snapCenter)
		Nest_SetBarGroupLabelFont(bg, media:Fetch("font", fp.labelFont), fp.labelFSize, fp.labelAlpha, fp.labelColor, fp.labelOutline)
		Nest_SetBarGroupTimeFont(bg, media:Fetch("font", fp.timeFont), fp.timeFSize, fp.timeAlpha, fp.timeColor, fp.timeOutline)
		Nest_SetBarGroupIconFont(bg, media:Fetch("font", fp.iconFont), fp.iconFSize, fp.iconAlpha, fp.iconColor, fp.iconOutline)
		Nest_SetBarGroupTextures(bg, fgtexture, fp.alpha, bgtexture, fp.bgAlpha, not bp.showNoDurationBackground,
			fp.fgSaturation, fp.fgBrightness, fp.bgSaturation, fp.bgBrightness)
		Nest_SetBarGroupVisibles(bg, not bp.hideIcon, not bp.hideClock, not bp.hideBar, not bp.hideSpark, not bp.hideLabel, not bp.hideValue)
		Nest_SetBarGroupAttribute(bg, "isAuto", bp.auto) -- save for tooltip
		Nest_SetBarGroupAttribute(bg, "attachment", bp.anchor) -- save for tooltip
		Nest_SetBarGroupTimeFormat(bg, bp.timeFormat, bp.timeSpaces, bp.timeCase)
		local sf = "alpha"; if bp.sor == "T" then sf = "time" elseif bp.sor == "D" then sf = "duration" end
		Nest_BarGroupSortFunction(bg, sf, bp.reverseSort)
	else
		MOD:ReleaseBarGroup(bp)
	end
end

-- Update the positions of all anchored bar groups plus make sure valid positions in all bar groups
local function UpdatePositions()
	for _, bp in pairs(MOD.db.profile.BarGroups) do -- update bar group positions including relative ones if anchored
		if IsOn(bp) then
			local bg = Nest_GetBarGroup(bp.name)
			if bg then
				if not bp.pointX or not bp.pointY then -- if not valid then move to center
					bp.pointX = 0.5; bp.pointY = 0.5
					Nest_SetAnchorPoint(bg, bp.pointX, bp.pointY, bp.scale or 1)
				end	
				if bp.anchor then
					local abp = MOD.db.profile.BarGroups[bp.anchor]
					if IsOn(abp) and abp.enabled then -- make sure the anchor is actually around to attach
						Nest_SetRelativeAnchorPoint(bg, bp.anchor, bp.anchorX, bp.anchorY, bp.anchorLastBar)
					end
				else
					Nest_SetRelativeAnchorPoint(bg, nil) -- reset the relative anchor point if none set
				end
				bp.pointX, bp.pointY = Nest_GetAnchorPoint(bg) -- get latest position so always accurate
			end
		end
	end
end

-- Update all the bar groups, this is necessary when changing stuff that can affect bars in multiple groups (e.g., buff colors and labels)
function MOD:UpdateAllBarGroups()
	for _, bp in pairs(MOD.db.profile.BarGroups) do -- update for changed bar group settings
		if IsOn(bp) then MOD:UpdateBarGroup(bp) end
	end
end

-- Lock or unlock all bar groups
function MOD:LockBarGroups(lock)
	for _, bp in pairs(MOD.db.profile.BarGroups) do
		if IsOn(bp) then bp.locked = lock end
	end
	MOD:UpdateAllBarGroups()
end

-- Toggle locking of bar groups 
function MOD:ToggleBarGroupLocks()
	-- Look in the profile table to determine current state
	local anyLocked = false
	for _, bp in pairs(MOD.db.profile.BarGroups) do
		if IsOn(bp) and bp.locked then anyLocked = true break end
	end
	-- Now go back through and set all to same state (if any locked then unlock all)
	MOD:LockBarGroups(not anyLocked)
end

-- Release all the bars in the named bar group in LibBars
function MOD:ReleaseBarGroup(bp)
	if bp then 
		local bg = Nest_GetBarGroup(bp.name)
		if bg then Nest_DeleteBarGroup(bg) end
	end
end

-- Show tooltip when entering a bar
local function Bar_OnEnter(frame, bgName, barName, ttanchor)
	if GetCVar("UberTooltips") == "1" then
		GameTooltip_SetDefaultAnchor(GameTooltip, frame)
	else
		GameTooltip:SetOwner(frame, ttanchor)
	end
	local bg = Nest_GetBarGroup(bgName)
	if not bg then return end
	local bar = Nest_GetBar(bg, barName)
	if not bar then return end	
	local tt = Nest_GetAttribute(bar, "tooltipType")
	local id = Nest_GetAttribute(bar, "tooltipID")
	local unit = Nest_GetAttribute(bar, "tooltipUnit")
	
	if tt == "text" then
		GameTooltip:SetText(id)
	elseif tt == "inventory" then
		local slot = GetInventorySlotInfo(id)
		if slot then
			GameTooltip:SetInventoryItem("player", slot)
		end
	elseif (tt == "spell link") or (tt == "item link") then
		GameTooltip:SetHyperlink(id)
	elseif tt == "buff" then
		GameTooltip:SetUnitBuff(unit, id)
	elseif tt == "debuff" then
		GameTooltip:SetUnitDebuff(unit, id)
	elseif tt == "vehicle buff" then
		GameTooltip:SetUnitBuff("vehicle", id)
	elseif tt == "vehicle debuff" then
		GameTooltip:SetUnitDebuff("vehicle", id)
	elseif tt == "tracking" then
		GameTooltip:SetText(id) -- id is localized name of tracking type
	elseif tt == "spell" then
		if id == "Eclipse" then
			if MOD.eclipseTab and MOD.eclipseIndex then GameTooltip:SetTalent(MOD.eclipseTab, MOD.eclipseIndex) end
		else
			GameTooltip:SetText(id)
		end
	elseif tt == "totem" then
		if id == "Fire Totem" then
			GameTooltip:SetTotem(1)
		elseif id == "Earth Totem" then
			GameTooltip:SetTotem(2)
		elseif id == "Water Totem" then
			GameTooltip:SetTotem(3)
		elseif id == "Air Totem" then
			GameTooltip:SetTotem(4)
		end
	elseif tt == "notification" then
		GameTooltip:AddDoubleLine(id, "Notification")
		local ct = MOD.db.profile.Conditions[MOD.myClass]
		if ct then
			local c = ct[unit]
			if IsOn(c) and c.tooltip then GameTooltip:AddLine(MOD:GetConditionText(c.name), 1, 1, 1, true) end
		end
		GameTooltip:Show()
	end
end

-- Hide tooltip when leaving a bar
local function Bar_OnLeave(frame, bgName, barName, ttanchor)
	GameTooltip:Hide()
end

-- Handle right clicking on a bar to remove a buff and shift-right clicking to disable a bar
local function Bar_OnClick(frame, bgName, barName, button)
	local bg = Nest_GetBarGroup(bgName)
	if not bg then return end
	local bar = Nest_GetBar(bg, barName)
	if not bar then return end	
	local tt = Nest_GetAttribute(bar, "tooltipType")
	local id = Nest_GetAttribute(bar, "tooltipID")
	local unit = Nest_GetAttribute(bar, "tooltipUnit")
	local shiftRightClick = (button == "RightButton") and (IsShiftKeyDown() == 1)
	
	if shiftRightClick then -- add this buff, debuff or cooldown to the bar group's appropriate filter list
		local bp = Nest_GetAttribute(bar, "saveBarGroup")
		local b = Nest_GetAttribute(bar, "saveBar") -- only valid for custom bar groups that store bar tables permanently!
		if bp and bp.auto then -- auto bar group so add to associated filter list
			local barType = Nest_GetAttribute(bar, "saveBarType")
			local action = Nest_GetAttribute(bar, "saveBarAction")
			local fb = "filter" .. barType
			if bp[fb] then
				local listName, linked = fb .. "List", fb .. "Link"
				local filterList = bp[listName]
				if not filterList then filterList = {}; bp[listName] = filterList end
				filterList[action] = action
				print(string.format("Raven: \"%s\" added to %s filter list for bar group \"%s\"", action, string.lower(barType), bp.name))
				if bp[linked] then -- have to save to the shared filter list too
					local shared = MOD.db.global["Filter" .. barType]
					local linkList = shared[bp.name]
					if not linkList then linkList = {}; shared[bp.name] = linkList end -- create shared list if necessary
					linkList[action] = action
				end
			end
		elseif b then -- custom bar group so disable the bar
			if b then b.enableBar = false end
		end
		MOD.updateOptions = true -- if options panel is open then update it in case viewing filter info
	elseif (button == "RightButton") and (unit == "player") then -- right clicking can turn off player buffs
		if tt == "inventory" then
			if id == "MainHandSlot" then
				CancelItemTempEnchantment(1)
			elseif id == "SecondaryHandSlot" then
				CancelItemTempEnchantment(2)
			end
		elseif tt == "totem" then
			if id == "Fire Totem" then
				DestroyTotem(1)
			elseif id == "Earth Totem" then
				DestroyTotem(2)
			elseif id == "Water Totem" then
				DestroyTotem(3)
			elseif id == "Air Totem" then
				DestroyTotem(4)
			end
		elseif tt == "buff" and GetSpellName then -- only supported pre-Cataclysm
			CancelUnitBuff(unit, id)
		end
	elseif (button == "LeftButton") and (unit == "player") and (tt == "tracking") then
		if GameTooltip:GetOwner() == frame then GameTooltip:Hide() end
		ToggleDropDownMenu(1, nil, MiniMapTrackingDropDown, frame, 0, 0)
	end
end

-- Return true if time and duration pass a bar group's filters
local function CheckTimeAndDuration(bp, timeLeft, duration)
	if (timeLeft == 0) and (duration == 0) then -- test for unlimited duration
		if not bp.showNoDuration then return false end
	else
		if bp.showNoDuration and bp.showOnlyNoDuration then return false end
		if bp.checkDuration and bp.filterDuration then
			if bp.minimumDuration then if duration < bp.filterDuration then return false end
			elseif duration >= bp.filterDuration then return false end
		end
		if bp.checkTimeLeft and bp.filterTimeLeft then
			if bp.minimumTimeLeft then if timeLeft < bp.filterTimeLeft then return false end
			elseif timeLeft >= bp.filterTimeLeft then return false end
		end
	end
	return true
end

-- Manage a bar, creating one if not currently active, otherwise updating as necessary
local function UpdateBar(bp, bg, b, icon, timeLeft, duration, count, btype, ttType, ttID, ttUnit)
	local bar, barname, label, src = nil, b.barLabel .. b.uniqueID, b.barLabel, b.barSource

	if (b.barType == "Buff") or (b.barType == "Debuff") then barname = barname .. ttID end -- distinguish between same name auras in different slots
	if bp.sor == "C" then -- prefixes are added to the barname to facilitate sorting
		if not src or (src == "Racial") or (src == "Spell") or (src == "Detected") then src = "ZZZZ" end
		barname = "ZZZZ" .. string.upper(src) .. barname
	end
	if bp.sor == "X" then barname = string.format("%05d ", b.sorder) .. barname end
	
	local maxTime = duration
	if bp.setDuration then maxTime = bp.uniformDuration end -- override with uniform duration for all bars in group

	local c = MOD:GetColorForBar(bp, b, btype)
	local iconCount = nil
	if count and (count > 1) then
		if bp.hideIcon then label = string.format("%s (%d)", label, count) else iconCount = count end
	end
	
	bar = Nest_GetBar(bg, barname)
	if not (bp.showNoDuration and bp.showOnlyNoDuration) and not ((timeLeft == 0) and (duration == 0)) then -- bar with duration
		if bar and Nest_IsTimer(bar) then -- existing timer bar
			local oldTimeLeft, oldDuration, oldMaxTime = Nest_GetTimes(bar)
			if (duration ~= oldDuration) or maxTime ~= oldMaxTime or (math.abs(timeLeft - oldTimeLeft) > 0.5) then
				Nest_StartTimer(bar, timeLeft, duration, maxTime) -- update if the bar is out of sync
			end
		else
			if bar then Nest_DeleteBar(bg, bar); bar = nil end
			bar = Nest_CreateBar(bg, barname)
			if bar then Nest_StartTimer(bar, timeLeft, duration, maxTime) end
		end
	elseif bp.showNoDuration or (b.barType == "Notification") then -- bars without duration
		if bar and Nest_IsTimer(bar) then Nest_DeleteBar(bg, bar); bar = nil end
		if not bar then bar = Nest_CreateBar(bg, barname) end
		if b.barType == "Notification" then Nest_SetFlash(bar, b.flash) end
	end
	if bar then
		Nest_SetAttribute(bar, "updated", true) -- for mark/sweep bar deletion
		Nest_SetLabel(bar, label)
		Nest_SetIcon(bar, icon)
		local bc = (bp.bgColors == "Custom") and bp.bgColor or c
		Nest_SetColor(bar, c.r, c.g, c.b, c.a, bc.r, bc.g, bc.b, bc.a)

		if not bp.hideIcon then
			Nest_SetCount(bar, iconCount) -- set the icon text to this count or blank if nil
			Nest_SetAttribute(bar, "tooltipType", ttType)
			Nest_SetAttribute(bar, "tooltipID", ttID)
			Nest_SetAttribute(bar, "tooltipUnit", ttUnit)
			Nest_SetAttribute(bar, "saveBar", b) -- not valid in auto bar group since it then points to a local not a permanent table!
			
			local sbg, sbt, sba = nil, nil, nil
			if b.barType ~= "Notification" then sbg = bp; sbt = b.barType; sba = b.action end
			Nest_SetAttribute(bar, "saveBarGroup", sbg)
			Nest_SetAttribute(bar, "saveBarType", sbt)
			Nest_SetAttribute(bar, "saveBarAction", sba)
			
			local click, onEnter, onLeave = nil, nil, nil
			if bp.combatClicks or not MOD.status.inCombat then click = Bar_OnClick end
			if bp.showTooltips and (bp.combatTips or not MOD.status.inCombat) then onEnter = Bar_OnEnter; onLeave = Bar_OnLeave end
			Nest_SetCallbacks(bar, click, onEnter, onLeave)
		end
		
		local alpha = 1 -- set flash and adjust alpha for special effects: combat fader, flash before expiring, fade after start
		if bp.combatFader and MOD.status.inCombat then alpha = bp.combatAlpha end
		if (IsOn(b.flashBar) and (b.flashBar == MOD:CheckCondition(b.flashCondition))) or -- conditional flashing
				(bp.flashExpiring and Nest_IsTimer(bar) and (timeLeft < bp.flashTime)) then -- bar group flash when expiring
			Nest_SetFlash(bar, true)
		elseif IsOn(b.fadeBar) then -- conditional fade for bars is higher priority than bar group setting for delayed fade
			if b.fadeBar == MOD:CheckCondition(b.fadeCondition) then
				if b.fadeAlpha then alpha = alpha * b.fadeAlpha end -- if fade condition is true then set to fade alpha
			else
				if b.normalAlpha then alpha = alpha * b.normalAlpha end -- otherwise set to normal alpha
			end
		elseif bp.fade then
			local _, _, _, startTime = Nest_GetTimes(bar) -- get time bar was created
			if GetTime() > (startTime + bp.fadeTime) then -- check if fade time has run out
				alpha = alpha * bp.bgAlpha -- after fade delay then lower bar alpha to that of background
			end
		else
			Nest_SetFlash(bar, false)
		end		
		Nest_SetAlpha(bar, alpha)
	end
end

-- Compare caster to enforce "cast by" restrictions
local function CheckCastBy(caster, cb)
	local isMine = (caster == "player")
	local isPet = (caster == "pet")
	local isOther = not (isMine or isPet)
	if not cb then cb = "Player" end
	return (((cb == "Player") and isMine) or ((cb == "Pet") and isPet) or ((cb == "Other") and isOther) or (cb == "Anyone"))
end

-- Return the source for a buff, debuff or cooldown
local function GetSpellSource(t)
	if not t then return "Unknown" end
	local s = classNames[t.class]
	if not s and t.race then s = "Racial" end
	if not s then s = "Detected" end
	return s
end

-- Check for detected buffs and create bars for them in the specified bar group
-- Detected auras that don't match current bar group settings may need to be added later if the settings change 
local function DetectNewBuffs(unit, n, aura, isBuff, bp, bg)
	if bp.filterBuff and bp.filterBuffList and bp.filterBuffList[n] then return end -- check if black listed
	local bt = MOD.BuffTable
	if not bt[n] then -- not in the tables already
		MOD:SetIcon(n, aura[8]) -- cache the icon since this is first time ever seen
		bt[n] = { det = true } -- newly detected aura goes into auras table
	end
		
	local label = MOD:GetLabel(n) -- check if there is a cached label for this action
	local checkTracking = (aura[4] ~= "Tracking") or bp.detectTracking
	if CheckCastBy(aura[6], bp.detectBuffsCastBy) and CheckTimeAndDuration(bp, aura[2], aura[5]) and checkTracking then -- apply filters
		local b = detectedBar
		b.action = n
		b.barLabel = label
		b.barType = "Buff"
		b.uniqueID = "Buff"
		b.barSource = GetSpellSource(bt[n])
		UpdateBar(bp, bg, b, aura[8], aura[2], aura[5], aura[3], aura[4], aura[11], aura[12], unit)
	end
end

-- Check for detected debuffs and create bars for them in the specified bar group 
local function DetectNewDebuffs(unit, n, aura, isBuff, bp, bg)
	if bp.filterDebuff and bp.filterDebuffList and bp.filterDebuffList[n] then return end -- check if black listed
	local bt = MOD.DebuffTable
	if not bt[n] then -- not in the tables already
		MOD:SetIcon(n, aura[8]) -- cache the icon since this is first time seen
		bt[n] = { det = true } -- newly detected aura goes into auras table
	end
		
	local label = MOD:GetLabel(n)
	
	if CheckCastBy(aura[6], bp.detectDebuffsCastBy) and CheckTimeAndDuration(bp, aura[2], aura[5]) then
		local b = detectedBar
		b.action = n
		b.barLabel = label
		b.barType = "Debuff"
		b.uniqueID = "Debuff"
		b.barSource = GetSpellSource(bt[n])
		UpdateBar(bp, bg, b, aura[8], aura[2], aura[5], aura[3], aura[4], aura[11], aura[12], unit)
	end
end

-- Check if a cooldown is of right type for the specified bar group
local function CheckCooldownType(cd, bp)
	local other, t, s = true, cd[5], cd[6]
	if (t == "spell link") or (t == "spell") then
		other = false
		if bp.detectSpellCooldowns then return true end
	end
	if (t == "inventory") and ((s == "Trinket0Slot") or (s == "Trinket1Slot")) then
		other = false
		if bp.detectTrinketCooldowns then return true end
	end
	if t == "text" then -- might be a potion or elixir
		if (s == "Shared Potion Cooldown") or (s == "Shared Elixir Cooldown") then
			other = false
			if bp.detectPotionCooldowns then return true end
		end
	end
	if other and bp.detectOtherCooldowns then return true end
	return false
end

-- Return true if cooldown is not one of the special case shared ones (e.g., paladin judgements)
local function CheckSharedCooldowns(n, b, bp)
	if bp.detectSharedJudgements then
		if n == "Judgement of Wisdom" or n == "Judgement of Light" then return false end
		if n == "Judgement of Justice" then b.barLabel = "Judgements" end
	end
	if bp.detectSharedFrostTraps then
		if n == "Freezing Trap" or n == "Freezing Arrow" then return false end
		if n == "Frost Trap" then b.barLabel = "Frost Traps" end
	end
	if bp.detectSharedShocks then
		if n == "Frost Shock" or n == "Flame Shock" then return false end
		if n == "Earth Shock" then b.barLabel = "Shocks" end
	end
	return true
end

-- Automatically generate rune cooldown bars for all six rune slots
local runeSlotPrefix = { "(1)  ", "(2)  ", "(5)  ", "(6)  ",  "(3)  ", "(4)  " }
local function AutoRuneBars(bp, bg)
	if MOD.myClass ~= "DEATHKNIGHT" then return end
	for i = 1, 6 do
		local rune = MOD.runeSlots[i]
		local icon = MOD.runeIcons[rune.rtype]
		local b = detectedBar
		b.action = MOD.runeTypes[rune.rtype]
		b.barLabel = runeSlotPrefix[i] .. b.action
		b.barType = "Cooldown"
		b.uniqueID = "Cooldown"
		b.barSource = "Detected"
		if rune.ready then -- generate ready bar with no duration
			if CheckTimeAndDuration(bp, 0, 0) then
				UpdateBar(bp, bg, b, icon, 0, 0, nil, nil, "text", b.action, nil)
			end
		else -- generate cooldown timer bar
			local timeLeft = rune.duration - (GetTime() - rune.start)
			if CheckTimeAndDuration(bp, timeLeft, rune.duration) then
				UpdateBar(bp, bg, b, icon, timeLeft, rune.duration, nil, nil, "text", b.action, nil)
			end
		end
	end
end

-- Check if there are detected cooldowns and conditionally create bars for them in the specified bar group
local function DetectNewCooldowns(n, cd, bp, bg)
	if bp.filterCooldown and bp.filterCooldownList and bp.filterCooldownList[n] then return end -- check if black listed
	local cdt = MOD.CooldownTable	
	if not cdt[n] then -- not in the cooldown tables already
		MOD:SetIcon(n, cd[2]) -- cache the icon since this is first time seen
		cdt[n] = { det = true } -- newly detected cooldown goes into cooldowns table
	end		
	local label = MOD:GetLabel(n)	
	if CheckCastBy(cd[7], bp.detectCooldownsBy) and CheckCooldownType(cd, bp) and CheckTimeAndDuration(bp, cd[1], cd[4]) then
		local b = detectedBar
		b.action = n
		b.barLabel = label
		b.barType = "Cooldown"
		b.uniqueID = "Cooldown"
		b.barSource = GetSpellSource(cdt[n])
		if CheckSharedCooldowns(n, b, bp) then
			UpdateBar(bp, bg, b, cd[2], cd[1], cd[4], nil, nil, cd[5], cd[6], nil)
		end
	end
end

-- Update all the bars in the bar group (bp), causing them to appear in the display bar group (bg).
-- The show/hide conditions are tested in this function so they are depending on the updating bar group.
local function UpdateBarGroupBars(bp, bg)
	local pst, stat = PartyInfo(), MOD.status	
	local show = (((pst == "solo") and bp.showSolo) or ((pst == "party") and bp.showParty) or ((pst == "raid") and bp.showRaid)) and
		((bp.showCombat and stat.inCombat) or (bp.showOOC and not stat.inCombat)) and (not stat.isResting or bp.showResting) and
		(not stat.isFishing or bp.showFishing) and (not stat.isMounted or bp.showMounted) and 
		(not bp.checkCondition or IsOff(bp.condition) or MOD:CheckCondition(bp.condition))
	if show then
		if bp.auto then -- if auto bar group then detect new auras and cooldowns
			if bp.detectBuffs then MOD:IterateAuras(bp.detectBuffsMonitor, DetectNewBuffs, true, bp, bg) end
			if bp.detectDebuffs then MOD:IterateAuras(bp.detectDebuffsMonitor, DetectNewDebuffs, false, bp, bg) end
			if bp.detectCooldowns then MOD:IterateCooldowns(DetectNewCooldowns, bp, bg) end
			if bp.detectRuneCooldowns then AutoRuneBars(bp, bg) end
		else
			for _, bar in pairs(bp.bars) do -- iterate over each bar in the bar group
				if bar.enableBar and (IsOff(bar.hideBar) or (bar.hideBar ~= MOD:CheckCondition(bar.hideCondition))) then
					local t = bar.barType
					if (t == "Buff")  or (t == "Debuff") then
						local aname, cb, saveLabel, count = bar.action, bar.castBy, bar.barLabel, 0
						local auraList = MOD:CheckAura(bar.monitor, aname, t == "Buff")
						for _, aura in pairs(auraList) do
							local isMine = (aura[6] == "player") -- enforce optional isMine restrictions 
							if (((cb == "Player") and isMine) or ((cb == "Other") and not isMine) or (cb == "Anyone")) then
								count = count + 1
								if count > 1 then bar.barLabel = bar.barLabel .. " " end -- add space at end to make unique
								if CheckTimeAndDuration(bp, aura[2], aura[5]) then
									UpdateBar(bp, bg, bar, aura[8], aura[2], aura[5], aura[3], aura[4], aura[11], aura[12], bar.monitor)
								end
							end
						end
						bar.barLabel = saveLabel -- restore in case of multiple bar copies
					elseif t == "Cooldown" then
						local aname = bar.action
						local cd = MOD:CheckCooldown(aname) -- look up in the active cooldowns table
						if cd and (cd[1] ~= nil) then
							if CheckTimeAndDuration(bp, cd[1], cd[4]) then
								UpdateBar(bp, bg, bar, cd[2], cd[1], cd[4], nil, nil, cd[5], cd[6], nil)
							end
						elseif bar.enableReady then -- see if need to create a ready bar for spell off cooldown
							if bar.readyNotUsable or IsUsableSpell(aname) then -- check if testing for spell actually usable
								if CheckTimeAndDuration(bp, 0, 0) then
									local ttype, link = "spell link", MOD:GetHyperlink(aname)
									if not link then ttype = "text"; link = aname end
									UpdateBar(bp, bg, bar, MOD:GetIcon(aname), 0, 0, nil, nil, ttype, link, nil)
								end
							end
						end
					elseif t == "Notification" then
						if MOD:CheckCondition(bar.action) then
							local icon = MOD:GetIconForBar(bar)
							UpdateBar(bp, bg, bar, icon, 0, 0, nil, nil, "notification", bar.barLabel, bar.action)
						end
					end
				end
			end
		end
	end
end

-- Update bars in all bar groups, checking for bar group visibility and removing expired bars
function MOD:UpdateBars()
	if hidden then -- if was hidden then need to re-initialize all bar groups
		hidden = false
		MOD:UpdateAllBarGroups()
	end
	
	for _, bp in pairs(MOD.db.profile.BarGroups) do -- iterate through the all bar groups
		if IsOn(bp) then
			local bg = Nest_GetBarGroup(bp.name) -- match the profile bar group to the LibBars bar group
			if bg then	
				if bp.enabled then -- check all the conditions under which the bar group might be hidden are not true
					Nest_SetAllAttributes(bg, "updated", false) -- first, mark all the bars in the group as not updated...
					if not bp.merged then
						UpdateBarGroupBars(bp, bg) -- then update all the bars for the bar group into the display bar group					
						for _, mbp in pairs(MOD.db.profile.BarGroups) do -- then look for bar groups merging into this bar group
							if IsOn(mbp) and mbp.enabled and mbp.merged and (mbp.mergeInto == bp.name) then
								UpdateBarGroupBars(mbp, bg) -- update all bars for merged bar group into same display bar group
							end
						end
					end				
					UpdateTestBars(bg) -- update any unexpired test bars in this bar group
					Nest_DeleteBarsWithAttribute(bg, "updated", false) -- then, remove any bars in the group that weren't updated
				else -- if not then hide any bars that might be around
					Nest_DeleteAllBars(bg)
				end
			end
		end
	end
	UpdatePositions()
end
