-- Raven is an addon to monitor auras and cooldowns, providing timer bars, action bar highlights, and helpful notifications.

-- Conditions.lua contains routines to test auras and status info to generate notifications.

-- Exported functions:
-- Raven:CheckCondition(name) evaluates a condition and returns its current value
-- Raven:GetConditionSpell(name) returns the spell associated with a condition
-- Raven:GetConditionText(name) returns a string containing a detailed description of the condition

local MOD = Raven
MOD.status = {} -- global status info cached on every update
local fishSpell = GetSpellInfo(7620)
local unitList = { player = "Player", pet = "Pet", target = "Target", focus = "Focus" }

-- Saved variables don't handle being set to nil properly so need to use alternate value to indicate an option has been turned off
local Off = 0 -- value used to designate an option is turned off
local function IsOff(value) return value == nil or value == Off end -- return true if option is turned off
local function IsOn(value) return value ~= nil and value ~= Off end -- return true if option is turned on

-- Initialization settings for each test type
MOD.conditionTests = {
	["Player Status"] = { enable = false, inCombat = nil, isResting = nil, hasPet = nil, isStealthed = nil, isMounted = nil, isFishing = nil,
		hasMainHand = nil, levelMainHand = 1, hasOffHand = nil, levelOffHand = 1, checkTalent = nil, talent = nil,
		checkLevel = nil, level = 80, checkStance = nil, stance = nil, checkHealth = nil, minHealth = 100, checkPower = nil, minPower = 100,
		checkComboPoints = nil, minComboPoints = 5, checkRune = nil, needBlood = nil, needFrost = nil, needUnholy = nil, },
	["Target Status"] = { enable = false, isEnemy = nil, isFriend = nil, inRange = nil, checkHealth = nil, minHealth = 100, checkPower = nil, minPower = 100 },
	["Focus Status"] = { enable = false, isEnemy = nil, isFriend = nil, inRange = nil, checkHealth = nil, minHealth = 100, checkPower = nil, minPower = 100 },
	["All Buffs"] = { enable = false, unit = "player", auras = {}, isMine = true, toggle = false },
	["Any Buffs"] = { enable = false, unit = "player", auras = {}, isMine = true, toggle = false },
	["Buff Time Left"] = { enable = false, unit = "player", aura = "", timeLeft = 10, isMine = true, toggle = false },
	["Buff Count"] = { enable = false, unit = "player", aura = "", count = 1, isMine = true, toggle = false },
	["Buff Type"] = { enable = false, hasBuff = nil, toggle = false },
	["All Debuffs"] = { enable = false, unit = "target", auras = {}, isMine = true, toggle = false },
	["Any Debuffs"] = { enable = false, unit = "target", auras = {}, isMine = true, toggle = false },
	["Debuff Time Left"] = { enable = false, unit = "target", aura = "", timeLeft = 10, isMine = true, toggle = false },
	["Debuff Count"] = { enable = false, unit = "player", aura = "", count = 1, isMine = true, toggle = false },
	["Debuff Type"] = { enable = false, hasDebuff = nil, toggle = false },
	["All Cooldowns"] = { enable = false, spells = {}, timeLeft = 10 },
	["Spell Ready"] = { enable = false, gcd = false, spell = "", inRange = false, notUsable = false },
	["Spell Casting"] = { enable = false, spell = "", unit = "player" },
	["Item Ready"] = { enable = false, item = "", toggle = false },
}

MOD.testOrder = { "Player Status", "Target Status", "Focus Status", "All Buffs", "Any Buffs", "Buff Time Left", "Buff Count", "Buff Type",
	"All Debuffs", "Any Debuffs", "Debuff Time Left", "Debuff Count", "Debuff Type", "All Cooldowns", "Spell Ready", "Spell Casting",
	"Item Ready" }

-- Initialize data used by this module
function MOD:InitializeConditions()
	local ct = MOD.db.profile.Conditions[MOD.myClass]
	if ct then
		for _, c in pairs(ct) do 
			if IsOn(c) and c.associatedSpell then MOD:GetIcon(c.associatedSpell) end -- forces lookups to happen at initialization
		end
	end
end

-- Finalize conditions by stripping out unnecessary values
function MOD:FinalizeConditions()
	for _, ct in pairs(MOD.db.profile.Conditions) do
		for _, c in pairs(ct) do 
			if IsOn(c) then c.result = nil; c.testResult = nil end -- these values are left over from evaluations
		end
	end
end

-- Check if a spell is ready to be cast by the player, if rangeCheck then make sure in range of unit too
local function CheckSpellReady(spell, gcd, unit, rangeCheck, usable)
	if not spell or (spell == "") then return true end
	if usable and not IsUsableSpell(spell) then return false end -- checks player has learned the spell, has mana and/or reagents, and reactive conditions are met
	local dcheck = 1.5 -- global cooldown threshold
	if gcd == true then dcheck = 0 end -- ignore global cooldown
	local start, duration, enabled = GetSpellCooldown(spell) -- checks if spell is on cooldown (includes cases where spell is on global cooldown, distinguished by duration of 1.5 seconds or less)
	if (start == nil) or ((enabled == 0) or (start > 0 and duration > dcheck)) then return false end -- start will be nil for invalid spell names
	if IsOn(rangeCheck) and IsOn(unit) and ((IsSpellInRange(spell, unit) == 1) ~= rangeCheck) then return false end
	return true
end

-- Check if a specified spell is currently being cast or channeled by the unit 
local function CheckSpellCast(spell, unit)
	if IsOff(unit) or not spell or (spell == "") then return true end
	local sp = UnitCastingInfo(unit)
	if (sp ~= nil) and (spell == sp) then return true end
	sp = UnitChannelInfo(unit)
	if (sp ~= nil) and (spell == sp) then return true end
	return false
end

-- Check if an item is ready to be used by the player,
local function CheckItemReady(item, toggle)
	if not item or (item == "") then return true end
	local start, duration, enabled = GetItemCooldown(item) -- start will be nil for invalid item names
	if (start == nil) or ((enabled == 0) or (start > 0 and duration > 0)) then return toggle end
	return not toggle
end

-- Check for either at least one active aura (hasAll is false) or all active auras (hasAll is true)
-- If isMine true then cast by the player, if false then cast by other, if nil cast by anyone
local function CheckAuras(unit, auras, hasAll, isMine, buff, toggle)
	if IsOff(unit) or not auras then return true end -- no test
	if (unit == "target") and MOD.status.noTarget then return not toggle end
	if (unit == "focus") and MOD.status.noFocus then return not toggle end
	local found = false
	for _, aname in pairs(auras) do -- look for each aura and check if buff/debuff and cast by player
		local auraList = MOD:CheckAura(unit, aname, buff)
		if #auraList > 0 then
			for _, aura in pairs(auraList) do
				if IsOff(isMine) or (isMine == (aura[6] == "player")) then -- check cast by setting
					found = true -- found at least one of the auras
				else
					if hasAll then return toggle end -- not all are found
				end
			end
		else
			if hasAll then return toggle end -- not all are found
		end
	end
	if toggle then return not found end -- flip result if necessary
	return found
end

-- Check an aura on the unit to see if active and either expiration time is less than or equal or count is greater
-- If isMine true then cast by the player, if false then cast by other, if nil cast by anyone
local function CheckAuraValue(unit, aname, timeLeft, count, isMine, buff, toggle)
	if IsOff(unit) or not aname or (aname == "") or (IsOff(timeLeft) and IsOff(count)) then return true end -- no test
	if (unit == "target") and MOD.status.noTarget then return not toggle end
	if (unit == "focus") and MOD.status.noFocus then return not toggle end
	local auraList = MOD:CheckAura(unit, aname, buff)
	for _, aura in pairs(auraList) do
		if IsOff(isMine) or (isMine == (aura[6] == "player")) then -- check if found one with right cast by setting
			if timeLeft and (aura[2] >= timeLeft) then return not toggle end -- check expiration time is greater or equal
			if count and (aura[3] >= count) then return not toggle end -- check count is greater or equal
		end
	end
	return toggle
end

-- Return true only if all spells in the list are on cooldown and have at least the specified amount of time left.
-- If toggle is true then check if they all have less than the specified time left instead.
-- If a spell is not ready for some other reason (e.g., too much target HP ala Kill Shot) then consider it on long cooldown.
local function CheckAllCooldowns(spells, timeLeft, toggle)
	for _, spell in pairs(spells) do -- look for each spell and check if on cooldown
		local cdt = 0
		if not IsUsableSpell(spell) then cdt = 3600 end -- checks player has learned the spell, has mana and/or reagents, and reactive conditions are met
		local cd = MOD:CheckCooldown(spell) -- look up in the active cooldowns table
		if cd and (cd[1] ~= nil) then cdt = cd[1] end
		if toggle then
			if cdt >= timeLeft then return false end
		else
			if cdt < timeLeft then return false end
		end
	end
	return true
end

-- Return the current stance or "none" if not in one
local function GetStance()
	local form = "none"
	local index = GetShapeshiftForm(nil)
	if index > 0 then
		 _, form = GetShapeshiftFormInfo(index)
	end
	return form
end

-- Return true only if the specified talent has points spent in it
local function CheckTalent(talent)
	local t = MOD.talents[talent]
	if t then
		local _, _, _, _, pts = GetTalentInfo(t.tab, t.index, false, false, nil)
		if pts and (pts > 0) then return true end
	end
	return false
end

-- Check if mounted, complicated by having to test for druid flight form
local function CheckMounted()
	local flying = false
	if MOD.myClass == "DRUID" then
		local form = GetShapeshiftForm(true)
		if form ~= 0 then
			local _, fname = GetShapeshiftFormInfo(form)
			flying = (fname == "Flight Form") or (fname == "Swift Flight Form")
		end
	end
	return (IsMounted() ~= nil) or flying
end

-- Check if weapon equipped with at least the minimum specified item level
local function CheckWeapon(slot, level)
	local islot = GetInventorySlotInfo(slot)
	local id = GetInventoryItemLink("player", islot)
	if id then
		local iname, _, _, ilevel = GetItemInfo(id)
		if iname and (ilevel >= level) then return true end
	end
	return false
end

-- Make sure the table contains at least one entry
local function HasTable(t)
	if IsOn(t) and (type(t) == "table") and (next(t) ~= nil) then return true end
	return false
end

-- Check the value of one test (see template at end of file)
local function CheckTest(ttype, t)
	local toggle = (t.toggle == true)
	local stat = MOD.status
	if ttype == "Player Status" then
		if IsOn(t.inCombat) and (t.inCombat ~= stat.inCombat) then return false end
		if IsOn(t.isResting) and (t.isResting ~= stat.isResting) then return false end
		if IsOn(t.isStealthed) and (t.isStealthed ~= stat.isStealthed) then return false end
		if IsOn(t.isFishing) and (t.isFishing ~= stat.isFishing) then return false end
		if IsOn(t.isMounted) and (t.isMounted ~= stat.isMounted) then return false end
		if IsOn(t.checkLevel) and IsOn(t.level) and (t.checkLevel ~= (stat.level >= t.level)) then return false end
		if IsOn(t.checkHealth) and IsOn(t.minHealth) and (t.checkHealth ~= (stat.health >= t.minHealth)) then return false end
		if IsOn(t.checkPower) and IsOn(t.minPower) and (t.checkPower ~= (stat.power >= t.minPower)) then return false end
		if IsOn(t.checkComboPoints) and IsOn(t.minComboPoints) and (t.checkComboPoints ~= (stat.combo >= t.minComboPoints)) then return false end
		if IsOn(t.hasPet) and (t.hasPet ~= (HasPetUI() ~= nil)) then return false end
		if IsOn(t.checkStance) and IsOn(t.stance) and (t.stance ~= stat.stance) then return false end
		if IsOn(t.checkRunes) and not IsRuneSpellReady(t.needBlood, t.needFrost, t.needUnholy) then return false end
		if IsOn(t.checkTalent) and IsOn(t.talent) and not CheckTalent(t.talent) then return false end
		if IsOn(t.hasMainHand) and not CheckWeapon("MainHandSlot", t.levelMainHand) then return false end
		if IsOn(t.hasOffHand) and not CheckWeapon("SecondaryHandSlot", t.levelOffHand) then return false end
	elseif ttype == "Target Status" then -- target must exist for these tests to be true
		if IsOn(t.isEnemy) and (stat.noTarget or (t.isEnemy ~= stat.targetEnemy)) then return false end
		if IsOn(t.isFriend) and (stat.noTarget or (t.isFriend ~= stat.targetFriend)) then return false end
		if IsOn(t.checkHealth) and IsOn(t.minHealth) and (stat.noTarget or (t.checkHealth ~= (stat.targetHealth >= t.minHealth))) then return false end
		if IsOn(t.checkPower) and IsOn(t.minPower) and (stat.noTarget or (t.checkPower ~= (stat.targetPower >= t.minPower))) then return false end
		if IsOn(t.inRange) and (stat.noTarget or (t.inRange ~= stat.targetInRange)) then return false end
	elseif ttype == "Focus Status" then -- focus must exist for these tests to be true
		if IsOn(t.isEnemy) and (stat.noFocus or (t.isEnemy ~= stat.focusEnemy)) then return false end
		if IsOn(t.isFriend) and (stat.noFocus or (t.isFriend ~= stat.focusFriend)) then return false end
		if IsOn(t.checkHealth) and IsOn(t.minHealth) and (stat.noFocus or (t.checkHealth ~= (focusHealth >= t.minHealth))) then return false end
		if IsOn(t.checkPower) and IsOn(t.minPower) and (stat.noFocus or (t.checkPower ~= (focusPower >= t.minPower))) then return false end
		if IsOn(t.inRange) and (stat.noFocus or (t.inRange ~= stat.focusInRange)) then return false end
	elseif ttype == "All Buffs" then
		if HasTable(t.auras) and not CheckAuras(t.unit, t.auras, true, t.isMine, true, toggle) then return false end
	elseif ttype == "Any Buffs" then
		if HasTable(t.auras) and not CheckAuras(t.unit, t.auras, false, t.isMine, true, toggle) then return false end
	elseif ttype == "Buff Time Left" then
		if not CheckAuraValue(t.unit, t.aura, t.timeLeft, nil, t.isMine, true, toggle) then return false end
	elseif ttype == "Buff Count" then
		if not CheckAuraValue(t.unit, t.aura, nil, t.count, t.isMine, true, toggle) then return false end
	elseif ttype == "Buff Type" then
		if IsOn(t.hasBuff) and (MOD:UnitHasBuff("player", t.hasBuff) == toggle) then return false end
	elseif ttype == "All Debuffs" then
		if HasTable(t.auras) and not CheckAuras(t.unit, t.auras, true, t.isMine, false, toggle) then return false end
	elseif ttype == "Any Debuffs" then
		if HasTable(t.auras) and not CheckAuras(t.unit, t.auras, false, t.isMine, false, toggle) then return false end
	elseif ttype == "Debuff Time Left" then
		if not CheckAuraValue(t.unit, t.aura, t.timeLeft, nil, t.isMine, false, toggle) then return false end
	elseif ttype == "Debuff Count" then
		if not CheckAuraValue(t.unit, t.aura, nil, t.count, t.isMine, false, toggle) then return false end
	elseif ttype == "Debuff Type" then
		if IsOn(t.hasDebuff) and (MOD:UnitHasDebuff("player", t.hasDebuff) == toggle) then return false end
	elseif ttype == "All Cooldowns" then
		if HasTable(t.spells) and not CheckAllCooldowns(t.spells, t.timeLeft, toggle) then return false end
	elseif ttype == "Spell Ready" then
		if not CheckSpellReady(t.spell, t.gcd, "target", t.inRange, not t.notUsable) then return false end
	elseif ttype == "Spell Casting" then
		if not CheckSpellCast(t.spell, t.unit) then return false end
	elseif ttype == "Item Ready" then
		if not CheckItemReady(t.item, toggle) then return false end
	else
		return false
	end
	return true
end

-- Check the value of one condition (see template at end of file), all individual tests must evaluate to true (i.e., logical and)
local function CheckCondition(c)
	if not c.enabled then return false end -- if enabled default value is true, if not enabled always return false
	if c.tests then
		for ttype, test in pairs(c.tests) do
			if test.enable and not CheckTest(ttype, test) then return false end
		end
	end
	return true
end

-- Process table of conditions in current profile, checking status and auras.
function MOD:UpdateConditions()
	-- update globally useful conditions
	local stat = MOD.status
	stat.inCombat = (UnitAffectingCombat("player") ~= nil)
	stat.isResting = (IsResting() ~= nil)
	stat.isFishing = (IsUsableSpell(fishSpell) ~= nil)
	stat.isMounted = CheckMounted()
	stat.isStealthed = (IsStealthed() ~= nil)
	stat.level = UnitLevel("player")
	stat.health = (100 * UnitHealth("player")/UnitHealthMax("player"))
	stat.power = (100 * UnitPower("player")/UnitPowerMax("player"))
	stat.combo = GetComboPoints("player")
	stat.stance = GetStance()
	stat.noTarget = (UnitExists("target") == nil)
	if not stat.noTarget then
		stat.targetEnemy = (UnitIsEnemy("player", "target") ~= nil)
		stat.targetFriend = (UnitIsFriend("player", "target") ~= nil)
		stat.targetHealth = (100 * UnitHealth("target")/UnitHealthMax("target"))
		stat.targetPower = (100 * UnitPower("target")/UnitPowerMax("target"))
		stat.targetInRange = (CheckInteractDistance("target", 1) ~= nil)
	end
	stat.noFocus = (UnitExists("focus") == nil)
	if not stat.noFocus then
		stat.focusEnemy = (UnitIsEnemy("player", "focus") ~= nil)
		stat.focusFriend = (UnitIsFriend("player", "focus") ~= nil)
		stat.focusHealth = (100 * UnitHealth("focus")/UnitHealthMax("focus"))
		stat.focusPower = (100 * UnitPower("focus")/UnitPowerMax("focus"))
		stat.focusInRange = (CheckInteractDistance("focus", 1) ~= nil)
	end
	
	-- only check conditions for the player's class
	local ct = MOD.db.profile.Conditions[MOD.myClass]
	if ct then
		-- set all conditions to false first (quicker to clear it than to check IsOn)
		for _, c in pairs(ct) do if IsOn(c) then c.testResult = false; c.result = false end end
		-- don't check conditions if dead or in vehicle or on a taxi
		if UnitIsDeadOrGhost("player") ~= nil then return end
		if UnitUsingVehicle("player") ~= nil then return end
		if UnitOnTaxi("player") ~= nil then return end		
		-- run the tests in each condition to get intermediate testResult
		for _, c in pairs(ct) do
			if IsOn(c) and c.name then c.testResult = CheckCondition(c) end
		end
		-- then check dependencies and overrides to get final result
		for _, c in pairs(ct) do 
			if IsOn(c) and c.name then
				c.result = c.testResult -- start with intermediate result
				if c.result and c.dependencies then -- if starting true then check dependencies
					for dname, result in pairs(c.dependencies) do -- check each dependency in the table
						local dep = ct[dname] -- dependency must not be missing from the conditions table
						if IsOff(dep) or (dep.testResult ~= result) then c.result = false end
					end
				end
				if c.toggleResult then c.result = not c.result end
			end
		end
	end
end

-- Return the current value of a condition, nil if not found
function MOD:CheckCondition(name)
	local ct = MOD.db.profile.Conditions[MOD.myClass]
	if ct then
		local c = ct[name]
		if IsOn(c) and c.name then return c.result end
	end
	return nil
end

-- Return the associated spell for a condition, nil if not found
function MOD:GetConditionSpell(name)
	local ct = MOD.db.profile.Conditions[MOD.myClass]
	if ct then
		local c = ct[name]
		if IsOn(c) and c.associatedSpell then return c.associatedSpell end
	end
	return nil
end

-- Return a text description of the current condition
function MOD:GetConditionText(name)
	local description, testString, depString, resultString = "", "", "", ""
	description = string.format("|cFF7adbf2Condition Name:|r %s", name)
	local etext = ""
	local ct, c = MOD.db.profile.Conditions[MOD.myClass], nil
	if ct then c = ct[name] end
	if not c then return description .. " (not found)" end
	if not c.enabled then etext = " |cFFef2929[DISABLED]|r" end
	description = description .. etext
	if c.tests then
		for _, tt in pairs(MOD.testOrder) do -- loop through tests in right order
			local t = c.tests[tt]
			if t and t.enable then -- check if a test is defined and enabled for this test type
				local a, d = "", "" -- a is string for current test conditions, d gets comma after first test clause added
				local at = string.format("\n|cFF7adbf2%s:|r", tt) -- test caption
				if tt == "Player Status" then
					if IsOn(t.inCombat) then if t.inCombat then a = a .. d .. " In Combat" else a = a .. d .. " Out Of Combat" end; d = "," end
					if IsOn(t.isResting) then if t.isResting then a = a .. d .. " Is Resting" else a = a .. d .. " Not Resting" end; d = "," end
					if IsOn(t.isStealthed) then if t.isStealthed then a = a .. d .. " Is Stealthed" else a = a .. d .. " Not Stealthed" end; d = "," end
					if IsOn(t.isFishing) then if t.isFishing then a = a .. d .. " Is Fishing" else a = a .. d .. " Not Fishing" end; d = "," end
					if IsOn(t.isMounted) then if t.isMounted then a = a .. d .. " Is Mounted" else a = a .. d .. " Not Mounted" end; d = "," end
					if IsOn(t.hasPet) then if t.hasPet then a = a .. d .. " Has Pet" else a = a .. d .. " No Pet" end; d = "," end
					if IsOn(t.checkLevel) and t.level then local x = "<"; if t.checkLevel then x = ">=" end;
						a = a .. d .. string.format(" Level %s %d", x, t.level); d = "," end
					if IsOn(t.checkHealth) and t.minHealth then local x = "<"; if t.checkHealth then x = ">=" end;
						a = a .. d .. string.format(" Health %s %d%%", x, t.minHealth); d = "," end
					if IsOn(t.checkPower) and t.minPower then local x = "<"; if t.checkPower then x = ">=" end;
						a = a .. d .. string.format(" Power %s %d", x, t.minPower); d = "," end
					if IsOn(t.checkComboPoints) and t.minComboPoints then local x = "<"; if t.checkComboPoints then x = ">=" end;
						a = a .. d .. string.format(" Combo Pts %s %d", x, t.minComboPoints); d = "," end
					if IsOn(t.checkStance) and t.stance then a = a .. d .. string.format(" Stance = \"%s\"", t.stance); d = "," end
					if IsOn(t.checkTalent) and t.talent then a = a .. d .. string.format(" Talent = \"%s\"", t.talent); d = "," end
					if IsOn(t.checkRunes) and (t.needBlood or t.needFrost or t.needUnholy) then
						a = a .. d .. " "
						local p, pc = "", 0
						if t.needBlood then a = a .. string.format("Blood"); p = "+"; pc = pc + 1 end
						if t.needFrost then a = a .. p .. string.format("Frost"); p = "+"; pc = pc + 1 end
						if t.needUnholy then a = a .. p .. string.format("Unholy"); pc = pc + 1 end
						a = a .. " Rune"
						if pc > 1 then a = a .. "s" end
						d = ","
					end
					if IsOn(t.hasMainHand) and IsOn(t.levelMainHand) then a = a .. d .. string.format(" Weapon Mainhand >= %d", t.levelMainHand); d = "," end
					if IsOn(t.hasOffHand) and IsOn(t.levelOffHand) then a = a .. d .. string.format(" Weapon Offhand >= %d", t.levelOffHand); d = "," end
				elseif tt == "Target Status" or tt == "Focus Status" then
					if IsOn(t.isEnemy) then if t.isEnemy then a = a .. d .. " Is Enemy" else a = a .. d .. " Not Enemy" end; d = "," end
					if IsOn(t.isFriend) then if t.isFriend then a = a .. d .. " Is Friendly" else a = a .. d .. " Not Friendly" end; d = "," end
					if IsOn(t.inRange) then if t.inRange then a = a .. d .. " In Range" else a = a .. d .. " Out Of Range" end; d = "," end
					if IsOn(t.checkHealth) and t.minHealth then local x = "<"; if t.checkHealth then x = ">=" end;
						a = a .. d .. string.format(" Health %s %d%%", x, t.minHealth); d = "," end
					if IsOn(t.checkPower) and t.minPower then local x = "<"; if t.checkPower then x = ">=" end;
						a = a .. d .. string.format(" Power %s %d", x, t.minPower); d = "," end
				elseif tt == "All Buffs" or tt == "All Debuffs" then
					if IsOn(t.unit) and t.auras then
						for _, v in pairs(t.auras) do a = a .. d .. string.format(" \"%s\"", v); d = "," end
						if d == "," then -- the list of auras contains at least one entry
							if t.toggle == true then a = a .. ", Not All Active" else a = a .. ", All Active" end
							a = a .. ", On " .. unitList[t.unit]
							if t.isMine == true then a = a .. ", Cast By Player" else if t.isMine == false then
								a = a .. ", Cast By Other" else a = a .. ", Cast By Anyone" end end
						end
					end
				elseif tt == "Any Buffs" or tt == "Any Debuffs" then
					if IsOn(t.unit) and t.auras then
						for _, v in pairs(t.auras) do a = a .. d .. string.format(" \"%s\"", v); d = "," end
						if d == "," then -- the list of auras contains at least one entry
							if t.toggle == true then a = a .. ", None Active" else a = a .. ", Any Active" end
							a = a .. ", On " .. unitList[t.unit]
							if t.isMine == true then a = a .. ", Cast By Player" else if t.isMine == false then
								a = a .. ", Cast By Other" else a = a .. ", Cast By Anyone" end end
						end
					end
				elseif tt == "Buff Time Left" or tt == "Debuff Time Left" then
					if IsOn(t.unit) and IsOn(t.timeLeft) and t.aura and t.aura ~= "" then
						a = a .. string.format(" \"%s\"", t.aura)
						a = a .. ", On " .. unitList[t.unit]
						local ds = string.format("%0d:%02d", math.floor(t.timeLeft / 60), t.timeLeft % 60)
						if t.toggle == true then a = a .. ", Less Than " .. ds else a = a .. ", " .. ds .. " Or More" end
						if t.isMine == true then a = a .. ", Cast By Player" else if t.isMine == false then
							a = a .. ", Cast By Other" else a = a .. ", Cast By Anyone" end end
					end
				elseif tt == "Buff Count" or tt == "Debuff Count" then
					if IsOn(t.unit) and t.aura and t.aura ~= "" then
						a = a .. string.format(" \"%s\"", t.aura)
						a = a .. ", On " .. unitList[t.unit]
						local d = t.count
						if IsOff(d) then d = 1 end -- default count is 1
						if t.toggle == true then a = a .. ", Less Than " .. d else a = a .. ", " .. d .. " Or More" end
						if t.isMine == true then a = a .. ", Cast By Player" else if t.isMine == false then
							a = a .. ", Cast By Other" else a = a .. ", Cast By Anyone" end end
					end
				elseif tt == "Buff Type" then
					if IsOn(t.hasBuff) then
						a = a .. string.format(" \"%s\"", t.hasBuff)
						if t.toggle == true then a = a .. " Missing" else a = a .. " Present" end
					end
				elseif tt == "Debuff Type" then
					if IsOn(t.hasDebuff) then
						a = a .. string.format(" \"%s\"", t.hasDebuff)
						if t.toggle == true then a = a .. " Missing" else a = a .. " Present" end
					end
				elseif tt == "All Cooldowns" then
					if IsOn(t.timeLeft) and t.spells then
						for _, v in pairs(t.spells) do a = a .. d .. string.format(" \"%s\"", v); d = "," end
						if d == "," then -- the list of spells contains at least one entry
							local ds = string.format("%0.1f", t.timeLeft)
							if t.toggle == true then a = a .. ", Less Than " .. ds .. " Seconds" else a = a .. ", " .. ds .. " Seconds Or More" end
						end
					end
				elseif tt == "Spell Ready" then
					if t.spell and t.spell ~= "" then
						a = a .. string.format(" \"%s\"", t.spell)
						if t.gcd == true then a = a .. ", Test Global Cooldown" else a = a .. ", Ignore Global Cooldown" end
						if t.notUsable == true then a = a .. ", Ignore Spell Usable" else a = a .. ", Test Spell Usable" end
						if IsOn(t.inRange) then
							if t.inRange == true then a = a .. ", Target In Range" else a = a .. ", Target Out Of Range" end
						end
					end
				elseif tt == "Spell Casting" then
					if t.spell and t.spell ~= "" and IsOn(t.unit) then
						a = a .. string.format(" \"%s\"", t.spell) .. ", Cast By " .. unitList[t.unit]							
					end
				elseif tt == "Item Ready" then
					if t.item and t.item ~= "" then
						a = a .. string.format(" \"%s\"", t.item) 
						if t.toggle == true then a = a .. ", Item Is Not Ready" else a = a .. ", Item Is Ready" end
					end
				end
				if a ~= "" then testString = testString .. at .. a end -- add caption and clauses for current test
			end
		end
	end
	if IsOn(c.dependencies) then -- have a dependencies list
		local at = string.format("\n|cFF7adbf2Dependencies:|r")
		local a, d = "", "" -- a is string for current test conditions, d gets comma after first test clause added
		local i, list = 0, {} -- build a sorted list of dependencies
		for dep in pairs(c.dependencies) do
			i = i + 1
			list[i] = dep
		end
		table.sort(list)
		for _, dep in pairs(list) do
			value = c.dependencies[dep]
			a = a .. d .. string.format(" [%s] = ", dep)
			if value then a = a .. "|cFF8ae234True|r" else a = a .. "|cFFef2929False|r" end
			d = ","
		end
		if a ~= "" then depString = depString .. at .. a end -- add caption and dependencies
	end
	if c.toggleResult then resultString = string.format("\n|cFF7adbf2Toggle Result|r") end
	return description .. testString .. depString .. resultString
end

--[[ Templates for conditions
conditionTemplate = {			-- indexed by the condition's name (must be unique)
	name = name,				-- condition name
	tests = {},					-- table of tests, indexed by testType
	testResult = false,			-- result from most recent update (without checking dependencies and overrides)
	dependencies = {},			-- table of condition, result pairs that must match for this one to be true
	result = false,				-- result of most recent update (after checking dependencies and overrides)
}

testTemplate = {
	testType = "type",			-- type of this test
	unit = "unit",				-- unit the test applies to
	toggle = true,				-- toggle result of the test (i.e., logical not)
	inCombat = nil,				-- if not nil, check if in or out of combat
	isResting = nil,			-- if not nil, check if resting
	isStealthed = nil,			-- if not nil, check if stealthed
	isMounted = nil,			-- if not nil, check if mounted
	hasPet = nil,				-- if not nil, check if player has pet
	inRange = nil,				-- if not nil, check if unit is in range
	isEnemy = nil,				-- if not nil, check if unit exists and is enemy
	isFriend = nil,				-- if not nil, check if unit exists and is friend
	hasMainHand = nil,			-- if not nil, check if unit has mainhand weapon equipped
	levelMainHand = nil,		-- if not nil, minimum mainhand weapon level
	hasOffHand = nil,			-- if not nil, check if unit has offhand weapon equipped
	levelOffHand = nil,			-- if not nil, miniumum offhand weapon level
	checkHealth = nil,			-- if not nil, check unit health versus targetHealth
	minHealth = nil,			-- if not nil, % minimum health required
	checkPower = nil,			-- if not nil, check unit power versus minPower
	minPower = nil,				-- if not nil, minimum % energy/rage/runic power required
	checkComboPoints = nil,		-- if not nil, check combo points versus minComboPoints
	minComboPoints = nil,		-- if not nil, minimum combo points required
	checkRunes = nil,			-- if not nil, check runes
	needBlood = nil				-- if true, must have a blood rune (or enough death runes)
	needFrost = nil				-- if true, must have a frost rune (or enough death runes)
	needUnholy = nil			-- if true, must have an unholy rune (or enough death runes)
	checkLevel = nil,			-- if not nil, check player's level
	level = nil,				-- if not nil, minimum level for player
	checkTalent = nil,			-- if not nil, check for a specific talent
	talent = nil,				-- if not nil, name of talent to check for
	item = nil,					-- if not nil, an item name
	spell = nil,				-- if not nil, a spell name
	gcd = nil,					-- if true then consider global cooldown for spell readiness
	notUsable = nil,			-- for spell ready, check if not usable due to mana, reagents, etc.
	anySpell = false,			-- spell check applies to any spell, not one specified spell
	channel = false,			-- check if spell is being channeled or not
	aura = nil,					-- aura for timeleft and count checks
	auras = {},					-- if not nil, a table of aura names
	isMine = true,				-- if not nil, check if aura cast by player
	timeLeft = nil,				-- if not nil, check time left on aura less than this value
	count = nil,				-- if not nil, check aura count is greater than or equal to this value
	hasDebuff = type,			-- if not nil, check if Poison/Curse/Magic/Disease
}
]]--
