-- Raven is an addon to monitor auras and cooldowns, providing timer bars, action bar highlights, and helpful notifications.

-- Main.lua contains initialization and update routines supporting Raven's core capability of tracking active auras and cooldowns.
-- It includes special cases for weapon buffs, stances, trinkets, and the hidden Eclipse cooldown.

-- Exported functions:
-- Raven:CheckAura(unit, name, isBuff) checks if an aura is active on a unit, returning detailed info if found
-- Raven:IterateAuras(unit, func, isBuff, p1, p2) calls func for each active aura, parameters include a table with detailed aura info 
-- Raven:CheckCooldown(name) checks if cooldown with the specified name is active, returning detailed info if found
-- Raven:IterateCooldowns(func, p1, p2) calls func for each active cooldown, parameters include a table with detailed cooldown info
-- Raven:UnitHasBuff(unit, type) returns true if the unit has an active buff of the specified type (e.g., "Mainhand")
-- Raven:UnitHasDebuff(unit, type) returns true if the unit has an active debuff of the specified type (e.g., "Poison")

Raven = LibStub("AceAddon-3.0"):NewAddon("Raven", "AceConsole-3.0", "AceEvent-3.0")
local MOD = Raven
local MOD_Options = "Raven_Options"
local optionsLoaded = false -- set when the load-on-demand options panel module has been loaded
local optionsFailed = false -- set if loading the option panel module failed
MOD.updateOptions = false -- set this to cause the options panel to update (checked every half second)

MOD.frame = nil
MOD.db = nil
MOD.ldb = nil
MOD.myClass = nil
MOD.myRace = nil
MOD.lockSpells = {} -- spells for testing lock out of each school of magic for current player
MOD.cdSpells = {} -- spells that can go on cooldown for the current player
MOD.classSpells = {} -- stores info about pre-defined spells for each class
MOD.classConditions = {} -- stores info about pre-defined conditions for each class
MOD.updatePlayer = false -- set when events indicate need to update player buffs
MOD.updatePet = false -- set when events indicate need to update pet buffs
MOD.updateTarget = false -- set when events indicate need to update target buffs
MOD.updateFocus = false -- set when events indicate need to update focus buffs
MOD.updateCooldowns = false -- set when actionbar or inventory slot cooldown starts or stops
MOD.updateActionbar = false -- set with actionbar change event happens
MOD.talents = {} -- table containing names and talent table location for each talent
MOD.talentList = {} -- table with list of talent names
MOD.runeSlots = {} -- cache information about each rune slot for DKs
MOD.runeTypes = {} -- cache types of runes
MOD.runeIcons = {} -- cache icons for Death Knight runes

local activeBuffs = { player = {}, pet = {}, target = {}, focus = {} } -- active buffs
local activeDebuffs = { player = {}, pet = {}, target = {}, focus = {} } -- active debuffs
local auraPool = {} -- pool of available aura descriptors
local activeCooldowns = {} -- spells/items that are currently on cooldown
local elapsedTime = 0 -- time in seconds since last update
local throttleTime = 0.1 -- minimum time in seconds between updates
local throttleCount = 0 -- secondary throttle incremented at each update, resets at 5
local bufftooltip = nil -- used to store tooltip for scanning weapon buffs
local bufftooltipline = {} -- caches lines in weapon tooltips
local mhLastBuff = nil -- saves name of most recent main hand weapon buff
local ohLastBuff = nil -- saves name of most recent off hand weapon buff
local iconEclipseSolar = nil -- used to determine type of Eclipse procs for Druids
local iconEclipseLunar = nil -- used to determine type of Eclipse procs for Druids
local iconGCD = nil -- icon for global cooldown
local iconPotion = nil -- icon for shared potions cooldown
local iconElixir = nil -- icon for shared elixirs cooldown
local expireEclipseSolar = nil -- set to expiration time when wrath Eclipse is active
local expireEclipseLunar = nil -- set to expiration time when starfire Eclipse is active
local lastTotems = {} -- cache last totems in each slot to see if changed
local lockouts = {} -- schools of magic that we are currently locked out of
local talentsInitialized = false -- set once talents have been initialized
local matchTable = {} -- passed from MOD:CheckAura with list of active auras

-- Initialization called when addon is loaded
function MOD:OnInitialize()
	_, MOD.myClass = UnitClass("player") -- cache the player's class
	_, MOD.myRace = UnitRace("player") -- cached the player's race
end

-- Functions called to trigger updates
local function TriggerCooldownUpdate() MOD.updateCooldowns = true end
local function TriggerActionsUpdate() MOD.updateActions = true end
local function TriggerPlayerUpdate() MOD.updatePlayer = true end

-- Event called when addon is enabled
function MOD:OnEnable()
	MOD:InitializeProfile() -- initialize the profile database
	MOD:InitializeLDB() -- initialize the data broker
	MOD:RegisterChatCommand("raven", function() MOD:OptionsPanel() end)
	
	-- Create a frame so that updates can be registered
	MOD.frame = CreateFrame('Frame')
	-- Set frame level high so visible above other addons
	MOD.frame:SetFrameLevel(MOD.frame:GetFrameLevel() + 8)
	-- Register the update method
	MOD.frame:SetScript('OnUpdate', function(self, elapsed) MOD:Update(elapsed) end)
	-- Register events called prior to starting play
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("UNIT_PET")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED")
	self:RegisterEvent("PLAYER_TALENT_UPDATE")
	self:RegisterEvent("SPELLS_CHANGED")
	self:RegisterEvent("UPDATE_SHAPESHIFT_FORM", TriggerPlayerUpdate)
	self:RegisterEvent("MINIMAP_UPDATE_TRACKING", TriggerPlayerUpdate)	
	self:RegisterEvent("UNIT_ENTERED_VEHICLE", TriggerPlayerUpdate)
	self:RegisterEvent("UNIT_EXITED_VEHICLE", TriggerPlayerUpdate)
	self:RegisterEvent("SPELL_UPDATE_COOLDOWN", TriggerCooldownUpdate)
	self:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN", TriggerCooldownUpdate)
	self:RegisterEvent("BAG_UPDATE_COOLDOWN", TriggerCooldownUpdate)
	self:RegisterEvent("PET_BAR_UPDATE_COOLDOWN", TriggerCooldownUpdate)
	self:RegisterEvent("ACTIONBAR_SLOT_CHANGED", TriggerActionsUpdate)
	self:RegisterEvent("ACTIONBAR_PAGE_CHANGED", TriggerActionsUpdate)
	self:RegisterEvent("RUNE_POWER_UPDATE", TriggerCooldownUpdate)
	self:RegisterEvent("RUNE_TYPE_UPDATE", TriggerCooldownUpdate)
	MOD:InitializeBars() -- initialize routine that manages the bar library
end

-- Event called when addon is disabled but this is probably never called
function MOD:OnDisable() end

-- Initialize rune information for Death Knights
local function InitializeRunes()
	local r, t = MOD.runeIcons, MOD.runeTypes
	r[1] = "Interface\\PlayerFrame\\UI-PlayerFrame-Deathknight-Blood" -- add DK rune info to cache
	r[2] = "Interface\\PlayerFrame\\UI-PlayerFrame-Deathknight-Unholy"
	r[3] = "Interface\\PlayerFrame\\UI-PlayerFrame-Deathknight-Frost"
	r[4] = "Interface\\PlayerFrame\\UI-PlayerFrame-Deathknight-Death"
	t[1] = "Blood Rune"; t[2] = "Unholy Rune"; t[3] = "Frost Rune"; t[4] = "Death Rune"
end

-- Cache icons for special purposes such as shared cooldowns
local function InitializeIcons()
	_, _, iconGCD = GetSpellInfo(28730) -- cached for global cooldown (using same icon as Arcane Torrent)
	_, _, iconEclipseSolar = GetSpellInfo(48517) -- cached for special case Eclipse check for Druids
	_, _, iconEclipseLunar = GetSpellInfo(48518) -- cached for special case Eclipse check for Druids
	iconPotion = GetItemIcon(31677) -- icon for shared potions cooldown
	iconElixir = GetItemIcon(28104) -- icon for shared elixirs cooldown
end

-- Initialize when play starts, deferred to allow system initialization to complete
function MOD:PLAYER_ENTERING_WORLD()
	MOD:InitializeHighlights() -- initialize routine that does action bar highlighting
	MOD:InitializeBuffTooltip() -- initialize tooltip used to monitor weapon buffs
	MOD:InitializeConditions() -- initialize routine that shows cooldown overlays and cooldown bars	
	InitializeRunes() -- death knight specific initialization
	InitializeIcons() -- cache special purpose icons

	MOD.updatePlayer, MOD.updatePet, MOD.updateTarget, MOD.updateFocus = true, true, true, true -- start tracking aura changes
	MOD.updateCooldowns, MOD.updateActions = true, true -- start tracking cooldowns and actionbar changes
	collectgarbage("collect") -- recover deleted preset data
end

-- Event called when an aura changes on a unit, returns the unit name
function MOD:UNIT_AURA(e, unit)
	if unit == "player" then
		MOD.updatePlayer = true
	elseif unit == "pet" then
		MOD.updatePet = true
	elseif unit == "target" then
		MOD.updateTarget = true
	elseif unit == "focus" then
		MOD.updateFocus = true
	end
end

-- Event called when the player changes talents
function MOD:PLAYER_TALENT_UPDATE()
	talentsInitialized = false
	MOD.updatePlayer = true
end

-- Event called when a pet changes
function MOD:UNIT_PET()
	MOD.updatePet = true
end

-- Event called when the target is changed
function MOD:PLAYER_TARGET_CHANGED()
	MOD.updateTarget = true
end

-- Event called when the focus is changed
function MOD:PLAYER_FOCUS_CHANGED()
	MOD.updateFocus = true
end

-- Event called when spells in spell book change
function MOD:SPELLS_CHANGED()
	MOD:SetIconDefaults()
	MOD:SetCooldownDefaults()
	MOD.updateCooldowns = true
end

-- Create cache of talent tab info plus special case for druid eclipse talent
function MOD:InitializeTalents()	
	local tabs = GetNumTalentTabs(false, false)
	if tabs == 0 then return end
	talentsInitialized = true
	MOD.talents = {}
	MOD.talentList = {}
	
	local select = 1
	for i = 1, tabs do
		local ts = GetNumTalents(i, false, false)
		for j = 1, ts do
			local name, rank, maxRank
			name, _, _, _, rank, maxRank = GetTalentInfo(i, j, false, false, nil) -- player's active talents
			MOD.talents[name] = { rank = rank, maxRank = maxRank, tab = i, index = j }
			MOD.talentList[select] = name
			select = select + 1
		end
	end
	table.sort(MOD.talentList)
	for i, t in pairs(MOD.talentList) do
		MOD.talents[t].select = i
	end
	
	if MOD.myClass == "DRUID" then
		MOD:SetIcon("Eclipse (Solar)", iconEclipseSolar)
		MOD:SetIcon("Eclipse (Lunar)", iconEclipseLunar)
		activeCooldowns["Eclipse (Solar)"] = { nil, nil, 0, 0, nil } -- make sure always have a cooldown table for Eclipse
		activeCooldowns["Eclipse (Lunar)"] = { nil, nil, 0, 0, nil } -- make sure always have a cooldown table for Eclipse	
		local t = MOD.talents["Eclipse"]
		if t then MOD.eclipseTab, MOD.eclipseIndex = t.tab, t.index end
	end
end

-- Check if the options panel is loaded, if not then get it loaded and ask it to toggle open/close status
function MOD:OptionsPanel()
    if not optionsLoaded then
        optionsLoaded = true
        local loaded, reason = LoadAddOn(MOD_Options)
        if not loaded then
            print("Failed to load " .. tostring(MOD_Options) .. ": " .. tostring(reason))
			optionsFailed = true
        end
	end
	if not optionsFailed then MOD:ToggleOptions() end
end

-- If the options panel is loaded then update it so it reflects any changes made thru anchors, etc.
function MOD:UpdateOptionsPanel()
    if optionsLoaded and not optionsFailed then MOD:UpdateOptions() end
	MOD.updateOptions = false
end

-- Tie into LibDataBroker
function MOD:InitializeLDB()
	local LDB = LibStub("LibDataBroker-1.1", true)
	MOD.ldb = LDB:NewDataObject("Raven", {
		type = "launcher",
		text = "Raven",
		icon = "Interface\\Icons\\Spell_Nature_RavenForm",
		OnClick = function(_, msg)
			if msg == "RightButton" then
				MOD:ToggleBarGroupLocks()
			elseif msg == "LeftButton" then
				MOD:OptionsPanel()
			end
		end,
		OnTooltipShow = function(tooltip)
			if not tooltip or not tooltip.AddLine then return end
			tooltip:AddLine("Raven")
			tooltip:AddLine("|cffffff00Left-click|r to open/close options menu")
			tooltip:AddLine("|cffffff00Right-click|r to toggle locking of bar groups")
		end,
	})
end

-- Show or hide the blizzard buff frames, called during update so synched with other changes
local function CheckBlizzFrames()
	local visible = BuffFrame:IsShown()
	if MOD.db.profile.hideBlizz then
		if visible then
			BuffFrame:UnregisterEvent("UNIT_AURA")
			BuffFrame:Hide()
			TemporaryEnchantFrame:Hide()
		end
	else
		if not visible then
			BuffFrame:RegisterEvent("UNIT_AURA")
			BuffFrame:Show()
			TemporaryEnchantFrame:Show()
			BuffFrame_Update()
		end
	end
end

-- Update routine called before each frame is displayed, throttled to minimize CPU usage
function MOD:Update(elapsed)
	CheckBlizzFrames() -- make sure blizzard frames are visible or not
	if MOD.db.profile.enabled then
		elapsedTime = elapsedTime + elapsed
		if elapsedTime >= throttleTime then
			throttleCount = throttleCount + 1; if throttleCount == 5 then throttleCount = 0 end
			if not talentsInitialized then MOD:InitializeTalents() end -- retry until talents initialized
			
			MOD:UpdateAuras() -- update table containing current auras on player, target and focus
			MOD:UpdateCooldowns() -- update table containing current cooldowns on action bar buttons and trinkets
			MOD:UpdateConditions() -- update table containing currently triggered conditions
			MOD:UpdateHighlights() -- update action bar buttons with highlights and cooldown text
			MOD:UpdateBars() -- update timer bars for auras and cooldowns
			Nest_Update() -- update the display using the Nest graphics package
			elapsedTime = 0
		else
			Nest_Refresh() -- refresh bars in the Nest graphics package (helps smooth animations)
		end
		if (throttleCount == 0) and MOD.updateOptions then MOD:UpdateOptionsPanel() end -- update option panel about every half second, if requested
	else
		MOD:HideHighlights()
		MOD:HideBars()
	end
end

-- Aura tables have this structure:
-- b[1] = isBuff, b[2] = timeLeft, b[3] = stackCount, b[4] = auraType, b[5] = duration, b[6] = caster, b[7] = isStealable
-- b[8] = icon, b[9] = rank, b[10] = expireTime, b[11] = tooltipType, b[12] = tooltipArgument, b[13] = name

-- Create and delete routines for managing aura tables, using a recycling pool to minimize garbage collection
local function AllocateAura() local b = next(auraPool); if b then auraPool[b] = nil else b = {} end return b end
local function ReleaseAura(b) table.wipe(b); auraPool[b] = true; return nil end

-- Calculate aura time left from expiration time and current time, this is always done before returning aura descriptors
-- If no duration or has expired then set to 0 (Blizzard may not yet have sent aura update event so could sit at 0 for a moment)
local function SetAuraTimeLeft(b) if b[5] > 0 then b[2] = b[10] - GetTime() if b[2] < 0 then b[2] = 0 end else b[2] = 0 end end
	
-- Add an active aura to the table for the specified unit
local function AddAura(unit, name, isBuff, count, btype, duration, caster, isStealable, icon, rank, expire, tt_type, tt_arg)
	local auraTable = isBuff and activeBuffs[unit] or activeDebuffs[unit]
	local b = AllocateAura() -- get an empty aura descriptor
	b[1], b[2], b[3], b[4], b[5], b[6], b[7], b[8], b[9], b[10], b[11], b[12], b[13] =
		isBuff, 0, count, btype, duration, caster, isStealable, icon, rank, expire, tt_type, tt_arg, name
	auraTable[#auraTable + 1] = b
	MOD:SetIcon(name, icon) --  cache icon in case this is first time the buff is detected
end

-- Check if aura(s) with given name are active on the unit (if isBuff is true only check buffs, otherwise only debuff)
-- Return a table with matching aura descriptors, potentially empty if none are found
-- The returned table is only valid until the next call to MOD:CheckAura since it is reused each time
function MOD:CheckAura(unit, name, isBuff)
	table.wipe(matchTable)
	local auraTable = isBuff and activeBuffs[unit] or activeDebuffs[unit]
	for _, b in pairs(auraTable) do
		if b[13] == name then SetAuraTimeLeft(b); matchTable[#matchTable + 1] = b end
	end
	return matchTable
end

-- For all active auras on a given unit (if isBuff is true only buffs, otherwise only debuff), call the function that
-- is passed in with the unit, aura name, aura descriptor table, isBuff, and two optional parameters passed in
function MOD:IterateAuras(unit, func, isBuff, p1, p2)
	local auraTable = isBuff and activeBuffs[unit] or activeDebuffs[unit]
	for _, b in pairs(auraTable) do
		SetAuraTimeLeft(b) -- update timeLeft from current time
		func(unit, b[13], b, isBuff, p1, p2)
	end
end

-- Release a particular player buff, including multiple copies, used for weapon buffs and tracking due to non-standard detection
local function ReleasePlayerBuff(name)
	local auraTable = activeBuffs.player
	for index, b in pairs(auraTable) do
		if b[13] == name then auraTable[index] = ReleaseAura(b) end
	end
end

-- Empty the aura tables for a unit by releasing all entries
local function ReleaseAuras(unit)
	local buffTable, debuffTable = activeBuffs[unit], activeDebuffs[unit]
	for index, b in pairs(buffTable) do buffTable[index] = ReleaseAura(b) end
	for index, b in pairs(debuffTable) do debuffTable[index] = ReleaseAura(b) end
end

-- Check all buffs on the unit to see if the specified buff type is currently active 
function MOD:UnitHasBuff(unit, btype)
	for _, b in pairs(activeBuffs[unit]) do
		if (b[4] == btype) and ((b[10] - GetTime()) > 0) then return true end
	end
	return false
end

-- Check all debuffs on the unit to see if the specified debuff type is currently active 
function MOD:UnitHasDebuff(unit, btype)
	for _, b in pairs(activeDebuffs[unit]) do
		if (b[4] == btype) and ((b[10] - GetTime()) > 0) then return true end
	end
	return false
end

-- Initialize tooltip to be used for determining weapon buffs
-- This code is based on the Pitbull implementation
function MOD:InitializeBuffTooltip()
	bufftooltip = CreateFrame("GameTooltip", nil, UIParent)
	bufftooltip:SetOwner(UIParent, "ANCHOR_NONE")
	local fs = bufftooltip:CreateFontString()
	fs:SetFontObject(_G.GameFontNormal)
	for i = 1, 30 do
		local ls = bufftooltip:CreateFontString()
		ls:SetFontObject(_G.GameFontNormal)
		bufftooltip:AddFontStrings(ls, fs)
		bufftooltipline[i] = ls
	end
end

-- No easy way to get this info, so scan item slot info for mainhand and offhand weapons using a tooltip
-- Weapon buffs are formatted in tooltips as name strings followed by remaining time in parentheses (e.g., "Wound Poison VI (15 min)")
-- This routine scans the tooltip for the first line that is in this format and extracts the weapon buff name without rank or time
local function GetWeaponBuff(weaponSlot)
	bufftooltip:ClearLines()
	bufftooltip:SetInventoryItem("player", weaponSlot)
	for i = 1, 30 do
		local text = bufftooltipline[i]:GetText()
		if text then
			local name = text:match("^(.+) %(%d+ [^$)]+%)$") -- extract name up to left paren in a line matching weapon buff format
			if name then
				name = (name:match("^(.*) %d+$")) or name -- remove any trailing numbers, as found in shaman weapon buffs
				-- local a, p = string.find(name, "Poison") -- special case for poisons: strip off ranks (e.g., "VI") from end of buff name
				-- if a then name = string.sub(name, 1, p) end -- commented out for now, since rogues actually might care about poison ranks
				return name
			end
		else
			break
		end
	end
	return nil
end

-- Get weapon buff duration, since this is not supplied by Blizzard look at current detected duration
-- and compare it to longest previous duration for the given weapon buff in order to find maximum ever detected
local function GetWeaponBuffDuration(buff, duration)
	local maxd = MOD.db.global.BuffDurations[buff]
	if not maxd or (duration > maxd) then
		MOD.db.global.BuffDurations[buff] = math.floor(duration + 0.5) -- round up
	else
		if maxd > duration then duration = maxd end
	end
	return duration
end

-- Add player weapon buffs for mainhand and offhand to the aura table 
-- The regen parameter is true if previous weapon buffs were released from the aura table and this routine must regenerate them
local function GetWeaponBuffs(regen)
	if not regen then -- if old buffs may still be valid then determine if it is time to regenerate them
		if throttleCount == 0 then return end -- limits weapon buff refreshes to every half second
		
		-- old weapons buffs are now out-of-date so release them before regenerating		
		if mhLastBuff then ReleasePlayerBuff(mhLastBuff) end
		if ohLastBuff then ReleasePlayerBuff(ohLastBuff) end 
	end
	mhLastBuff = nil; ohLastBuff = nil
	
	-- first check if there are weapon auras then, only if necessary, use tooltip to scan for the buff names
	local mh, mhms, mhc, oh, ohms, ohc = GetWeaponEnchantInfo()
	
	if mh then -- add the mainhand buff, if any, to the table
		local islot = GetInventorySlotInfo("MainHandSlot")
		local mhbuff = GetWeaponBuff(islot)
		if mhbuff then 
			local icon = GetInventoryItemTexture("player", islot)					
			local timeLeft = mhms / 1000
			local duration, expire = GetWeaponBuffDuration(mhbuff, timeLeft), GetTime() + timeLeft
			AddAura("player", mhbuff, true, mhc, "Mainhand", duration, "player", false, icon, nil, expire, "inventory", "MainHandSlot")
			mhLastBuff = mhbuff -- caches the name of the weapon buff so can clear it later
		end
	end
	
	if oh then -- add the offhand buff, if any, to the table
		local islot = GetInventorySlotInfo("SecondaryHandSlot")
		local ohbuff = GetWeaponBuff(islot)
		if ohbuff then
			local icon = GetInventoryItemTexture("player", islot)
			local timeLeft = ohms / 1000
			local duration, expire = GetWeaponBuffDuration(ohbuff, timeLeft), GetTime() + timeLeft
			AddAura("player", ohbuff, true, ohc, "Offhand", duration, "player", false, icon, nil, expire, "inventory", "SecondaryHandSlot")
			ohLastBuff = ohbuff -- caches the name of the weapon buff so can clear it later
		end
	end
end

-- Add tracking auras (updated for Cataclysm which allows multiple active tracking types)
local function GetTracking()
	local notTracking, notTrackingIcon, found = "Not Tracking", "Interface\\Minimap\\Tracking\\None", false
	for i = 1, GetNumTrackingTypes() do
		local tracking, trackingIcon, active = GetTrackingInfo(i)
		if active == 1 then
			found = true
			AddAura("player", tracking, true, 1, "Tracking", 0, "player", false, trackingIcon, nil, 0, "tracking", tracking)
		end
	end
	if not found then
		AddAura("player", notTracking, true, 1, "Tracking", 0, "player", false, notTrackingIcon, nil, 0, "tracking", notTracking)
	end
end

-- See if totems have changed since last update, can't count on events, and save for future checks
local function ChangedTotems()
	local changed = false
	if MOD.myClass == "SHAMAN" then
		for i = 1, 4 do 
			local _, name = GetTotemInfo(i)
			if lastTotems[i] ~= name then -- one of the totem slots has changed
				lastTotems[i] = name -- cache for next time checking the totems
				changed = true
			end
		end
	end
	return changed
end

-- Add buffs for the specified unit to the active buffs table
local function GetBuffs(unit)
	local name, rank, icon, count, btype, duration, expire, caster, isStealable, icon
	local i = 1
	repeat
		name, rank, icon, count, btype, duration, expire, caster, isStealable = UnitBuff(unit, i)
		if name then
			if caster == "vehicle" then caster = "player" end -- include buffs cast by vehicle controlled by the player
			AddAura(unit, name, true, count, btype, duration, caster, isStealable, icon, rank, expire, "buff", i)
		end
		i = i + 1
	until not name
	if unit ~= "player" then return end -- done for all but player, players also need to add vehicle buffs
	i = 1
	repeat
		name, rank, icon, count, btype, duration, expire, caster, isStealable = UnitBuff("vehicle", i)
		if name then
			if caster == "vehicle" then caster = "player" end -- include buffs cast by vehicle controlled by the player
			AddAura(unit, name, true, count, btype, duration, caster, isStealable, icon, rank, expire, "vehicle buff", i)
		end
		i = i + 1
	until not name	
end

-- Add debuffs for the specified unit to the active debuffs table
local function GetDebuffs(unit)
	local name, rank, icon, count, btype, duration, expire, caster, isStealable
	local i = 1
	repeat
		name, rank, icon, count, btype, duration, expire, caster, isStealable = UnitDebuff(unit, i)
		if name then
			if caster == "vehicle" then caster = "player" end -- include debuffs cast by vehicle controlled by the player
			AddAura(unit, name, false, count, btype, duration, caster, isStealable, icon, rank, expire, "debuff", i)
		end
		i = i + 1
	until not name
	if unit ~= "player" then return end -- done for all but player, players also need to add vehicle debuffs
	i = 1
	repeat
		name, rank, icon, count, btype, duration, expire, caster, isStealable = UnitDebuff("vehicle", i)
		if name then
			if caster == "vehicle" then caster = "player" end -- include debuffs cast by vehicle controlled by the player
			AddAura(unit, name, false, count, btype, duration, caster, isStealable, icon, rank, expire, "vehicle debuff", i)
		end
		i = i + 1
	until not name	
end

-- Update aura table with current player, target and focus auras and debuffs, include player weapon buffs
local function UpdateUnitAuras(unit) ReleaseAuras(unit); GetBuffs(unit); GetDebuffs(unit) end -- update unit auras
function MOD:UpdateAuras()
	if ChangedTotems() or UnitIsControlling("player") then MOD.updatePlayer = true end -- check conditions that can't count on an event
	local regen = MOD.updatePlayer -- tells whether to regenerate weapons and tracking even if not normally time to do so
	if MOD.updatePlayer then UpdateUnitAuras("player"); GetTracking(); MOD.updatePlayer = false end
	GetWeaponBuffs(regen) -- get buff info for mainhand and offhand weapons
	if MOD.updatePet then UpdateUnitAuras("pet"); MOD.updatePet = false end
	if MOD.updateTarget then UpdateUnitAuras("target"); MOD.updateTarget = false end
	if MOD.updateFocus then UpdateUnitAuras("focus"); MOD.updateFocus = false end
end

-- Cooldown tables have this structure (name of the cooldown is the index into the activeCooldowns table):
-- b[1] = timeLeft, b[2] = icon, b[3] = startTime, b[4] = duration, b[5] = tooltipType, b[6] = tooltipArgument, b[7] = unit

-- Check if valid cooldown table, if so then calculate time left from start time and duration and invalidate if cooldown has expired
-- Returns either the updated cooldown table or nil if not valid
local function ValidateCooldown(b) 
	if b[1] ~= nil then
		b[1] = b[3] + b[4] - GetTime() -- calculate timeLeft from start time and duration
		if b[1] > 0 then return b end -- check if the cooldown has expired 
		b[1] = nil -- this cooldown is no longer valid
	end
	return nil
end

-- Add a cooldown to the current list of active cooldowns, cached info includes icon, start time, duration, tt_type, tt_arg, unit
local function AddCooldown(name, icon, start, duration, tt_type, tt_arg, unit)
	local t = activeCooldowns -- shared for player and pet cooldowns
	if not t[name] then
		MOD:SetIcon(name, icon) --  cache icon first time see this inventory slot item
		t[name] = { 0, icon, start, duration, tt_type, tt_arg, unit }
	else
		local b = t[name]
		b[1], b[2], b[3], b[4], b[5], b[6], b[7] = 0, icon, start, duration, tt_type, tt_arg, unit
	end
end

-- Check if the named spell or item is on cooldown, return a cooldown table
function MOD:CheckCooldown(name)
	local b = activeCooldowns[name]
	if b then return ValidateCooldown(b) end
	return nil
end

-- Iterate over current cooldowns, calling the function with cooldown name, cooldown table, and optional parameters
function MOD:IterateCooldowns(func, p1, p2)
	for n, cd in pairs(activeCooldowns) do if ValidateCooldown(cd) then func(n, cd, p1, p2) end end
end

-- Empty the cooldowns table by reseting all entries
local function ReleaseCooldowns()
	for _, b in pairs(activeCooldowns) do b[1] = nil end -- set first field to nil to indicate is not active
end

-- Update the expiration time for cooldowns, releasing any that have lapsed
local function UpdateCooldownTimes()
	for _, b in pairs(activeCooldowns) do ValidateCooldown(b) end
end

-- Get cooldown info for an inventory slot
local function CheckInventoryCooldown(slot)
	local id = GetInventorySlotInfo(slot)
	if id then
		local start, duration, enable = GetInventoryItemCooldown("player", id)
		if start > 0 and enable == 1 and duration > 1.5 then
			local name, _, _, _, _, _, _, _, _, icon = GetItemInfo(GetInventoryItemLink("player", id))
			AddCooldown(name, icon, start, duration, "inventory", slot, "player")
		end
	end
end

-- Table of spells for each class to use for checking glabal cooldown, spells must be known at level 1
-- and must be subject to GCD but without their own cooldowns
local classGCD = {
	DEATHKNIGHT = "Blood Strike", DRUID = "Mark of the Wild", HUNTER = "Track Beasts", MAGE = "Arcane Intellect",
	PALADIN = "Devotion Aura", PRIEST = "Power Word: Fortitude", ROGUE = "Sinister Strike", SHAMAN = "Rockbiter Weapon",
	WARLOCK = "Demon Skin", WARRIOR = "Battle Shout"
}

-- See if on global cooldown, depends on current class
local function CheckGlobalCooldown()
	local start, duration, enable = GetSpellCooldown(classGCD[MOD.myClass])
	if start and start > 0 and enable == 1 and duration <= 1.5 then -- max GCD is 1.5 seconds
		AddCooldown("GCD", iconGCD, start, duration, "text", "Global Cooldown", "player")
	end
end

-- Check for Eclipse procs and create a 30 second cooldown bar
local function CheckEclipse()
	local solars = MOD:CheckAura("player", "Eclipse (Solar)", true)
	local bs = solars[1] -- can only be one entry in the table
	
	if expireEclipseSolar then -- already tracking solar Eclipse
		local c = activeCooldowns["Eclipse (Solar)"]
		if GetTime() >= expireEclipseSolar then
			expireEclipseSolar = nil
			c[1] = nil
		else
			c[1] = 0 -- restore in case was cleared by normal cooldown code
		end
	elseif bs and (bs[8] == iconEclipseSolar) then -- if not already tracking wrath Eclipse then check if one proc'd
		expireEclipseSolar = GetTime() + 30 -- hidden cooldown expiration time
		AddCooldown("Eclipse (Solar)", bs[8], GetTime(), 30, "spell", "Eclipse", "player")
	end
	
	local lunars = MOD:CheckAura("player", "Eclipse (Lunar)", true)
	local bl = lunars[1] -- can only be one entry in the table
	if expireEclipseLunar then -- already tracking lunar Eclipse
		local c = activeCooldowns["Eclipse (Lunar)"]
		if GetTime() >= expireEclipseLunar then
			expireEclipseLunar = nil
			c[1] = nil
		else
			c[1] = 0 -- restore in case was cleared by normal cooldown code
		end
	elseif bl and (bl[8] == iconEclipseLunar) then -- if not already tracking starfire Eclipse then check if one proc'd
		expireEclipseLunar = GetTime() + 30 -- hidden cooldown expiration time
		AddCooldown("Eclipse (Lunar)", bl[8], GetTime(), 30, "spell", "Eclipse", "player")
	end
end

-- Update info about the rune slots and add rune cooldowns
local function CheckRunes()
	local blood, frost, unholy, death = 0, 0, 0, 0
	for i = 1, 6 do
		local rune, rtype = MOD.runeSlots[i], GetRuneType(i)
		local start, duration, ready = GetRuneCooldown(i)
		if not rune then
			rune = { rtype = rtype, start = start, duration = duration, ready = ready }
			MOD.runeSlots[i] = rune
		else
			rune.rtype = rtype; rune.start = start; rune.duration = duration; rune.ready = ready
		end
	end
end

-- Return true only if the specified runes are available, with death runes serving as wildcards
function IsRuneSpellReady(blood, frost, unholy)
	local b, f, u, d = 0, 0, 0, 0
	for i = 1, 6 do
		if MOD.runeSlots[i].ready then
			local t = MOD.runeSlots[i].rtype
			if t == 1 then b = b + 1 elseif t == 2 then u = u + 1 elseif t == 3 then f = f + 1 elseif t == 4 then d = d + 1 end
		end
	end
	if blood and (b < 1) then if d < 1 then return false else d = d - 1 end end
	if frost and (f < 1) then if d < 1 then return false else d = d - 1 end end
	if unholy then if u < 1 and d < 1 then return false end end
	return true
end

-- Check if the spell is on cooldown because a rune is not available, return true only if on real cooldown
local function CheckRuneCooldown(name, duration)
	local runes = MOD.runeSpells[name]
	if runes then
		if IsRuneSpellReady(runes.blood, runes.frost, runes.unholy) then return true end -- runes are available so real cooldown
		if (duration >= 9) and (duration <= 10) then return false end -- right duration to be a rune cooldown
	end
	return true
end

-- Check for new and expiring cooldowns associated with all action bar slots plus trinkets (might want to add inventory slots someday)
function MOD:UpdateCooldowns()
	if not MOD.updateCooldowns then -- no new cooldown so just check for expiring ones
		UpdateCooldownTimes()
	elseif HasFullControl() or UnitIsControlling("player") then -- don't update if player doesn't have full control
		ReleaseCooldowns() -- mark all cooldowns as not active
		if MOD.myClass == "DEATHKNIGHT" then CheckRunes() end
		for school in pairs(lockouts) do lockouts[school] = 0 end -- clear any previous settings in lockout table
		for name, ls in pairs(MOD.lockSpells) do
			if not lockouts[ls.school] then lockouts[ls.school] = 0 end -- initialize when school seen for first time
			if ls.index and (lockouts[ls.school] == 0) then
				local start, duration = GetSpellCooldown(ls.index, "spell")
				if (start > 0) and (duration > 1.5) then -- locked out!
					lockouts[ls.school] = duration
					AddCooldown(ls.label, iconGCD, start, duration, "spell", ls.text, "player")
				end
			end
		end
		for name, cd in pairs(MOD.cdSpells) do
			if cd.index then
				local start, duration, enable = GetSpellCooldown(cd.index, "spell")
				if (start > 0) and (enable == 1) and (duration > 1.5) then
					if ((lockouts[cd.school] == nil) or (duration > lockouts[cd.school])) and -- not locked out or on longer cooldown
							((MOD.myClass ~= "DEATHKNIGHT") or CheckRuneCooldown(name, duration)) then -- death knight rune check
						AddCooldown(name, cd.icon, start, duration, "spell link", cd.link, "player")
					end
				end
			end
		end	
		local numSpells = HasPetSpells()
		if numSpells then
			for i = 1, numSpells do
				local start, duration, enable = GetSpellCooldown(i, "pet")
				if (start > 0) and (enable == 1) and (duration > 1.5) then -- don't include global cooldowns
					local name, _, icon = GetSpellInfo(i, "pet")
					if name then
						local hyperlink = GetSpellLink(i, "pet")
						AddCooldown(name, icon, start, duration, "spell link", hyperlink, "pet")
					end
				end
			end
		end
		if UnitIsControlling("player") then
			for slot = 121, 132 do
				local start, duration, enable = GetActionCooldown(slot)
				if (start > 0) and (enable == 1) and (duration > 1.5) then -- don't include global cooldowns
					local actionType, _, _, spellID = GetActionInfo(slot)
					if actionType == "spell" then
						local name, _, icon = GetSpellInfo(spellID)
						if name then
							local hyperlink = GetSpellLink(spellID)
							AddCooldown(name, icon, start, duration, "spell link", hyperlink, "player")
						end
					end
				end
			end
		end	
		for bag = 1, NUM_BAG_SLOTS do
			local numSlots = GetContainerNumSlots(bag)
			for i = 1, numSlots do
				local itemID = GetContainerItemID(bag, i)
				if itemID then
					local start, duration, enable = GetItemCooldown(itemID)
					if (start > 0) and (enable == 1) and (duration > 1.5) then -- don't include global cooldowns or really short cooldowns
						local name, _, _, _, _, itemType, itemSubType, _, _, icon = GetItemInfo(itemID)
						if name then
							if itemType == "Consumable" then -- check for shared cooldowns for potions/elixirs/flasks
								if itemSubType == "Potion" then
									if not MOD:CheckCooldown("Potions") then
										AddCooldown("Potions", iconPotion, start, duration, "text", "Shared Potion Cooldown", "player")
									end
								elseif (itemSubType == "Elixir") or (itemSubType == "Flask") then
									if not MOD:CheckCooldown("Elixirs") then
										AddCooldown("Elixirs", iconElixir, start, duration, "text", "Shared Elixir Cooldown", "player")
									end
								end
							else
								local hyperlink = GetContainerItemLink(bag, i)
								AddCooldown(name, icon, start, duration, "item link", hyperlink, "player")
							end
						end
					end
				end
			end
		end
		CheckInventoryCooldown("HeadSlot")
		CheckInventoryCooldown("NeckSlot")
		CheckInventoryCooldown("BackSlot")
		CheckInventoryCooldown("ShoulderSlot")
		CheckInventoryCooldown("ChestSlot")
		CheckInventoryCooldown("ShirtSlot")
		CheckInventoryCooldown("TabardSlot")
		CheckInventoryCooldown("WristSlot")
		CheckInventoryCooldown("HandsSlot")
		CheckInventoryCooldown("WaistSlot")
		CheckInventoryCooldown("LegsSlot")
		CheckInventoryCooldown("FeetSlot")
		CheckInventoryCooldown("Finger0Slot")
		CheckInventoryCooldown("Finger1Slot")
		CheckInventoryCooldown("Trinket0Slot")
		CheckInventoryCooldown("Trinket1Slot")
		CheckInventoryCooldown("MainHandSlot")
		CheckInventoryCooldown("SecondaryHandSlot")
		CheckInventoryCooldown("RangedSlot")	
		CheckGlobalCooldown()
		MOD.updateCooldowns = false
	end
	if MOD.myClass == "DRUID" then CheckEclipse() end -- special case Eclipse spell
end
