-- Raven is an addon to monitor auras and cooldowns, providing timer bars, action bar highlights, and helpful notifications.

-- Profile.lua contains the default profile for initializing the player's selected profile.
-- It includes routines to process preset aura and cooldown info for all classes and races.
-- Profile settings are accessed by reference to MOD.db.global or MOD.db.profile.
-- It also maintains a persistent database that caches icons, colors and labels for auras and spells.

-- Exported functions for looking up aura and spell-related info:
-- Raven:SetIcon(name) save icon in cache
-- Raven:GetIcon(name) returns cached icon for spell with the specified name, nil if not found
-- Raven:SetColor(name) save color in cache
-- Raven:GetColor(name) returns cached color for spell with the specified name, nil if not found
-- Raven:SetLabel(name) save label in cache
-- Raven:GetLabel(name) returns cached label for spell with the specified name, nil if not found
-- Raven:GetHyperlink(name) return a hyperlink for the spell with the specified name, nil if not found

local MOD = Raven
MOD.BuffTable = {}
MOD.DebuffTable = {}
MOD.CooldownTable = {}

-- Saved variables don't handle being set to nil properly so need to use alternate value to indicate an option has been turned off
local Off = 0 -- value used to designate an option is turned off
local function IsOff(value) return value == nil or value == Off end -- return true if option is turned off
local function IsOn(value) return value ~= nil and value ~= Off end -- return true if option is turned on

-- Convert color codes from hex number to array with r, g, b, a fields (alpha set to 1.0)
function HexColorConvert(hex)
	local n = tonumber(hex, 16)
	local red = math.floor(n / (256 * 256))
	local green = math.floor(n / 256) % 256
	local blue = n % 256

	return { r = red/255, g = green/255, b = blue/255, a = 1.0 }
end

-- Return a copy of a color
function CopyColor(c)
	return { r = c.r, g = c.g, b = c.b, a = c.a }
end

-- Global color palette containing the standard colors for this addon
MOD.ColorPalette = {
	Yellow1=HexColorConvert("fce94f"), Yellow2=HexColorConvert("edd400"), Yellow3=HexColorConvert("c4a000"),
	Orange1=HexColorConvert("fcaf3e"), Orange2=HexColorConvert("f57900"), Orange3=HexColorConvert("ce5c00"),
	Brown1=HexColorConvert("e9b96e"), Brown2=HexColorConvert("c17d11"), Brown3=HexColorConvert("8f5902"),
	Green1=HexColorConvert("8ae234"), Green2=HexColorConvert("73d216"), Green3=HexColorConvert("4e9a06"),
	Blue1=HexColorConvert("729fcf"), Blue2=HexColorConvert("3465a4"), Blue3=HexColorConvert("204a87"),
	Purple1=HexColorConvert("ad7fa8"), Purple2=HexColorConvert("75507b"), Purple3=HexColorConvert("5c3566"),
	Red1=HexColorConvert("ef2929"), Red2=HexColorConvert("cc0000"), Red3=HexColorConvert("a40000"),
	Pink=HexColorConvert("ff6eb4"), Cyan=HexColorConvert("7adbf2"), Gray=HexColorConvert("888a85"),
}

-- Remove unneeded variables from the profile before logout
local function OnProfileShutDown()
	for n, k in pairs(MOD.db.global.SpellIDs) do if k == 0 then MOD.db.global.SpellIDs[n] = nil end end
	MOD:FinalizeBars() -- strip out all default values to significantly reduce profile file size
	MOD:FinalizeConditions() -- strip out temporary values from conditions
	MOD:FinalizeSettings() -- strip default values from layouts
end
	
-- Initialize profile used to customize the addon
function MOD:InitializeProfile()
	-- Add preset aura, cooldown and notification information to the default profile
	MOD:SetSpellDefaults()
	MOD:SetCooldownDefaults()
	MOD:SetConditionDefaults()
	MOD:SetHighlightDefaults()
	MOD:SetIconDefaults()
	MOD:SetDimensionDefaults(MOD.DefaultProfile.global.Defaults)
	MOD:SetFontTextureDefaults(MOD.DefaultProfile.global.Defaults)
	
	-- Get profile from database, providing default profile for initialization
	MOD.db = LibStub("AceDB-3.0"):New("RavenDB", MOD.DefaultProfile)
	MOD.db.RegisterCallback(MOD, "OnDatabaseShutdown", OnProfileShutDown)
	
	MOD:CleanupProfile() -- Cleanup from previous versions, if necessary
	MOD:InitializeSettings() -- Initialize bar group settings with default values
end

-- Cleanup any old profile information, if any
function MOD:CleanupProfile()
end

-- Process an aura description and add info to tables for buffs, debuffs, cooldowns
-- p[1] = name, p[2] = isBuff (nil if not actually an aura), p[3] = color
-- isBuff=true for BUFFs, false for DEBUFFs, nil for other spells
-- color = array of r, g, b, a to be used for associated bars
-- cooldown = true for spells that have a cooldown
-- school = school of magic that the spell belongs to
-- lockout = true for spells that can be used to test if a school is locked out
local function ProcessSpellInfo(p, class, race)
	local name, isBuff, c = p[1], p[2], p[3]	
	MOD.DefaultProfile.global.SpellColors[name] = c -- sets default color in the shared color table
	if p.cooldown then -- add to the cooldown table
		local cd = { class = class, race = race, school = p.school }
		MOD.CooldownTable[name] = cd -- add to table, indexed by name
	end
	if ((class == MOD.myClass) or (race == MOD.myRace)) then
		if p.lockout and p.school then MOD.lockSpells[name] = { school = p.school } end
		if p.cooldown then MOD.cdSpells[name] = { school = p.school } end
	end
	if isBuff ~= nil then -- add to the buff or debuff table
		local aura = { class = class, race = race }
		if isBuff then MOD.BuffTable[name] = aura else MOD.DebuffTable[name] = aura end
	end	
end
	
-- Initialize spells for class auras and cooldowns, also scan other classes for group buffs and cooldowns
-- Buffs, debuffs and cooldowns are tracked in tables containing name, color, class, race
function MOD:SetSpellDefaults()
	for class, at in pairs(MOD.classSpells) do -- look at all classes
		for _, p in pairs(at) do ProcessSpellInfo(p, class, nil) end
	end	
	for race, at in pairs(MOD.racialSpells) do -- look at racial spells
		for _, p in pairs(at) do ProcessSpellInfo(p, nil, race) end
	end	
	for _, p in pairs(MOD.generalSpells) do -- look at general purpose spells
		ProcessSpellInfo(p, nil, nil)
	end
	
	MOD.classSpells = nil -- not used again after initialization so okay to delete
	MOD.racialSpells = nil
	MOD.generalSpells = nil
end

-- Initialize cooldown info from spellbook, should be called whenever spell book changes
function MOD:SetCooldownDefaults()
	for name, cd in pairs(MOD.cdSpells) do cd.index = nil; cd.icon = nil; cd.link = nil end
	local numTabs, numSpells = GetNumSpellTabs(), 0
	for i = 1, numTabs do local _, _, _, n = GetSpellTabInfo(i); numSpells = numSpells + n end
	if numSpells then
		for i = 1, numSpells do
			local name, _, icon = GetSpellInfo(i, "spell")
			local cd = MOD.cdSpells[name]
			if cd then cd.index = i; cd.icon = icon; cd.link = GetSpellLink(i, "spell") end
			local ls = MOD.lockSpells[name]
			if ls then ls.index = i; ls.label = ls.school .. " School"; ls.text = "Locked out of " .. ls.school .. " school of magic." end
		end
	end
end

-- Reset a particular spell color to its default value
function MOD:ResetColorDefault(name)
	if name then
		local dct = MOD.DefaultProfile.global.SpellColors
		local sct = MOD.db.global.SpellColors
		local c = dct[name]
		if not c then
			sct[name] = nil -- if not default value then just clear the spell color
		else
			local t = sct[name]
			if t then
				t.r, t.g, t.b, t.a = c.r, c.g, c.b, c.a
			else
				sct[name] = CopyColor(c)
			end
		end	
	end
end

-- Reset all colors to default values
function MOD:ResetColorDefaults()
	local dct = MOD.DefaultProfile.global.SpellColors
	local sct = MOD.db.global.SpellColors
	for n, c in pairs(dct) do -- copy all original values from the default color table
		local t = sct[n]
		if t then
			t.r, t.g, t.b, t.a = c.r, c.g, c.b, c.a
		else
			sct[n] = CopyColor(c)
		end
	end 
	for n in pairs(sct) do if not dct[n] then sct[n] = nil end end -- remove any extras
end

-- Initialize cache of icons
local iconCache = {}
function MOD:SetIconDefaults()
	local tcount = GetNumSpellTabs()
	for i = 1, tcount do
		local tname, _, toffset, tspells = GetSpellTabInfo(i)	
		
		for k = toffset + 1, toffset + tspells do
			local name, icon
			if GetSpellName then -- pre-Cataclysm
				name = GetSpellName(k, BOOKTYPE_SPELL)
				icon = GetSpellTexture(k, BOOKTYPE_SPELL)
			else -- Cataclysm changes calls to get name and texture for spell book entries
				name = GetSpellBookItemName(k, BOOKTYPE_SPELL)
				icon = GetSpellBookItemTexture(k, BOOKTYPE_SPELL)
			end
			iconCache[name] = icon
		end
	end
end
	
-- Initialize conditions for the player's class
function MOD:SetConditionDefaults()
	local nt = MOD.classConditions[MOD.myClass]
	MOD.DefaultProfile.profile.Conditions[MOD.myClass] = nt
end

-- Initialize dimension defaults
function MOD:SetDimensionDefaults(p)
	p.barWidth = 150; p.barHeight = 15; p.iconSize = 15; p.scale = 1
	p.spacingX = 0; p.spacingY = 0; p.iconOffsetX = 0; p.iconOffsetY = 0
end

-- Initialize default fonts and textures
function MOD:SetFontTextureDefaults(p)
	p.labelFont = "Arial Narrow"; p.labelFSize = 10; p.labelAlpha = 1; p.labelColor = { r = 1, g = 1, b = 1, a = 1 }; p.labelOutline = false
	p.timeFont = "Arial Narrow"; p.timeFSize = 10; p.timeAlpha = 1; p.timeColor = { r = 1, g = 1, b = 1, a = 1 }; p.timeOutline = false
	p.iconFont = "Arial Narrow"; p.iconFSize = 10; p.iconAlpha = 1; p.iconColor = { r = 1, g = 1, b = 1, a = 1 }; p.iconOutline = true
	p.texture = "Blizzard"; p.bgtexture = "Blizzard"; p.alpha = 1; p.bgAlpha = 0.65
	p.fgSaturation = 0; p.fgBrightness = 0; p.bgSaturation = 0; p.bgBrightness = 0
end

-- Find and cache spell ids (this should be used rarely, primarily when entering spell names manually
local function GetSpellID(name)
	local id = nil
	id = MOD.db.global.SpellIDs[name]
	if id == 0 then return nil end -- only scan invalid ones once in a session
	if id and (name ~= GetSpellInfo(id)) then id = nil end -- verify it is still valid

	if not id then
		id = 0
		while id < 100000 do
			id = id + 1
			local n = GetSpellInfo(id)
			if n == name then
				MOD.db.global.SpellIDs[n] = id
				return id
			end
		end
		MOD.db.global.SpellIDs[name] = 0 -- this marks an invalid spell name
		id = nil
	end
	return id
end

-- Add a texture to the icons cache
function MOD:SetIcon(name, texture)
	if name and texture then iconCache[name] = texture end
end

-- Get a texture from the icons cache, if not there try to get by spell name and cache if found.
-- If not found then look up spell identifier and use it to locate a texture.
function MOD:GetIcon(name)
	if not name or (name == "none") or (name == "") then return nil end
	local tex = iconCache[name] -- this is initialized from the player's spell book
	if not tex then
		local n
		n, _, tex = GetSpellInfo(name) -- first try to find it based on the name
		if n and tex then
			iconCache[name] = tex -- only cache textures found by looking up the name
		else
			local id = GetSpellID(name)
			if id then _, _, tex = GetSpellInfo(id) end -- then try based on id
		end
	end
	return tex
end

-- Get a hyperlink for a spell, looking up as a spell identifier if name is unknown
function MOD:GetHyperlink(name)
	if not name or (name == "") then return nil end
	local link = GetSpellLink(name) -- first try to find it based on the name
	if not link then
		local id = GetSpellID(name)
		if id then link = GetSpellLink(id) end -- then try based on id
	end
	return link
end

-- Add a color to the cache, update values in case they have changed
function MOD:SetColor(name, c)
	if name and c then
		local t = MOD.db.global.SpellColors[name]
		if t then
			t.r, t.g, t.b, t.a = c.r, c.g, c.b, c.a
		else
			MOD.db.global.SpellColors[name] = CopyColor(c)
		end
	end
end

-- Get a color from the cache of given name, but if not in cache return default value for the specified type
-- If type is not "Buff", "Debuff", "Cooldown", or "Notification" then it is possible for this to return nil
function MOD:GetColor(name, t)
	local c = MOD.db.global.SpellColors[name]
	if not c then
		if t == "Buff" then
			c = MOD.db.global.DefaultBuffColor
		elseif t == "Debuff" then
			c = MOD.db.global.DefaultDebuffColor
		elseif t == "Cooldown" then
			c = MOD.db.global.DefaultCooldownColor
		elseif t == "Notification" then
			c = MOD.db.global.DefaultNotificationColor
		end
	end
	return c
end

-- Add a label to the cache but only if different from name
function MOD:SetLabel(name, label)
	if name and label then
		if name == label then MOD.db.global.Labels[name] = nil else MOD.db.global.Labels[name] = label end
	end
end

-- Get a label from the cache, but if not in the cache then return the name
function MOD:GetLabel(name)
	local label = MOD.db.global.Labels[name]
	if not label then label = name end
	return label
end

-- Default profile description used to initialize the SavedVariables persistent database
MOD.DefaultProfile = {
	global = {
		Labels = {},					-- cache of labels for actions and spells
		SpellColors = {},				-- cache of colors for actions and spells
		SpellIDs = {},					-- cache of spell ids that had to be looked up
		Settings = {},					-- settings table indexed by bar group names
		Defaults = {},					-- default settings for bar group fonts and textures
		FilterBuff = {},				-- shared bar group filter for buffs
		FilterDebuff = {},				-- shared bar group filter for debuffs
		FilterCooldown = {},			-- shared bar group filter for cooldowns
		BuffDurations = {},				-- cache of buff durations used for weapon buffs
		DefaultBuffColor = HexColorConvert("8ae234"), -- Green1
		DefaultDebuffColor = HexColorConvert("fcaf3e"), -- Orange1
		DefaultCooldownColor = HexColorConvert("fce94f"), -- Yellow1
		DefaultNotificationColor = HexColorConvert("729fcf"), -- Blue1
		DefaultPoisonColor = CopyColor(DebuffTypeColor["Poison"]),
		DefaultCurseColor = CopyColor(DebuffTypeColor["Curse"]),
		DefaultMagicColor = CopyColor(DebuffTypeColor["Magic"]),
		DefaultDiseaseColor = CopyColor(DebuffTypeColor["Disease"]),
		PlayerBuffHighlights = true, 	-- enable highlights for player buffs
		PlayerDebuffHighlights = true,	-- enable highlights for player debuffs
		TargetBuffHighlights = true,	-- enable highlights for target buffs
		TargetDebuffHighlights = true,	-- enable highlights for target debuffs
		FocusBuffHighlights = true,		-- enable highlights for target buffs
		FocusDebuffHighlights = true,	-- enable highlights for target debuffs
		PlayerBuffColor = HexColorConvert("8ae234"), -- Green1
		PlayerDebuffColor = HexColorConvert("fcaf3e"), -- Orange1
		TargetBuffColor = HexColorConvert("3465a4"), -- Blue2
		TargetDebuffColor = HexColorConvert("cc0000"), -- Red2
		FocusBuffColor = HexColorConvert("ad7fa8"), -- Purple1
		FocusDebuffColor = HexColorConvert("fce94f"), -- Yellow1
		FlashExpiring = true,			-- enable flashing of expiring buffs and debuffs
		FlashTime = 5,					-- how many seconds to flash before expiration
		CooldownText = true,			-- enable cooldown text overlays
		CooldownFont = "Arial Narrow",	-- default font for cooldown text
		CooldownFsize = 10,				-- default font size for cooldown text
		CooldownTimeFormat = 24,		-- default time format for cooldown text
		CooldownTimeSpaces = false,		-- default is no spaces in cooldown text
		CooldownTimeCase = false,		-- default is lowercase for cooldown text
		ButtonFacade = false,			-- enable highlights support for ButtonFacade if it is loaded
	},
	profile = {
		enabled = true,					-- enable Raven
		hideBlizz = true,				-- enable hiding the Blizzard buff and temp enchant frames
		BarGroups = {},					-- bar group options to be filled in and saved between sessions
		Conditions = {}, 				-- includes default reminders for common buffs
		ButtonFacadeSkin = {}			-- skin settings from ButtonFacade
	},
}
