-- Raven is an addon to monitor auras and cooldowns, providing timer bars, action bar highlights, and helpful notifications.
-- Raven_Options is a load-on-demand module that only gets loaded and initialized when the user requests the GUI

-- Options.lua contains the tables used by the options panel as well as all the supporting functions for displaying
-- current settings, bar groups, conditions, etc.

local MOD = Raven
local acereg = LibStub("AceConfigRegistry-3.0")
local acedia = LibStub("AceConfigDialog-3.0")
local initialized = false -- set when options are first accessed

local standardPlayerBuffs = false -- standard bar groups for "getting started"
local standardPlayerDebuffs = false
local standardCooldowns = false
local standardTarget = false
local standardFocus = false
local standardRunes = false
local standardNotifications = false
local standardLongBuffs = false

local enterNewBarGroup = false -- trigger showing bar group name entry
local toggleNewBarGroup = false -- used for automatic cancel
local typeNewBarGroup = false -- if true then auto-added bars, otherwise manually added
local enterNewBar = false -- used to trigger "enter bar mode"
local newbar = nil -- new bar template for editing
local oldbar = nil -- cache previous new bar settings
local configlist = nil -- cache pointer to generated configuration list

local enterNewCondition= false -- trigger showing notification name entry
local toggleNewCondition = false -- used for automatic cancel
local selectedDependency = nil -- used for selecting dependency condition
local selectedCondition = nil -- used for finding out when switching conditions
local tableSelection = {} -- stores current selection for a profile table
local selectedSpellName = nil -- stores current selection for a spell name in defaults panel

-- Saved variables don't handle being set to nil properly so need to use alternate value to indicate an option has been turned off
local Off = 0 -- value used to designate an option is turned off
local function IsOff(value) return value == nil or value == Off end -- return true if option is turned off
local function IsOn(value) return value ~= nil and value ~= Off end -- return true if option is turned on

local function InMode(t) -- return true if locked into entry mode of specified type
	if t == "Bar" then return enterNewBar end
	if t == "BG" then return enterNewBar or enterNewBarGroup end
	if t == "Not" then return enterNewCondition end
	return enterNewBar or enterNewBarGroup or enterNewCondition
end 

-- Check if spell name is valid and display a warning if it is not, return validation result
local function ValidateSpellName(name)
	if not name or (name == "") then return false end
	if not MOD:GetIcon(name) then
		print(string.format("Raven: warning \"%s\" is not a valid spell name", name))
		return false
	end
	return true
end

-- Update the addon when the profile changes
local function OnProfileChanged()
	tableSelection = {} -- necessary since this is not cached in the profile itself
	MOD:InitializeBars() -- required because of the linkage between the profile and LibBars
end

-- Initialize options for the configuration GUI
-- Register the options table and link to the Blizz addons interface
local function InitializeOptions()
	initialized = true -- only do this once	
	local options = MOD.OptionsTable	
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(MOD.db) -- fill in the profile section
	options.args.profile.disabled = function(info) return InMode() end,
	
	acereg:RegisterOptionsTable("Raven", options)
	acereg:RegisterOptionsTable("Raven: "..options.args.FrontPage.name, options.args.FrontPage)
	acereg:RegisterOptionsTable("Raven: "..options.args.BarGroups.name, options.args.BarGroups)
	acereg:RegisterOptionsTable("Raven: "..options.args.Conditions.name, options.args.Conditions)
	acereg:RegisterOptionsTable("Raven: "..options.args.Highlights.name, options.args.Highlights)
	acereg:RegisterOptionsTable("Raven: "..options.args.profile.name, options.args.profile)
	acereg:RegisterOptionsTable("Raven Blizz", MOD.TopOptionsTable)

	acedia:AddToBlizOptions("Raven Blizz", "Raven")
	acedia:AddToBlizOptions("Raven: "..options.args.FrontPage.name, options.args.FrontPage.name, "Raven")
	acedia:AddToBlizOptions("Raven: "..options.args.BarGroups.name, options.args.BarGroups.name, "Raven")
	acedia:AddToBlizOptions("Raven: "..options.args.Conditions.name, options.args.Conditions.name, "Raven")
	acedia:AddToBlizOptions("Raven: "..options.args.Highlights.name, options.args.Highlights.name, "Raven")
	acedia:AddToBlizOptions("Raven: "..options.args.profile.name, options.args.profile.name, "Raven")

	local w, h = 860, 680 -- somewhat arbitrary numbers that seem to work for the configuration dialog layout
	acedia:SetDefaultSize("Raven", w, h)
	
	MOD.db.RegisterCallback(MOD, "OnProfileChanged", OnProfileChanged)
	MOD.db.RegisterCallback(MOD, "OnProfileCopied", OnProfileChanged)
	MOD.db.RegisterCallback(MOD, "OnProfileReset", OnProfileChanged)
end

-- Update options in case anything changes
function MOD:UpdateOptions()
	if initialized and acedia.OpenFrames["Raven"] then
		acereg:NotifyChange("Raven")
	end
end

-- Toggle display of the options panel
function MOD:ToggleOptions()
	if not initialized then InitializeOptions() end
	if acedia.OpenFrames["Raven"] then
		acedia:Close("Raven")
	else
		acedia:Open("Raven")
	end
	collectgarbage("collect")
end

-- Generic function to build and return a sorted list of names stored in a profile table
-- Assumes table entries are arrays with the names indexed by "name"
local function GetNameList(id, t)
	local i, list = 0, {} -- build the name list from the profile table
	
	if IsOff(t) then
		tableSelection[id] = nil -- no selection for empty list
		return list -- empty list if t is not defined
	end

	local selection = tableSelection[id] -- make sure still a valid selection
	local found = false
	for _, n in pairs(t) do
		if IsOn(n) and n.name then
			i = i + 1
			list[i] = n.name
			if n.name == selection then found = true end
		end
	end
	table.sort(list)
	
	if not found and (i > 0) then -- check that current selection, if any, was found in the table
		tableSelection[id] = list[1] -- default is first entry in the list
	end
	return list
end

-- Generic function to add a new entry to a profile table, assumes "name" is field in the entry
local function AddNameEntry(id, t, newEntry)
	t[newEntry.name] = newEntry -- add the new entry indexed by its name to facilitate reverse lookup
	GetNameList(id, t) -- initializes the cached settings, if necessary
	tableSelection[id] = newEntry.name
end

-- Generic function to delete a name entry from a profile table
local function DeleteNameEntry(id, t)
	-- Get the cached list of names and the selected name
	local list = GetNameList(id, t)
	local selection = tableSelection[id]
	if not selection then return end -- empty table
	
	-- Look in the profile table to find the matching entry
	for k, n in pairs(t) do
		if IsOn(n) and (n.name == selection) then
			t[k] = Off -- remove the table entry by setting it to Off so that it will not come back when reload profile
			tableSelection[id] = nil -- will be set on next call to GetNameList, if table not empty
			return
		end
	end
end

-- Generic function to return index into the sorted name list of current selection
local function GetNameSelection(id, t)
	-- Get the cached list of names and the selected name
	local list = GetNameList(id, t)
	local selection = tableSelection[id]
	if not selection then return nil end -- empty table
	
	-- Look in the cached list to find the index of the selection
	for pos, n in pairs(list) do
		if n == selection then return pos end
	end
	return nil -- should never get here, just in case return nil
end

-- Generic function to set the selection in a name list
local function SetNameSelection(id, t, value)
	local list = GetNameList(id, t)
	if value > 0 and value <= table.maxn(list) then
		tableSelection[id] = list[value]
	else
		tableSelection[id] = nil
	end
end

-- Generic function to return table position (nil if empty table) and pointer to the currently selection in the profile table
local function GetNameEntry(id, t)
	-- Get the cached list of names and the selected name
	local list = GetNameList(id, t)
	local selection = tableSelection[id]
	if not selection then return nil end -- empty table
	
	-- Look in the profile table to find the matching entry
	for _, n in pairs(t) do
		if IsOn(n) then
			if n.name == selection then return n end
		end
	end
	return nil -- should never get here, just in case return nil
end

-- Functions for accessing bar group lists
local function GetBarGroupList() return GetNameList("BarGroups", MOD.db.profile.BarGroups) end
local function GetBarGroupEntry() return GetNameEntry("BarGroups", MOD.db.profile.BarGroups) end
local function GetSelectedBarGroup() return GetNameSelection("BarGroups", MOD.db.profile.BarGroups) end
local function SetSelectedBarGroup(value) SetNameSelection("BarGroups", MOD.db.profile.BarGroups, value) end
local function NoBarGroup() return enterNewBarGroup or (GetSelectedBarGroup() == nil) end

local function CreateBarGroup(name, auto, link, offsetX, offsetY)
	local bp = {}
	for n, k in pairs(MOD.db.global.Defaults) do bp[n] = k end -- current default settings for layouts fonts textures
	for n, k in pairs(MOD.BarGroupTemplate) do bp[n] = k end -- add settings in the template
	bp.name = name
	bp.auto = auto
	bp.linkSettings = link
	bp.locked = false
	bp.bars = {}

	AddNameEntry("BarGroups", MOD.db.profile.BarGroups, bp)
	MOD:InitializeBarGroup(bp, offsetX, offsetY)
	MOD:UpdateAllBarGroups()
	return bp
end

local function DeleteBarGroup()
	local bg = GetBarGroupEntry()
	local oldName = bg.name
	MOD:ReleaseBarGroup(bg)
	DeleteNameEntry("BarGroups", MOD.db.profile.BarGroups)
	for _, bp in pairs(MOD.db.profile.BarGroups) do if IsOn(bp) and (bp.anchor == oldName) then bp.anchor = false end end
	MOD:UpdateAllBarGroups()
end

-- Rename involves deleting current bar group and then adding it back renamed
local function RenameBarGroup(newName)
	local bg = GetBarGroupEntry()
	local oldName = bg.name
	MOD:ReleaseBarGroup(bg)
	DeleteNameEntry("BarGroups", MOD.db.profile.BarGroups)
	bg.name = newName
	AddNameEntry("BarGroups", MOD.db.profile.BarGroups, bg)
	MOD:InitializeBarGroup(bg, 0, 0)
	for _, bp in pairs(MOD.db.profile.BarGroups) do if IsOn(bp) and (bp.anchor == oldName) then bp.anchor = newName end end
	MOD:UpdateAllBarGroups()
end

-- Check if any bar groups currently exist...
local function CheckBarGroupsExist()
	local bgs = MOD.db.profile.BarGroups
	for n, bg in pairs(bgs) do
		if IsOn(bg) then
			if standardPlayerBuffs and (n == "Short Buffs") then return true end
			if standardLongBuffs and (n == "Long Buffs") then return true end
			if standardPlayerDebuffs and (n == "Debuffs") then return true end
			if standardCooldowns and (n == "Cooldowns") then return true end
			if standardTarget and (n == "Target") then return true end
			if standardFocus and (n == "Focus") then return true end
			if standardRunes and (n == "Runes") then return true end
			if standardNotifications and (n == "Notifications") then return true end
		end
	end
	return false
end

-- Check if a bar group with the name already exists, return a confirmation string if it does, otherwise return false
local function ConfirmNewBarGroup(name)
	local bg = MOD.db.profile.BarGroups[name]
	if IsOn(bg) then -- name already exists
		return "DUPLICATE NAME\nThere is already a bar group named " .. name .. ". Do you want to overwrite it?"
	end
	return false
end

-- Copy layout configuration settings from another bar group
local function CopyBarGroupConfiguration(name)
	local bg = GetBarGroupEntry()
	local abg = MOD.db.profile.BarGroups[name]
	if IsOn(abg) and (bg.name ~= name) then -- make sure not same bar groups
		for n in pairs(MOD.BarGroupLayoutTemplate) do bg[n] = abg[n] end
		MOD:UpdateAllBarGroups()
	end
end

-- Functions for accessing bars, special case since the group tree function to list bars
local function GetBarEntryWithLabel(bname)
	local bg = GetBarGroupEntry()
	if not bg then return nil end

	for k, n in pairs(bg.bars) do if n.barLabel == bname then return n, k end end
	return nil
end

local function GetBarEntry(info)
	local barlist = MOD.OptionsTable.args.BarGroups.args.BarTab.args
	local bname = barlist[info[#info-2]].name
	return GetBarEntryWithLabel(bname)
end

-- Return true if there is no bar (or we are in the middle of creating a new bar)
local function NoBar()
	local bg = GetBarGroupEntry()
	if not bg or InMode("Bar") then return true end
	return next(bg.bars) == nil
end

-- Functions for sorting bar lists
local function SortAlphaUp(a, b) return string.upper(a) < string.upper(b) end
local function SortCustomOrder(a, b) return GetBarEntryWithLabel(a).sorder < GetBarEntryWithLabel(b).sorder end

-- Populate the options table for all the bars in the currently selected bar group
local function UpdateBarList()
	local bg = GetBarGroupEntry()
	if not bg then return end

	local barlist = MOD.OptionsTable.args.BarGroups.args.BarTab.args
	for j in pairs(barlist) do -- first remove any old entries
		if string.find(j, "XXX") then barlist[j] = nil end
	end
	if bg.auto then return end -- no bars if auto bar group
	
	local i, bp = 0, {} -- build a list of bar names from the profile table
	for _, n in pairs(bg.bars) do
		i = i + 1
		bp[i] = n.barLabel
	end

	if bg.sor ~= "X" then table.sort(bp, SortAlphaUp) else table.sort(bp, SortCustomOrder) end -- sort names by appropriate sorting algorithm
	
	for k, bname in pairs(bp) do -- now populate the groups in the tree with bar options
		barlist["XXX"..k] = {
			type = "group", order = k + 10, name = bname,
			disabled = function(info) return InMode("Bar") end,
			args = MOD.barOptions,
		}
		local bar = GetBarEntryWithLabel(bname)
		bar.sorder = k -- normalizes the sorder values to ascending integers
	end
	
	local key = "EnterNewBar"
	if enterNewBar then -- support entering new bar into the current bar group
		barlist[key] = { type = "group", order = 10, name = "-- New Bar --", args = MOD.barOptions }
		acedia:SelectGroup("Raven", "BarGroups", "BarTab", key)
	else
		barlist[key] = nil
	end
end

-- Select a particular bar
local function SelectBar(label)
	local barlist = MOD.OptionsTable.args.BarGroups.args.BarTab.args
	for k, b in pairs(barlist) do
		if string.find(k, "XXX") then -- only check special keys for bars
			if b.name == label then
				acedia:SelectGroup("Raven", "BarGroups", "BarTab", k)
				return
			end
		end
	end
end

-- Delete the current bar
local function DeleteBar()
	local bg = GetBarGroupEntry()
	if not bg then return end
	
	local statustable = acedia:GetStatusTable("Raven", { "BarGroups", "BarTab" })
	local selected = statustable.groups.selected

	if string.find(selected, "XXX") then -- only check special keys for bars
		local barlist = MOD.OptionsTable.args.BarGroups.args.BarTab.args
		local bname = barlist[selected].name
		local bar, key = GetBarEntryWithLabel(bname)
		if bar then -- remove the table entry, essential since recycling would require reset of special values
			bg.bars[key] = nil
			UpdateBarList()
		end
	end
end

-- Default bar information tables
local barSources = { "Death Knight", "Druid", "Hunter", "Mage", "Paladin", "Priest", "Rogue", "Shaman", "Warlock", "Warrior", "Racial", "Spell" }
local classIndex = { DEATHKNIGHT = 1, DRUID = 2, HUNTER = 3, MAGE = 4, PALADIN = 5, PRIEST = 6, ROGUE = 7, SHAMAN = 8, WARLOCK = 9, WARRIOR = 10 }
local classList = { "DEATHKNIGHT", "DRUID", "HUNTER", "MAGE", "PALADIN", "PRIEST", "ROGUE", "SHAMAN", "WARLOCK", "WARRIOR" }
local unitList = { player = "Player", pet = "Pet", target = "Target", focus = "Focus" }

-- Add a new bar to a bar group, initializing the sort order value and adding a unique identifier
-- Verify that the bar is not a duplicate of one already in the bar group before adding
local lastdatecheck = ""
local lastdatecount = 0
function AddBarToGroup(bg, bar)
	for k, n in pairs(bg.bars) do -- scan current bars and see if this bar is a duplicate, only tricky part is the castBy setting
		local ok = true
		if (bar.action == n.action) and (bar.barType == n.barType) and (bar.monitor == n.monitor) then
			if bar.castBy == n.castBy then return false end -- everything matches so is a duplicate for sure
			if n.castBy == "Anyone" then return false end -- if already have everything then duplicate
			if bar.castBy == "Anyone" then bg.bars[k] = nil; ok = false end -- more inclusive that the original so need to delete the original
		end
		if ok and (bar.barLabel == n.barLabel) then -- found duplicate label, add a suffix 
			bar.barLabel = bar.barLabel .. "*"
			bar.labelLink = true -- make sure not linked (true means not linked)
		end
	end
	
	bar.sorder = 0
	local d = date("%m%d%y-%H%M%S-")
	if d == lastdatecheck then
		lastdatecount = lastdatecount + 1
	else
		lastdatecheck = d
		lastdatecount = 0
	end
	
	bar.uniqueID = d .. lastdatecount
	table.insert(bg.bars, bar)
	return true
end

-- Process entering a new bar, state = "start", "ok", "cancel" to move between stages of bar entry
local function EnterNewBar(state)
	local bg = GetBarGroupEntry()
	if not bg then return end

	if state == "start" then
		enterNewBar = true
		if oldbar then -- cache of previous settings
			newbar = { -- only preserve selected fields from saved bar description
				barType = oldbar.barType, barSource = oldbar.barSource, barLabel = "",
				monitor = oldbar.monitor, isMine = oldbar.isMine, isOther = oldbar.isOther
			}
			oldbar = nil
		else
			newbar = { barType = "Notification", barSource = classIndex[MOD.myClass], barLabel = "", monitor = "player", isMine = true, isOther = false }
		end
	elseif state == "ok" then
		enterNewBar = false
		
		local castBy = "Player"
		if newbar.isOther then if newbar.isMine then castBy = "Anyone" else castBy = "Other" end end
		local source = barSources[newbar.barSource]
		
		if (newbar.barType ~= "Notification") and (newbar.barSource == 12) and ValidateSpellName(selectedSpellName) then
			if not newbar.actionList then newbar.actionList = {} end
			if not newbar.selectList then newbar.selectList = {} end
			newbar.actionList[1] = selectedSpellName
			newbar.selectList[1] = true
		end
		
		local sel = nil
		if (newbar.barType == "Buff") or (newbar.barType == "Debuff") then -- buff or debuff
			if newbar.actionList then
				for k, n in pairs(newbar.selectList) do
					if n then
						local bname = newbar.actionList[k]
						local label = MOD:GetLabel(bname)
						local bar = { action = bname, enableBar = true, barLabel = label, barType = newbar.barType, barSource = source,
								monitor = newbar.monitor, castBy = castBy }
						if AddBarToGroup(bg, bar) then
							if not sel then sel = bar.barLabel end
						end
					end
				end
			end
		elseif newbar.barType == "Cooldown" then -- cooldown
			if newbar.actionList then
				for k, n in pairs(newbar.selectList) do
					if n then
						local bname = newbar.actionList[k]
						local label = MOD:GetLabel(bname)
						local bar = { action = bname, enableBar = true, barLabel = label, barType = newbar.barType, barSource = source,
								monitor = nil, castBy = nil, enableReady = nil, readyNotUsable = nil }
						if AddBarToGroup(bg, bar) then
							if not sel then sel = bar.barLabel end
						end
					end
				end
			end
		elseif newbar.barType == "Notification" then -- notification
			if newbar.conditionList and newbar.selectCondition then
				for k, n in pairs(newbar.selectCondition) do
					if n then
						local bname = newbar.conditionList[k]
						local bar = { action = bname, enableBar = true, barLabel = bname, barType = newbar.barType, barSource = nil,
								monitor = nil, castBy = nil, }
						if AddBarToGroup(bg, bar) then
							if not sel then sel = bar.barLabel end
						end
					end
				end
			end
		end
		if sel then UpdateBarList(); SelectBar(sel) end
		oldbar = newbar
		newbar = nil
	elseif state == "cancel" then
		enterNewBar = false
		oldbar = newbar
		newbar = nil
	end
end

-- Get a formatted text description of the current bar
local function GetBarDescription(info)
	local a = ""
	local n = GetBarEntry(info)
	if n then
		if n.barType == "Notification" then
			a = string.format("|cFF7adbf2Type:|r %s\n|cFF7adbf2Condition:|r %s", n.barType, n.action)
			local spell = MOD:GetAssociatedSpellForBar(n)
			if spell then a = a .. string.format("\n|cFF7adbf2Associated Spell:|r %s", spell) end
		elseif n.barType == "Buff" or n.barType == "Debuff" or n.barType == "Cooldown" then
			if n.action then
				a = string.format("|cFF7adbf2Action:|r %s\n|cFF7adbf2Type:|r %s\n|cFF7adbf2Source:|r %s", n.action, n.barType, n.barSource)
			else
				a = string.format("|cFF7adbf2Type:|r %s\n|cFF7adbf2Source:|r %s", n.barType, n.barSource)
			end
			if n.barType ~= "Cooldown" then
				a = a .. string.format("\n|cFF7adbf2Action On:|r %s\n|cFF7adbf2Cast By:|r %s", unitList[n.monitor], n.castBy)
			end
		end
	end
	return a
end

-- Move a bar up or down in the list
local function MoveBarInList(info, direction)
	local bar = GetBarEntry(info)
	if bar then
		if direction == "up" then bar.sorder = bar.sorder - 1.5 else bar.sorder = bar.sorder + 1.5 end
		UpdateBarList()
		SelectBar(bar.barLabel)
	end
end

-- Return a list for selection of configuration
local function GetOrientationList()
	if not configlist then -- build it from the supported configurations table
		configlist = {}
		for i, config in pairs(Nest_SupportedConfigurations) do configlist[i] = config.name end
	end
	return configlist
end

-- Set a field in the current bar
local function SetBarField(info, fname, value)
	local n = GetBarEntry(info)
	if n then
		n[fname] = value
	end
end

-- Get a field in the current bar
local function GetBarField(info, fname)
	local n = GetBarEntry(info)
	if n then return n[fname] end
	return nil
end

-- Get the icon associated with the current bar
local function GetBarIcon(info)
	local b = GetBarEntry(info)
	return MOD:GetIconForBar(b)
end

-- Check for duplicate labels, including "other labels"
local function IsDuplicateLabel(bg, newLabel)
	for _, n in pairs(bg.bars) do if n.barLabel == newLabel then return true end end
	return false
end

-- Change a bar's label, have to redo the list afterwards, also do some validation and return false if not valid
local function SetBarLabel(info, newLabel, isLinked)
	local bg = GetBarGroupEntry()
	if not bg then return false end
	
	local bar = GetBarEntry(info)
	if not bar then return false end
	
	if bar.barLabel == newLabel then return true end  -- make sure new label is actually a change
	if newLabel == "" then return false end -- make sure not an empty string
	
	if IsDuplicateLabel(bg, newLabel) then -- make sure not a duplicate within this bar group
		print('"' .. newLabel .. '" is a duplicate of an existing bar label')
		return false
	end

	local a, b = string.find(newLabel, "^%a[%a%d%b()%[%] %-%+%!%*_':]*") -- make sure new label has correct format and doesn't contain bad characters
	if a ~= 1 and b ~= string.len(newLabel) then
		print('"' .. newLabel .. '" does not begin with a letter or contains restricted characters')
		return false
	end 

	DeleteBar()
	if isLinked and ((bar.barType == "Buff") or (bar.barType == "Debuff") or (bar.barType == "Cooldown")) then MOD:SetLabel(bar.action, newLabel) end
	bar.barLabel = newLabel
	table.insert(bg.bars, bar)
	UpdateBarList()
	SelectBar(newLabel)
	return true
end

-- Generate a list of auras that meet the criteria of type and source specified in the newbar table
local function GenerateBarActionList()
	local i = 0
	newbar.actionList = {}
	newbar.selectList = {}
	selectedSpellName = nil -- for entered spell names

	local btype = newbar.barType
	if (btype == "Buff") or (btype == "Debuff") or (btype == "Cooldown") then
		local btable = MOD[btype .. "Table"]
		if newbar.barSource <= 10 then
			local c = classList[newbar.barSource]
			for name, action in pairs(btable) do
				if (action.class == c) then i = i + 1; newbar.actionList[i] = name; end
			end
		elseif newbar.barSource == 11 then -- racial buffs
			for name, action in pairs(btable) do
				if action.race then i = i + 1; newbar.actionList[i] = name; end
			end
		end
	end

	table.sort(newbar.actionList) -- alphabetical order to improve readability
	for k in pairs(newbar.actionList) do newbar.selectList[k] = false end -- default to all not selected
	newbar.actionListCount = i
end

-- Functions for accessing bar type lists
local function SetSelectedBarType(value)
	if newbar.barType ~= value then
		newbar.barType = value
		GenerateBarActionList()
	end
end

-- Functions for accessing bar source lists
local function SetSelectedBarSource(value)
	if newbar.barSource ~= value then 
		newbar.barSource = value
		GenerateBarActionList()
	end
end

-- Generate a list of conditions to select from
local function GetSelectConditionList()
	local i, t = 0, {}
	for _, n in pairs(MOD.db.profile.Conditions[MOD.myClass]) do
		if IsOn(n) and n.name then
			i = i + 1
			t[i] = n.name
		end
	end
	table.sort(t)
	return t, i
end

-- Get the, potentially cached, list of conditions to select from
local function GetBarConditionList()
	if not newbar.conditionList then
		newbar.conditionList, newbar.conditionListCount = GetSelectConditionList()
		newbar.selectCondition = {}
		for k in pairs(newbar.conditionList) do newbar.selectCondition[k] = false end -- default to all not selected
	end
	return newbar.conditionList
end

local function GetBarConditionListCount() GetBarConditionList(); return newbar.conditionListCount end
local function GetSelectedBarCondition(key) return newbar.selectCondition[key] end
local function SetSelectedBarCondition(key, value) newbar.selectCondition[key] = value end
local function SetAllBarConditions(value) for k in pairs(newbar.conditionList) do newbar.selectCondition[k] = value end end

-- Generate a list of auras that meet the criteria of type and source specified in the newbar table
local function GetBarActionList() if not newbar.actionList then GenerateBarActionList() end; return newbar.actionList end
local function GetBarActionListCount() if not newbar.actionList then GenerateBarActionList() end; return newbar.actionListCount end
local function GetSelectedBarAction(key) return newbar.selectList[key] end
local function SetSelectedBarAction(key, value) newbar.selectList[key] = value end
local function SetAllBarActions(value) for k in pairs(newbar.actionList) do newbar.selectList[k] = value end end

-- Functions for accessing condition lists
local function GetConditionList() return GetNameList("Conditions", MOD.db.profile.Conditions[MOD.myClass]) end
local function GetCondition() return GetNameEntry("Conditions", MOD.db.profile.Conditions[MOD.myClass]) end
local function GetSelectedCondition() return GetNameSelection("Conditions", MOD.db.profile.Conditions[MOD.myClass]) end
local function SetSelectedCondition(value) SetNameSelection("Conditions", MOD.db.profile.Conditions[MOD.myClass], value) end
local function DeleteCondition() DeleteNameEntry("Conditions", MOD.db.profile.Conditions[MOD.myClass]) end
local function NoCondition() return enterNewCondition or (GetSelectedCondition() == nil) end

local function CreateCondition(name)
	local n = { enabled = true, name = name, notify = true, tooltip = true, tests = {}, testResult = false, dependencies = {}, result = false }
	AddNameEntry("Conditions", MOD.db.profile.Conditions[MOD.myClass], n)
end

-- Check if a condition with the name already exists, return a confirmation string if it does, otherwise return false
local function ConfirmNewCondition(name)
	local cs = MOD.db.profile.Conditions[MOD.myClass]
	local c = cs[name]
	if IsOn(c) and c.name then -- name already exists
		return "DUPLICATE NAME\nThere is already a condition named " .. name .. ". Do you want to overwrite it?"
	end
	return false
end

-- Purge an existing condition, have to delete from bar groups, notifications, dependencies
local function PurgeCondition()
	local con = GetCondition()
	local oldName = con.name
	for _, bp in pairs(MOD.db.profile.BarGroups) do -- check bar group conditions
		if IsOn(bp) then
			if bp.checkCondition and (bp.condition == oldName) then bp.condition = nil; bp.checkCondition = false end
			for _, bar in pairs(bp.bars) do -- check for notifications
				if (bar.barType == "Notification") and (bar.action == oldName) then bar.enableBar = false end
			end
		end
	end
	for _, c in pairs(MOD.db.profile.Conditions[MOD.myClass]) do -- check dependencies
		if IsOn(c) and c.name and c.dependencies then
			local result = c.dependencies[oldName]
			if result ~= nil then
				c.dependencies[oldName] = nil
			end
		end
	end
end

-- Rename an existing condition, have to change for bar groups, bars, dependencies
local function RenameCondition(newName)
	local con = GetCondition()
	local oldName = con.name
	DeleteNameEntry("Conditions", MOD.db.profile.Conditions[MOD.myClass])
	con.name = newName
	AddNameEntry("Conditions", MOD.db.profile.Conditions[MOD.myClass], con)
	for _, bp in pairs(MOD.db.profile.BarGroups) do -- check bar group and bar conditions
		if IsOn(bp) then
			if bp.checkCondition and (bp.condition == oldName) then bp.condition = newName end
			if not bp.auto then -- have to scan bars in custom bar groups
				for _, bar in pairs(bp.bars) do
					if (bar.barType == "Notification") and (bar.action == oldName) then bar.action = newName end
					if bar.hideCondition == oldName then bar.hideCondition = newName end
					if bar.flashCondition == oldName then bar.flashCondition = newName end
					if bar.fadeCondition == oldName then bar.fadeCondition = newName end
				end
			end
		end
	end
	for _, c in pairs(MOD.db.profile.Conditions[MOD.myClass]) do -- check dependencies
		if IsOn(c) and c.name and c.dependencies then
			local result = c.dependencies[oldName]
			if result ~= nil then
				c.dependencies[oldName] = nil
				c.dependencies[newName] = result
			end
		end
	end
end

-- Return the selected condition associated with a bar group
local function GetBarGroupSelectedCondition(list)
	local bp = GetBarGroupEntry()
	if bp and bp.checkCondition and bp.condition then
		for j, c in pairs(list) do
			if c == bp.condition then return j end
		end
	end
	return nil
end

-- Return the index of the condition in the list
local function GetBarSelectedCondition(list, condition)
	for j, c in pairs(list) do
		if c == condition then return j end
	end
	return nil
end

-- Get a list of dependencies to select from
local function GetDependenciesList()
	local con = GetCondition()
	local i, t = 0, {}
	local nt = MOD.db.profile.Conditions[MOD.myClass]
	for _, n in pairs(nt) do
		if IsOn(n) and n.name and (n.name ~= con.name) then
			i = i + 1
			t[i] = n.name
		end
	end
	table.sort(t)
	return t
end

-- Set a dependency with the specified value in the current condition's dependencies list
local function SetDependency(dep, value)
	local con = GetCondition()
	if not con.dependencies then con.dependencies = {} end
	con.dependencies[dep] = value
end

-- Check if the dependency is included in the current condition's dependencies list
local function CheckDependency(dep)
	local con = GetCondition()
	if con.dependencies then
		for d in pairs(con.dependencies) do
			if dep == d then return true end
		end
	end
	return false
end

-- Return the selected dependency, clear it when switching conditions
local function GetSelectedDependency()
	local con = GetCondition()
	if con ~= selectedCondition then
		selectedCondition = con
		selectedDependency = nil
	end
	return selectedDependency
end

-- Set a field in the current bar group
local function SetBarGroupField(fname, value)
	local n = GetBarGroupEntry()
	if n then
		n[fname] = value
		if fname == "linkSettings" then MOD:InitializeBarGroupSettings(n) end -- special case for linked layouts
		if (fname == "pointX") or (fname == "pointY") then MOD:SetBarGroupPosition(n) end -- special case for display position
		if fname == "filterBuffLink" then MOD:InitializeFilterList(n, "Buff") end -- special case for linked buff filters
		if fname == "filterDebuffLink" then MOD:InitializeFilterList(n, "Debuff") end -- special case for linked debuff filters
		if fname == "filterCooldownLink" then MOD:InitializeFilterList(n, "Cooldown") end -- special case for linked cooldown filters
		MOD:UpdateAllBarGroups() -- any field changes need to be reflected in the actual bars on the display
	end
end

-- Get a field in the current bar group
local function GetBarGroupField(fname)
	local n = GetBarGroupEntry()
	if n then return n[fname] end
	return nil
end

-- Return the current merge bar group for the selected bar group
local function GetMergeBarGroup()
	local bg = GetBarGroupEntry()
	if bg then
		local bglist = GetBarGroupList()
		for i, bgname in pairs(bglist) do if bgname == bg.mergeInto then return i end end
	end
	return nil
end

-- Set the merge bar group for the selected bar group
local function SetMergeBarGroup(value)
	local bg = GetBarGroupEntry()
	if bg then
		local bglist = GetBarGroupList()
		local mergeInto = bglist[value]
		local mbg = MOD.db.profile.BarGroups[mergeInto]
		if IsOn(mbg) and mbg.enabled and not mbg.merged then
			bg.mergeInto = mergeInto		
			MOD:UpdateAllBarGroups() -- any changes need to be reflected in the actual bars on the display
		end
	end
end

-- Return the current anchor bar group for the selected bar group
local function GetBarGroupAnchor()
	local bg = GetBarGroupEntry()
	if bg then
		local anchor = bg.name
		if bg.anchor and (bg.anchor ~= bg.name) then anchor = bg.anchor end
		local bglist = GetBarGroupList()
		for i, bgname in pairs(bglist) do if bgname == anchor then return i end end
	end
	return nil
end

-- Set the current anchor bar group for the selected bar group
local function SetBarGroupAnchor(value)
	local bg = GetBarGroupEntry()
	if bg then
		if not value then
			bg.anchor = false
			bg.anchorX = 0
			bg.anchorY = 0
			bg.anchorLastBar = false
			bg.locked = false
		else
			local bglist = GetBarGroupList()
			local anchor = bglist[value]
			if anchor ~= bg.name then -- need to make sure no loops being set up
				local i, a = 0, anchor
				while i < 1000 do -- safety net, should never get this high
					i = i + 1
					local bp = MOD.db.profile.BarGroups[a]
					if not bp.anchor then break end -- found end of chain without looping back
					a = bp.anchor
					if a == bg.name then -- found a loop so print a warning and return
						print(string.format("Raven: \"%s\" cannot be attached to \"%s\" because that would create a circular dependency", a, anchor))
						return
					end
				end
				bg.anchor = anchor
			else bg.anchor = false end
		end
		MOD:UpdateAllBarGroups() -- any changes need to be reflected in the actual bars on the display
	end
end

-- Return the filter list of a given type for a bar group
local function GetBarGroupFilter(actionType)
	local n = GetBarGroupEntry()
	if n then
		local listName, selectName = "filter" .. actionType .. "List", "filter" .. actionType .. "Selection"
		local filterList = n[listName]
		if filterList then 
			table.sort(filterList)
			if not n[selectName] then n[selectName] = next(filterList, nil) end
			return filterList
		end
		n[selectName] = nil
	end
	return {}
end

-- Return selection within the filter list of a given type for a bar group
local function GetBarGroupFilterSelection(actionType)
	local n = GetBarGroupEntry()
	if n then
		local listName, selectName = "filter" .. actionType .. "List", "filter" .. actionType .. "Selection"
		local filterList = n[listName]
		local filterSelect = n[selectName]
		if filterList then 
			table.sort(filterList)
			if not filterSelect or not filterList[filterSelect] then n[selectName] = next(filterList, nil) end
			return n[selectName]
		end
	end
	return nil
end

-- Reset the filter list of a given type for a bar group
local function ResetBarGroupFilter(actionType)
	local n = GetBarGroupEntry()
	if n then
		local listName, selectName = "filter" .. actionType .. "List", "filter" .. actionType .. "Selection"
		n[listName] = nil
		n[selectName] = nil
	end
end

-- Add an action to be filtered by a bar group
local function AddBarGroupFilter(actionType, actionName)
	local n = GetBarGroupEntry()
	if n then
		local listName, selectName = "filter" .. actionType .. "List", "filter" .. actionType .. "Selection"
		local filterList = n[listName]
		if not filterList then filterList = {}; n[listName] = filterList end
		filterList[actionName] = actionName
		n[selectName] = actionName
		MOD:UpdateAllBarGroups() -- any field changes need to be reflected in the actual bars on the display
	end
end

-- Delete an action from a bar group filter list
local function DeleteBarGroupFilter(actionType, actionName)
	local n = GetBarGroupEntry()
	if n then
		local listName, selectName = "filter" .. actionType .. "List", "filter" .. actionType .. "Selection"
		local filterList = n[listName]
		if filterList and actionName then
			filterList[actionName] = nil
			n[selectName] = nil
		end
		MOD:UpdateAllBarGroups() -- any field changes need to be reflected in the actual bars on the display
	end
end

-- Set a field in the current notification
local function SetConditionField(fname, value)
	local n = GetCondition()
	if n then n[fname] = value end
	MOD:UpdateAllBarGroups() -- any field changes need to be reflected in the actual bars on the display
end

-- Get a field in the current notification
local function GetConditionField(fname)
	local n = GetCondition()
	if n then return n[fname] end
	return nil
end

-- Return a text description of the current condition
local function GetConditionDescription()
	local c = GetCondition()
	if c then return MOD:GetConditionText(c.name) end
	return ""
end

-- Get current setting for a test field in the currently selected condition
local function GetTestField(ttype, fname)
	local c = GetCondition()
	if c and c.tests then
		local t = c.tests[ttype] -- find the right test, if it exists
		if t then return t[fname] end -- found the field
	end
	return nil
end

-- Set value of a test field in the current condition, initializing fields as necessary
local function SetTestField(ttype, fname, value)
	local c = GetCondition()
	if c then
		if not c.tests then
			c.tests = {}
		else
			local t = c.tests[ttype] -- find the right test, if it has already been defined
			if t then -- found the field
				t[fname] = value
				MOD:UpdateAllBarGroups()
				return
			end
		end
		local newTest = {}
		for f, v in pairs(MOD.conditionTests[ttype]) do newTest[f] = v end -- initialize test parameters
		newTest[fname] = value -- after initializing the test can set the field
		c.tests[ttype] = newTest
		MOD:UpdateAllBarGroups()
	end
end

-- Check if an option field in the current test is turned on
local function IsTestFieldOn(ttype, fname)
	local value = GetTestField(ttype, fname)
	return (value ~= nil) and (value ~= Off)
end

-- Check if an option field in the current test is turned off
local function IsTestFieldOff(ttype, fname)
	local value = GetTestField(ttype, fname)
	return value == nil or value == Off
end

-- Set a field associated with a test, strip off white space at start and end of the string
local function SetTestFieldString(ttype, fname, s)
	local whiteStart, whiteEnd = string.find(s, "%s*", 1) -- skip white space
	if whiteStart == 1 then s = string.sub(s, whiteEnd + 1) end	
	s = string.reverse(s) -- reverse so can check for end white space
	whiteStart, whiteEnd = string.find(s, "%s*", 1) -- skip white space
	if whiteStart == 1 then s = string.sub(s, whiteEnd + 1) end
	local str = string.reverse(s) -- get the string after white space, reverse again to restore original order
	if str ~= "" then -- make sure not empty string
		if fname ~= "item" then ValidateSpellName(str) end
		SetTestField(ttype, fname, str)
	else
		SetTestField(ttype, fname, nil)
	end
end

-- Get a comma-separated list of the comma-seperated strings associated with a test
local function GetTestFieldSpellList(ttype, fname)
	local a, d = "", ""
	local strs = GetTestField(ttype, fname)
	if strs then
		for _, str in pairs(strs) do
			a = a .. d .. str -- append each to the comma-seperated string being built
			d = ", "
		end
	end
	return a
end

-- Set the auras associated with a test from a string with comma-separated names
local function SetTestFieldSpellList(ttype, fname, st)
	local start, strlist, s = 1, {}, st .. ',' -- initialize array, add ending comma
	repeat
		local whiteStart, whiteEnd = string.find(s, "%s*", start) -- skip white space
		if whiteStart == start then start = whiteEnd + 1 end
		if start > string.len(s) then break end
		local comma, nexts = string.find(s, '%s*,', start) -- find next comma, skipping white space preceding it
		if (comma - start) > 1 then -- make sure there is a string left after stripping white space
			table.insert(strlist, string.sub(s, start, comma - 1)) -- store the string in the table of aura names
		end
		start = nexts + 1
	until start > string.len(s)
	for _, k in pairs(strlist) do ValidateSpellName(k) end -- validates spell names
	SetTestField(ttype, fname, strlist)
end

-- Set a standard color
local function SetStandardColor(i, r, g, b, a)
	local c = MOD.ColorPalette[i]
	if c then c.r = r; c.g = g; c.b = b; c.a = a end
end

-- Get a standard color
local function GetStandardColor(i)
	local c = MOD.ColorPalette[i]
	if c then return c.r, c.g, c.b, c.a end
	return 0, 0, 0, 0
end

-- Generate text array using the standard colors
local function GetStandardColorList()
	local t = { }
	for name, c in pairs(MOD.ColorPalette) do
		t[name] = string.format("|cFF%02x%02x%02x%s", c.r*255, c.g*255, c.b*255, name)
	end
	return t
end

-- Add class, group and racial buffs or debuffs from the known auras table to a bar group
local function AddAurasToBarGroup(bg, buff, monitor, castBy)
	if buff then
		bg.detectBuffs = true
		bg.detectBuffsMonitor = monitor
		bg.detectBuffsCastBy = castBy
	else
		bg.detectDebuffs = true
		bg.detectDebuffsMonitor = monitor
		bg.detectDebuffsCastBy = castBy
	end
end

-- Add cooldowns from the known cooldowns table to a bar group
local function AddCooldownsToBarGroup(bg)
	bg.detectCooldowns = true
end

-- Add notifications for conditions to a bar group
local function AddNotificationsToBarGroup(bg)
	-- only check conditions for the player's class
	local ct = MOD.db.profile.Conditions[MOD.myClass]
	if ct then
		for _, c in pairs(ct) do
			if IsOn(c) and c.name and c.notify then
				local bname = c.name
				local bar = { action = bname, enableBar = true, barLabel = bname, barType = "Notification", barSource = nil,
						monitor = nil, castBy = nil, }
				AddBarToGroup(bg, bar)
			end
		end
	end
end

-- Configure Raven with the selected standard bar groups
local function ConfigureBarGroups()
	local offsetY = 0
	if standardNotifications then
		local bg = CreateBarGroup("Notifications", false, true, 0, offsetY)
		AddNotificationsToBarGroup(bg)
		offsetY = offsetY + 20
		standardNotifications = false
	end
	if standardRunes then
		local bg = CreateBarGroup("Runes", true, true, 0, offsetY)
		bg.showNoDuration = true
		bg.detectRuneCooldowns = true
		offsetY = offsetY + 20
		standardRunes = false
	end
	if standardFocus then
		local bg = CreateBarGroup("Focus", true, true, 0, offsetY)
		AddAurasToBarGroup(bg, true, "focus", "Player")
		AddAurasToBarGroup(bg, false, "focus", "Player")
		bg.showNoDuration = false
		offsetY = offsetY + 20
		standardFocus = false
	end
	if standardTarget then
		local bg = CreateBarGroup("Target", true, true, 0, offsetY)
		AddAurasToBarGroup(bg, true, "target", "Player")
		AddAurasToBarGroup(bg, false, "target", "Player")
		bg.showNoDuration = false
		offsetY = offsetY + 20
		standardTarget = false
	end
	if standardCooldowns then
		local bg = CreateBarGroup("Cooldowns", true, true, 0, offsetY)
		bg.checkDuration = true
		bg.minimumDuration = true
		bg.filterDuration = 2
		AddCooldownsToBarGroup(bg)
		offsetY = offsetY + 20
		standardCooldowns = false
	end
	if standardPlayerDebuffs then
		local bg = CreateBarGroup("Debuffs", true, true, 0, offsetY)
		AddAurasToBarGroup(bg, false, "player", "Anyone")
		bg.showNoDuration = false
		offsetY = offsetY + 20
		standardPlayerDebuffs = false
	end
	if standardPlayerBuffs then
		local bg = CreateBarGroup("Short Buffs", true, true, 0, offsetY)
		bg.checkDuration = true
		bg.minimumDuration = false
		bg.filterDuration = 120
		bg.showNoDuration = false
		AddAurasToBarGroup(bg, true, "player", "Anyone")
		offsetY = offsetY + 20
		standardPlayerBuffs = false
	end
	if standardLongBuffs then -- minimum duration of two minutes
		local bg = CreateBarGroup("Long Buffs", true, true, 0, offsetY)
		bg.checkDuration = true
		bg.minimumDuration = true
		bg.filterDuration = 120
		bg.detectTracking = true
		bg.showNoDuration = true
		bg.barColors = "Class"
		bg.sor = "C"
		AddAurasToBarGroup(bg, true, "player", "Anyone")
		offsetY = offsetY + 20
		standardLongBuffs = false
	end
end

-- Prepare a table of time format options showing examples to choose from with select dropdown menu
local function GetTimeFormatList(s, c)
	local i, menu = 1, {}
	while i <= #Nest_TimeFormatOptions do
		local f = Nest_FormatTime
		local t1, t2, t3, t4, t5 = f(8125.8, i, s, c), f(343.8, i, s, c), f(75.3, i, s, c), f(42.7, i, s, c), f(3.6, i, s, c)
		menu[i] = t1 .. ", " .. t2 .. ", " .. t3 .. ", " .. t4 .. ", " .. t5
		i = i + 1
	end
	return menu
end

-- Create a mini-options table to be inserted at top level in the Bliz interface
MOD.TopOptionsTable = {
	type = "group", order = 1,
	args = {
		intro = {
			type = "description", order = 1,
			name = "This addon lets you monitor buffs and debuffs for player, target and focus. " ..
			"Colorful highlights can be added to buttons to show status. " ..
			"Monitored buffs and debuffs can trigger notifications.\n",
		},
	},
}

-- Create the options table to be used by the configuration GUI
MOD.OptionsTable = {
	type = "group", childGroups = "tab",
	args = {
		FrontPage = {
			type = "group", order = 10, name = "Setup",
			disabled = function(info) return InMode() end,
			args = {
				EnableGroup = {
					type = "group", order = 1, name = "Enable", inline = true,
					args = {
						EnableGroup = {
							type = "toggle", order = 10, name = "Enable Raven",
							desc = "If checked, Raven is enabled, otherwise all features are disabled.",
							get = function(info) return MOD.db.profile.enabled end,
							set = function(info, value) MOD.db.profile.enabled = value end,
						},
						HideBlizzBuffsGroup = {
							type = "toggle", order = 20, name = "Hide Blizzard Buffs",
							desc = "If checked, Raven will hide the default user interface for buffs.",
							get = function(info) return MOD.db.profile.hideBlizz end,
							set = function(info, value) MOD.db.profile.hideBlizz = value end,
						},
					},
				},
				StandardGroups = {
					type = "group", order = 10, name = "Standard Bar Groups", inline = true,
					args = {
						Anchors = {
							type = "description", order = 1,
							name = "Raven lets you display bars for buffs, debuffs and cooldowns. " ..
							'You can configure bars using the many options under the "Bar Groups" tab to create a personalized user interface. ' ..
							"To get started, however, it may be easier to begin with the standard bar groups listed here. " ..
							'Select the ones you want and then click the "Configure" button.\n'
						},
						Buffs = {
							type = "toggle", order = 10, name = "Short Buffs",
							get = function(info) return standardPlayerBuffs end,
							set = function(info, value) standardPlayerBuffs = value end,
						},
						spacer0 = { type = "description", name = "Buffs on you lasting less than 2 minutes.", order = 11, width = "double" },
						spacer0A = { type = "description", name = "", order = 12 },
						LongBuffs = {
							type = "toggle", order = 20, name = "Long Buffs",
							get = function(info) return standardLongBuffs end,
							set = function(info, value) standardLongBuffs = value end,
						},
						spacer1 = { type = "description", name = "Buffs on you lasting at least 2 minutes, color and sort by class.", order = 21, width = "double" },
						spacer1A = { type = "description", name = "", order = 22 },
						Debuffs = {
							type = "toggle", order = 25, name = "Debuffs",
							get = function(info) return standardPlayerDebuffs end,
							set = function(info, value) standardPlayerDebuffs = value end,
						},
						spacer2 = { type = "description", name = "Debuffs on you.", order = 26, width = "double" },
						spacer2A = { type = "description", name = "", order = 27 },
						Cooldowns = {
							type = "toggle", order = 30, name = "Cooldowns",
							get = function(info) return standardCooldowns end,
							set = function(info, value) standardCooldowns = value end,
						},
						spacer3 = { type = "description", name = "Cooldowns on your spells and trinkets lasting at least 2 seconds.", order = 31, width = "double" },
						spacer3A = { type = "description", name = "", order = 32 },
						Target = {
							type = "toggle", order = 35, name = "Target",
							get = function(info) return standardTarget end,
							set = function(info, value) standardTarget = value end,
						},
						spacer4 = { type = "description", name = "Buffs and debuffs you cast on the target.", order = 36, width = "double" },
						spacer4A = { type = "description", name = "", order = 37 },
						Focus = {
							type = "toggle", order = 40, name = "Focus",
							get = function(info) return standardFocus end,
							set = function(info, value) standardFocus = value end,
						},
						spacer5 = { type = "description", name = "Buffs and debuffs you cast on the focus.", order = 41, width = "double" },
						spacer5A = { type = "description", name = "", order = 42,
							hidden = function(info) return MOD.myClass ~= "DEATHKNIGHT" end,
						},
						Runes = {
							type = "toggle", order = 45, name = "Runes",
							hidden = function(info) return MOD.myClass ~= "DEATHKNIGHT" end,
							get = function(info) return standardRunes end,
							set = function(info, value) standardRunes = value end,
						},
						spacer6 = { type = "description", name = "Rune cooldown bars for Death Knights.", order = 46, width = "double",
							hidden = function(info) return MOD.myClass ~= "DEATHKNIGHT" end,
						},
						spacer6A = { type = "description", name = "", order = 47 },
						Notifications = {
							type = "toggle", order = 50, name = "Notifications",
							get = function(info) return standardNotifications end,
							set = function(info, value) standardNotifications = value end,
						},
						spacer7 = { type = "description", name = "Common class-specific notifications.", order = 51, width = "double" },
						spacer7A = { type = "description", name = "", order = 52 },
						Configure = {
							type = "execute", order = 90, name = "Configure",
							desc = "Configure Raven with the selected standard bar groups.",
							func = function(info) ConfigureBarGroups() end,
							confirm = function(info) if CheckBarGroupsExist() then return "CONFIGURE\nAre you sure you want to set up the " ..
								"selected standard bar groups, overwriting existing bar groups with the same names?" end end,
						},
					},
				},
				AnchorGroup = {
					type = "group", order = 50, name = "Bar Group Anchors", inline = true,
					args = {
						Anchors = {
							type = "description", order = 10,
							name = "Bar groups have anchors that show where bars will be displayed. " ..
									"You can left-click and drag an anchor to an appropriate location. " ..
									"You can right-click on an anchor to lock it when you are satisfied with its position. " ..
									'Locked anchors are hidden--use the "Unlock All Anchors" button to show them again.\n'
						},
						LockBars = {
							type = "execute", order = 11, name = "Lock All Anchors", 
							desc = "Lock and hide the anchors for all bar groups.",
							func = function(info) MOD:LockBarGroups(true) end,
						},
						UnlockBars = {
							type = "execute", order = 12, name = "Unlock All Anchors",
							desc = "Unlock and show the anchors for all bar groups.",
							func = function(info) MOD:LockBarGroups(false) end,
						},
					},
				},
			},
		},
		DefaultsPage = {
			type = "group", order = 15, name = "Defaults",
			disabled = function(info) return InMode() end,
			args = {
				Defaults = {
					type = "description", order = 10,
					name = "Bars in a bar group share appearance options such as dimensions and fonts/textures. " ..
							"You can adjust defaults for these options with settings in this section. " ..
							"You can also change colors assigned to particular spells in this section. " ..
							"These settings are shared across all profiles to simplify use with multiple characters.\n"
				},
				DimensionGroup = {
					type = "group", order = 10, name = "Dimensions", inline = true,
					args = {
						BarWidth = {
							type = "range", order = 10, name = "Bar Width", min = 5, max = 500, step = 1,
							desc = "Set width of bars.",
							get = function(info) return MOD.db.global.Defaults.barWidth end,
							set = function(info, value) MOD.db.global.Defaults.barWidth = value; MOD:UpdateAllBarGroups() end,
						},
						BarHeight = {
							type = "range", order = 15, name = "Bar Height", min = 5, max = 100, step = 1,
							desc = "Set height of bars.",
							get = function(info) return MOD.db.global.Defaults.barHeight end,
							set = function(info, value) MOD.db.global.Defaults.barHeight = value; MOD:UpdateAllBarGroups() end,
						},
						IconSize = {
							type = "range", order = 20, name = "Icon Size", min = 5, max = 100, step = 1,
							desc = "Set width/height for icons.",
							get = function(info) return MOD.db.global.Defaults.iconSize end,
							set = function(info, value) MOD.db.global.Defaults.iconSize = value; MOD:UpdateAllBarGroups() end,
						},
						BarScale = {
							type = "range", order = 25, name = "Scale", min = 0.1, max = 2, step = 0.05,
							desc = "Set scale factor for bars and icons.",
							get = function(info) return MOD.db.global.Defaults.scale end,
							set = function(info, value) MOD.db.global.Defaults.scale = value; MOD:UpdateAllBarGroups() end,
						},
						Space1 = { type = "description", name = "", order = 40 },
						HorizontalSpacing = {
							type = "range", order = 60, name = "Horizontal Spacing", min = 0, max = 100, step = 1,
							desc = "Adjust horizontal spacing between bars.",
							get = function(info) return MOD.db.global.Defaults.spacingX end,
							set = function(info, value) MOD.db.global.Defaults.spacingX = value; MOD:UpdateAllBarGroups() end,
						},
						VerticalSpacing = {
							type = "range", order = 65, name = "Vertical Spacing", min = 0, max = 100, step = 1,
							desc = "Adjust vertical spacing between bars.",
							get = function(info) return MOD.db.global.Defaults.spacingY end,
							set = function(info, value) MOD.db.global.Defaults.spacingY = value; MOD:UpdateAllBarGroups() end,
						},
						IconOffsetX = {
							type = "range", order = 70, name = "Icon Inset", min = -100, max = 100, step = 1,
							desc = "Set icon's horizontal inset from bar.",
							get = function(info) return MOD.db.global.Defaults.iconOffsetX end,
							set = function(info, value) MOD.db.global.Defaults.iconOffsetX = value; MOD:UpdateAllBarGroups() end,
						},
						IconOffsetY = {
							type = "range", order = 75, name = "Icon Offset", min = -100, max = 100, step = 1,
							desc = "Set vertical offset between icon and bar.",
							get = function(info) return MOD.db.global.Defaults.iconOffsetY end,
							set = function(info, value) MOD.db.global.Defaults.iconOffsetY = value; MOD:UpdateAllBarGroups() end,
						},
						Space2 = { type = "description", name = "", order = 80 },
						ResetDimensions = {
							type = "execute", order = 90, name = "Reset Dimensions",
							desc = "Reset dimensions to default settings.",
							confirm = function(info) return 'RESET DIMENSIONS\nAre you sure you want to reset the dimension options?' end,
							func = function(info) MOD:SetDimensionDefaults(MOD.db.global.Defaults); MOD:UpdateAllBarGroups() end,
						},
					},
				},
				FontsGroup = {
					type = "group", order = 20, name = "Fonts and Textures", inline = true,
					args = {
						LabelFont = {
							type = "select", order = 21, name = " Label Font", 
							desc = "Select font for label text.",
							dialogControl = 'LSM30_Font',
							values = AceGUIWidgetLSMlists.font,
							get = function(info) return MOD.db.global.Defaults.labelFont end,
							set = function(info, value) MOD.db.global.Defaults.labelFont = value; MOD:UpdateAllBarGroups() end,
						},
						LabelFontSize = {
							type = "range", order = 23, name = "Font Size", min = 5, max = 30, step = 1,
							desc = "Set font size for labels.",
							get = function(info) return MOD.db.global.Defaults.labelFSize end,
							set = function(info, value) MOD.db.global.Defaults.labelFSize = value; MOD:UpdateAllBarGroups() end,
						},
						LabelAlpha = {
							type = "range", order = 25, name = "Opacity", min = 0, max = 1, step = 0.05,
							desc = "Set opacity for labels.",
							get = function(info) return MOD.db.global.Defaults.labelAlpha end,
							set = function(info, value) MOD.db.global.Defaults.labelAlpha = value; MOD:UpdateAllBarGroups() end,
						},
						LabelColor = {
							type = "color", order = 27, name = "Color", hasAlpha = false, width = "half",
							get = function(info)
								local t = MOD.db.global.Defaults.labelColor
								if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.Defaults.labelColor
								if not t then MOD.db.global.Defaults.labelColor = { r = r, g = g, b = b, a = a }
									else t.r = r; t.g = g; t.b = b; t.a = a end
								MOD:UpdateAllBarGroups()
							end,
						},
						LabelOutline = {
							type = "toggle", order = 29, name = "Outline", width = "half",
							desc = "Add black outline to label.",
							get = function(info) return MOD.db.global.Defaults.labelOutline end,
							set = function(info, value) MOD.db.global.Defaults.labelOutline = value; MOD:UpdateAllBarGroups() end,
						},
						Space2 = { type = "description", name = "", order = 30 },
						TimeFont = {
							type = "select", order = 31, name = " Time Font", 
							desc = "Select font for time text.",
							dialogControl = 'LSM30_Font',
							values = AceGUIWidgetLSMlists.font,
							get = function(info) return MOD.db.global.Defaults.timeFont end,
							set = function(info, value) MOD.db.global.Defaults.timeFont = value; MOD:UpdateAllBarGroups() end,
						},
						TimeFontSize = {
							type = "range", order = 33, name = "Font Size", min = 5, max = 30, step = 1,
							desc = "Set font size for time text.",
							get = function(info) return MOD.db.global.Defaults.timeFSize end,
							set = function(info, value) MOD.db.global.Defaults.timeFSize = value; MOD:UpdateAllBarGroups() end,
						},
						TimeAlpha = {
							type = "range", order = 35, name = "Opacity", min = 0, max = 1, step = 0.05,
							desc = "Set opacity for time text.",
							get = function(info) return MOD.db.global.Defaults.timeAlpha end,
							set = function(info, value) MOD.db.global.Defaults.timeAlpha = value; MOD:UpdateAllBarGroups() end,
						},
						TimeColor = {
							type = "color", order = 37, name = "Color", hasAlpha = false, width = "half",
							get = function(info)
								local t = MOD.db.global.Defaults.timeColor
								if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.Defaults.timeColor
								if not t then MOD.db.global.Defaults.timeColor = { r = r, g = g, b = b, a = a }
									else t.r = r; t.g = g; t.b = b; t.a = a end
								MOD:UpdateAllBarGroups()
							end,
						},
						TimeOutline = {
							type = "toggle", order = 39, name = "Outline", width = "half",
							desc = "Add black outline to time text.",
							get = function(info) return MOD.db.global.Defaults.timeOutline end,
							set = function(info, value) MOD.db.global.Defaults.timeOutline = value; MOD:UpdateAllBarGroups() end,
						},
						Space3 = { type = "description", name = "", order = 40 },
						IconFont = {
							type = "select", order = 41, name = " Icon Font", 
							desc = "Select font for icon text (i.e., spell stack count).",
							dialogControl = 'LSM30_Font',
							values = AceGUIWidgetLSMlists.font,
							get = function(info) return MOD.db.global.Defaults.iconFont end,
							set = function(info, value) MOD.db.global.Defaults.iconFont = value; MOD:UpdateAllBarGroups() end,
						},
						IconFontSize = {
							type = "range", order = 43, name = "Font Size", min = 5, max = 30, step = 1,
							desc = "Set font size for icon text.",
							get = function(info) return MOD.db.global.Defaults.iconFSize end,
							set = function(info, value) MOD.db.global.Defaults.iconFSize = value; MOD:UpdateAllBarGroups() end,
						},
						IconAlpha = {
							type = "range", order = 45, name = "Opacity", min = 0, max = 1, step = 0.05,
							desc = "Set opacity for icon text.",
							get = function(info) return MOD.db.global.Defaults.iconAlpha end,
							set = function(info, value) MOD.db.global.Defaults.iconAlpha = value; MOD:UpdateAllBarGroups() end,
						},
						IconColor = {
							type = "color", order = 47, name = "Color", hasAlpha = false, width = "half",
							get = function(info)
								local t = MOD.db.global.Defaults.iconColor
								if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.Defaults.iconColor
								if not t then MOD.db.global.Defaults.iconColor = { r = r, g = g, b = b, a = a }
									else t.r = r; t.g = g; t.b = b; t.a = a end
								MOD:UpdateAllBarGroups()
							end,
						},
						IconOutline = {
							type = "toggle", order = 49, name = "Outline", width = "half",
							desc = "Add black outline to icon text.",
							get = function(info) return MOD.db.global.Defaults.iconOutline end,
							set = function(info, value) MOD.db.global.Defaults.iconOutline = value; MOD:UpdateAllBarGroups() end,
						},
						Space7 = { type = "description", name = "", order = 70 },
						BarTexture = {
							type = "select", order = 72, name = " Bar Texture", 
							desc = "Select texture for bars.",
							dialogControl = 'LSM30_Statusbar',
							values = AceGUIWidgetLSMlists.statusbar,
							get = function(info) return MOD.db.global.Defaults.texture end,
							set = function(info, value) MOD.db.global.Defaults.texture = value; MOD:UpdateAllBarGroups() end,
						},
						BarAlpha = {
							type = "range", order = 74, name = "Bar Opacity", min = 0, max = 1, step = 0.05,
							desc = "Set alpha for bars.",
							get = function(info) return MOD.db.global.Defaults.alpha end,
							set = function(info, value) MOD.db.global.Defaults.alpha = value; MOD:UpdateAllBarGroups() end,
						},
						BackgroundTexture = {
							type = "select", order = 76, name = " Background Texture", 
							desc = "Select background texture for bars.",
							dialogControl = 'LSM30_Statusbar',
							values = AceGUIWidgetLSMlists.statusbar,
							get = function(info) return MOD.db.global.Defaults.bgtexture end,
							set = function(info, value) MOD.db.global.Defaults.bgtexture = value; MOD:UpdateAllBarGroups() end,
						},
						Background = {
							type = "range", order = 78, name = "Background Opacity", min = 0, max = 1, step = 0.05,
							desc = "Set background opacity for bars.",
							get = function(info) return MOD.db.global.Defaults.bgAlpha end,
							set = function(info, value) MOD.db.global.Defaults.bgAlpha = value; MOD:UpdateAllBarGroups() end,
						},
						Space8 = { type = "description", name = "", order = 80 },
						ForegroundSaturation = {
							type = "range", order = 82, name = "Foreground Saturation", min = -1, max = 1, step = 0.05,
							desc = "Set saturation for foreground colors.",
							get = function(info) return MOD.db.global.Defaults.fgSaturation end,
							set = function(info, value) MOD.db.global.Defaults.fgSaturation = value; MOD:UpdateAllBarGroups() end,
						},
						ForegroundBrightness = {
							type = "range", order = 84, name = "Foreground Brightness", min = -1, max = 1, step = 0.05,
							desc = "Set brightness for foreground colors.",
							get = function(info) return MOD.db.global.Defaults.fgBrightness end,
							set = function(info, value) MOD.db.global.Defaults.fgBrightness = value; MOD:UpdateAllBarGroups() end,
						},
						BackgroundSaturation = {
							type = "range", order = 86, name = "Background Saturation", min = -1, max = 1, step = 0.05,
							desc = "Set saturation for foreground colors.",
							get = function(info) return MOD.db.global.Defaults.bgSaturation end,
							set = function(info, value) MOD.db.global.Defaults.bgSaturation = value; MOD:UpdateAllBarGroups() end,
						},
						BackgroundBrightness = {
							type = "range", order = 88, name = "Background Brightness", min = -1, max = 1, step = 0.05,
							desc = "Set brightness for foreground colors.",
							get = function(info) return MOD.db.global.Defaults.bgBrightness end,
							set = function(info, value) MOD.db.global.Defaults.bgBrightness = value; MOD:UpdateAllBarGroups() end,
						},
						Space9 = { type = "description", name = "", order = 90 },
						ResetFonts = {
							type = "execute", order = 90, name = "Reset Fonts/Textures",
							desc = "Reset fonts and textures to default settings.",
							confirm = function(info) return 'RESET FONTS/TEXTURES\nAre you sure you want to reset the font and texture options?' end,
							func = function(info) MOD:SetFontTextureDefaults(MOD.db.global.Defaults); MOD:UpdateAllBarGroups() end,
						},
					},
				},
				ColorsGroup = {
					type = "group", order = 30, name = "Standard Colors", inline = true,
					args = {
						ColorText = { type = "description", name = "Bar Colors:", order = 1, width = "half" },
						NotificationColor = {
							type = "color", order = 10, name = "Notify", hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultNotificationColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultNotificationColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						BuffColor = {
							type = "color", order = 11, name = "Buff", hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultBuffColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultBuffColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						DebuffColor = {
							type = "color", order = 12, name = "Debuff", hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultDebuffColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultDebuffColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						CooldownColor = {
							type = "color", order = 13, name = "Cooldown", hasAlpha = false,
							get = function(info) local t = MOD.db.global.DefaultCooldownColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultCooldownColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						Space0 = { type = "description", name = "", order = 14 },
						DebuffText = { type = "description", name = "Debuff Colors:", order = 15, width = "half" },
						PoisonColor = {
							type = "color", order = 15, name = "Poison", hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultPoisonColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultPoisonColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						CurseColor = {
							type = "color", order = 16, name = "Curse", hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultCurseColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultCurseColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						MagicColor = {
							type = "color", order = 17, name = "Magic", hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultMagicColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultMagicColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						DiseaseColor = {
							type = "color", order = 18, name = "Disease", hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultDiseaseColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultDiseaseColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						Space1 = { type = "description", name = "", order = 20 },
						SpellName = {
							type = "input", order = 30, name = " Spell Name",
							desc = "Enter spell name (or numeric identifier) whose color you want to change.",
							get = function(info) return selectedSpellName end,
							set = function(info, value)
								local n, t = value, tonumber(value)
								if t then n = GetSpellInfo(t) end
								ValidateSpellName(n)
								selectedSpellName = n
							end,
						},
						StandardColors = {
							type = "select", order = 40, name = "  Color", width = "half",
							desc = "Select a standard color or click to set a custom color.",
							disabled = function(info) return not selectedSpellName or (selectedSpellName == "") end,
							get = function(info) return nil end,
							set = function(info, value)
								if MOD:GetIcon(selectedSpellName) then -- make sure this is a valid spell before setting a color in save table
									local t = {}
									t.r, t.g, t.b, t.a = GetStandardColor(value)
									MOD:SetColor(selectedSpellName, t)
									MOD:UpdateAllBarGroups()
								end
							end,
							values = function(info) return GetStandardColorList() end,
							style = "dropdown",
						},
						SpellColor = {
							type = "color", order = 50, name = "", hasAlpha = false, width = "half",
							disabled = function(info) return not selectedSpellName or (selectedSpellName == "") end,
							get = function(info)
								local t = MOD:GetColor(selectedSpellName, nil)
								if not t then return nil end
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								if MOD:GetIcon(selectedSpellName) then -- make sure this is a valid spell before setting a color in save table
									local t = {}
									t.r = r; t.g = g; t.b = b; t.a = a
									MOD:SetColor(selectedSpellName, t)
									MOD:UpdateAllBarGroups()
								end
							end,
						},
						SpellIcon = {
							type = "description", order = 60, name = "", width = "half", 
							disabled = function(info) return not selectedSpellName or (selectedSpellName == "") end,
							image = function(info) return MOD:GetIcon(selectedSpellName) end,
							imageWidth = 24, imageHeight = 24,
						},
						ResetSpellColor = {
							type = "execute", order = 70, name = "Reset", width = "half",
							desc = function(info) return ("Reset color for " .. selectedSpellName .. " back to default.") end,
							disabled = function(info) return not selectedSpellName or (selectedSpellName == "") end,
							confirm = function(info) return ('RESET SPELL COLOR\nAre you sure you want to reset the color for ' .. selectedSpellName .. ' back to default?') end,
							func = function(info)
								if MOD:GetIcon(selectedSpellName) then -- make sure this is a valid spell name before changing the color
									MOD:ResetColorDefault(selectedSpellName)
									MOD:UpdateAllBarGroups()
								end
							end,
						},
						Space2 = { type = "description", name = "", order = 80 },
						ResetBarColors = {
							type = "execute", order = 85, name = "Reset Bar Colors",
							desc = "Reset bar colors back to default.",
							confirm = function(info) return ('RESET BAR COLORS\nAre you sure you want to reset bar colors back to default?') end,
							func = function(info)
								MOD.db.global.DefaultBuffColor = HexColorConvert("8ae234") -- Green1
								MOD.db.global.DefaultDebuffColor = HexColorConvert("fcaf3e") -- Orange1
								MOD.db.global.DefaultCooldownColor = HexColorConvert("fce94f") -- Yellow1
								MOD.db.global.DefaultNotificationColor = HexColorConvert("729fcf") -- Blue1								
							end,
						},
						ResetDebuffColors = {
							type = "execute", order = 86, name = "Reset Debuff Colors",
							desc = "Reset debuff colors back to default.",
							confirm = function(info) return ('RESET DEBUFF COLORS\nAre you sure you want to reset debuff colors back to default?') end,
							func = function(info)
								MOD.db.global.DefaultPoisonColor = CopyColor(DebuffTypeColor["Poison"])
								MOD.db.global.DefaultCurseColor = CopyColor(DebuffTypeColor["Curse"])
								MOD.db.global.DefaultMagicColor = CopyColor(DebuffTypeColor["Magic"])
								MOD.db.global.DefaultDiseaseColor = CopyColor(DebuffTypeColor["Disease"])						
							end,
						},
						ResetSpellColors = {
							type = "execute", order = 87, name = "Reset Spell Colors",
							desc = "Reset spell colors back to defaults.",
							confirm = function(info) return 'RESET SPELL COLORS\nAre you sure you want to reset all spell colors back to defaults?' end,
							func = function(info) MOD:ResetColorDefaults(); MOD:UpdateAllBarGroups() end,
						},
					},
				},
			},
		},
		BarGroups = {
			type = "group", order = 20, name = "Bar Groups", childGroups = "tab",
			disabled = function(info) return InMode("Not") end,
			args = {
				SelectBarGroup = {
					type = "select", order = 1, name = "  Bar Group",
					get = function(info) UpdateBarList(); return GetSelectedBarGroup() end,
					set = function(info, value) SetSelectedBarGroup(value) end,
					disabled = function(info) return NoBarGroup() or InMode("Bar") end,
					values = function(info) return GetBarGroupList() end,
					style = "dropdown",
				},
				Space1 = { type = "description", name = "", order = 2, width = "half" },
				NewBarGroupButton = {
					type = "execute", order = 3, name = "New Custom Group",
					desc = "Create a new bar group with manually added bars.",
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return enterNewBarGroup end,
					func = function(info) enterNewBarGroup, toggleNewBarGroup, typeNewBarGroup = true, true, false end,
				},
				NewAutoBarGroupButton = {
					type = "execute", order = 4, name = "New Auto Group",
					desc = "Create a new bar group with automatically displayed bars.",
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return enterNewBarGroup end,
					func = function(info) enterNewBarGroup, toggleNewBarGroup, typeNewBarGroup = true, true, true end,
				},
				NewCustomBarGroupName = {
					type = "input", order = 5, name = " Enter Custom Group Name",
					desc = " Enter name of new custom bar group.",
					hidden = function(info) return not enterNewBarGroup or typeNewBarGroup end,
					validate = function(info, n) if not n or (n == "") then return "Invalid name." else return true end end,
					confirm = function(info, value) return ConfirmNewBarGroup(value) end,
					get = function(info)
						enterNewBarGroup = toggleNewBarGroup
						enterNewBarGroupType = false
						if toggleNewBarGroup then toggleNewBarGroup = false end
						if not enterNewBarGroup then MOD:UpdateOptions() end
						return false
					end,
					set = function(info, value) enterNewBarGroup = false; CreateBarGroup(value, false, false, 0, 0) end,
				},
				NewAutoBarGroupName = {
					type = "input", order = 6, name = " Enter Auto Group Name",
					desc = " Enter name of new auto bar group.",
					hidden = function(info) return not enterNewBarGroup or not typeNewBarGroup end,
					validate = function(info, n) if not n or (n == "") then return "Invalid name." else return true end end,
					confirm = function(info, value) return ConfirmNewBarGroup(value) end,
					get = function(info)
						enterNewBarGroup = toggleNewBarGroup
						enterNewBarGroupType = false
						if toggleNewBarGroup then toggleNewBarGroup = false end
						if not enterNewBarGroup then MOD:UpdateOptions() end
						return false
					end,
					set = function(info, value) enterNewBarGroup = false; CreateBarGroup(value, true, false, 0, 0) end,
				},
				CancelNewBarGroup = {
					type = "execute", order = 7, name = "Cancel", width = "half",
					desc = "Cancel creating a new bar group.",
					hidden = function(info) return not enterNewBarGroup end,
					func = function(info) enterNewBarGroup, toggleNewBarGroup = false, false end,
				},
				DeleteBarGroup = {
					type = "execute", order = 8, name = "Delete Bar Group",
					desc = "Delete the selected bar group.",
					disabled = function(info) return NoBarGroup() or InMode("Bar") end,
					hidden = function(info) return enterNewBarGroup end,
					func = function(info) DeleteBarGroup() end,
					confirm = function(info) return 'DELETE BAR GROUP\nAre you sure you want to delete the "' .. GetBarGroupField("name") .. '" bar group?' end,
				},
				GeneralTab = {
					type = "group", order = 10, name = "General",
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() end,
					args = {
						SettingsGroup = {
							type = "group", order = 1, name = "Settings", inline = true,
							args = {
								EnableBarGroup = {
									type = "toggle", order = 10, name = "Enable Bar Group",
									desc = "If checked, the bar group is enabled and its anchor and bars will be displayed, " ..
										"otherwise the bar group is disabled and neither anchor nor bars will be displayed.",
									get = function(info) return GetBarGroupField("enabled") end,
									set = function(info, value) SetBarGroupField("enabled", value) end,
								},
								EnableMerge = {
									type = "toggle", order = 20, name = "Merge Bar Group",
									desc = "If checked, merge bars from this bar group into another bar group. " ..
										"Bars are generated based on this bar group's custom or auto settings, including \"Show When\" and timer options, " ..
										"but are sorted and displayed using the other bar group's layout and appearance settings.",
									get = function(info) return GetBarGroupField("merged") end,
									set = function(info, value) SetBarGroupField("merged", value) end,
								},
								LinkSettings = {
									type = "toggle", order = 30, name = "Link Settings",
									desc = "If checked, the bar group's position, layout, appearance, and timer options are linked " ..
											"to bar groups in other profiles with the same name.",
									get = function(info) return GetBarGroupField("linkSettings") end,
									set = function(info, value) SetBarGroupField("linkSettings", value) end,
								},
								SaveSettings = {
									type = "execute", order = 40, name = "Save Settings",
									hidden = function(info) return GetBarGroupField("linkSettings") end,
									desc = "Click to save current settings to linked settings.",
									func = function(info) MOD:SaveBarGroupSettings(GetBarGroupEntry()) end,
								},
								spacer = { type = "description", name = "", order = 50, hidden = function(info) return not GetBarGroupField("merged") end },
								MergeBarGroup = {
									type = "select", order = 60, name = "  Bar Group To Merge Into",
									desc = "Select a bar group to merge into.",
									hidden = function(info) return not GetBarGroupField("merged") end,
									get = function(info) return GetMergeBarGroup() end,
									set = function(info, value) SetMergeBarGroup(value) end,
									values = function(info) return GetBarGroupList() end,
									style = "dropdown",
								},
							},
						},
						RenameGroup = {
							type = "group", order = 5, name = "Rename Bar Group", inline = true,
							args = {
								Rename = {
									type = "input", order = 10, name = " New Name",
									validate = function(info, n) if not n or (n == "") then return "Invalid name." else return true end end,
									confirm = function(info, value) return ConfirmNewBarGroup(value) end,
									desc = " Enter new name for the bar group.",
									get = function(info) return GetBarGroupField("name") end,
									set = function(info, value) RenameBarGroup(value) end,
								},
							},
						},
						SortingGroup = {
							type = "group", order = 10, name = "Sort Order", inline = true,
							hidden = function(info) return GetBarGroupField("merged") end,
							args = {
								AtoZOrder = {
									type = "toggle", order = 10, name = "A to Z", width = "half", 
									desc = "If checked, sort in ascending alphabetical order starting at bar closest to the anchor.",
									get = function(info) return GetBarGroupField("sor") == "A" end,
									set = function(info, value) SetBarGroupField("sor", "A") end,
								},
								TimeLeftOrder = {
									type = "toggle", order = 20, name = "Time Left", width = "half", 
									desc = "If checked, sort by time left in ascending order starting at bar closest to the anchor.",
									get = function(info) return GetBarGroupField("sor") == "T" end,
									set = function(info, value) SetBarGroupField("sor", "T") end,
								},
								DurationOrder = {
									type = "toggle", order = 30, name = "Duration", width = "half", 
									desc = "If checked, sort by overall duration in ascending order starting at bar closest to the anchor.",
									get = function(info) return GetBarGroupField("sor") == "D" end,
									set = function(info, value) SetBarGroupField("sor", "D") end,
								},
								ClassOrder = {
									type = "toggle", order = 40, name = "Class", width = "half", 
									desc = "If checked, sort by associated class name, if any, in ascending alphabetical order starting at bar closest to the anchor.",
									get = function(info) return GetBarGroupField("sor") == "C" end,
									set = function(info, value) SetBarGroupField("sor", "C") end,
								},
								CustomOrder = {
									type = "toggle", order = 50, name = "Custom", width = "half",
									desc = "If checked, allow manually setting the order of bars.",
									hidden = function(info) return GetBarGroupField("auto") end,
									get = function(info) return GetBarGroupField("sor") == "X" end,
									set = function(info, value) SetBarGroupField("sor", "X") end,
								},
								spacer = { type = "description", name = "", order = 60, width = "half" },
								ReverseSortOrder = {
									type = "toggle", order = 70, name = "Reverse Order", 
									desc = 'If checked, reverse the sort order (e.g., "A to Z" becomes "Z to A").',
									get = function(info) return GetBarGroupField("reverseSort") end,
									set = function(info, value) SetBarGroupField("reverseSort", value) end,
								},
							},
						},
						ShowWhenGroup = {
							type = "group", order = 20, name = "Show When", inline = true,
							args = {
								InCombatGroup = {
									type = "toggle", order = 20, name = "In Combat",
									desc = "If checked, bar group is shown when the player is in combat.",
									get = function(info) return GetBarGroupField("showCombat") end,
									set = function(info, value) SetBarGroupField("showCombat", value) end,
								},
								OutOfCombatGroup = {
									type = "toggle", order = 21, name = "Out Of Combat",
									desc = "If checked, bar group is shown when the player is out of combat.",
									get = function(info) return GetBarGroupField("showOOC") end,
									set = function(info, value) SetBarGroupField("showOOC", value) end,
								},
								RestingGroup = {
									type = "toggle", order = 22, name = "Resting",
									desc = "If checked, bar group is shown when the player is resting.",
									get = function(info) return GetBarGroupField("showResting") end,
									set = function(info, value) SetBarGroupField("showResting", value) end,
								},
								MountedGroup = {
									type = "toggle", order = 23, name = "Mounted",
									desc = "If checked, bar group is shown when the player is mounted.",
									get = function(info) return GetBarGroupField("showMounted") end,
									set = function(info, value) SetBarGroupField("showMounted", value) end,
								},
								SoloGroup = {
									type = "toggle", order = 30, name = "Solo",
									desc = "If checked, bar group is shown when the player is not in a party or raid.",
									get = function(info) return GetBarGroupField("showSolo") end,
									set = function(info, value) SetBarGroupField("showSolo", value) end,
								},
								PartyGroup = {
									type = "toggle", order = 31, name = "In Party",
									desc = "If checked, bar group is shown when the player is in a party.",
									get = function(info) return GetBarGroupField("showParty") end,
									set = function(info, value) SetBarGroupField("showParty", value) end,
								},
								RaidGroup = {
									type = "toggle", order = 32, name = "In Raid",
									desc = "If checked, bar group is shown when the player is in a raid.",
									get = function(info) return GetBarGroupField("showRaid") end,
									set = function(info, value) SetBarGroupField("showRaid", value) end,
								},
								FishingGroup = {
									type = "toggle", order = 33, name = "Fishing",
									desc = "If checked, bar group is shown when the player is ready to fish (i.e., trained with fishing pole equipped).",
									get = function(info) return GetBarGroupField("showFishing") end,
									set = function(info, value) SetBarGroupField("showFishing", value) end,
								},
								spacer = { type = "description", name = "", order = 50, },
								CheckCondition = {
									type = "toggle", order = 60, name = "Condition Is True",
									desc = "If checked, bar group is shown only when the selected condition is true.",
									get = function(info) return GetBarGroupField("checkCondition") end,
									set = function(info, value) if not value then SetBarGroupField("condition", nil) end; SetBarGroupField("checkCondition", value) end,
								},
								SelectCondition = {
									type = "select", order = 65, name = "  Condition",
									disabled = function(info) return not GetBarGroupField("checkCondition") end,
									get = function(info) return GetBarGroupSelectedCondition(GetSelectConditionList()) end,
									set = function(info, value) SetBarGroupField("condition", GetSelectConditionList()[value]) end,
									values = function(info) return GetSelectConditionList() end,
									style = "dropdown",
								},
							},
						},
						CombatGroup = {
							type = "group", order = 30, name = "Combat Options", inline = true,
							disabled = function(info) return not GetBarGroupField("showCombat") end,
							hidden = function(info) return GetBarGroupField("merged") end,
							args = {
								CombatClicks = {
									type = "toggle", order = 10, name = "Allow Clicks", 
									desc = "If checked, clicking bars is allowed during combat. " ..
											"Right-clicking buff icons for the player will remove the buff. " ..
											"Shift-right-clicking the icon for any bar in a custom group will disable it and "..
											"in an auto group will add the associated spell to the filter list.",
									get = function(info) return GetBarGroupField("combatClicks") end,
									set = function(info, value) SetBarGroupField("combatClicks", value) end,
								},
								CombatTooltips = {
									type = "toggle", order = 20, name = "Show Tooltips", 
									desc = "If checked, tooltips are shown during combat.",
									get = function(info) return GetBarGroupField("combatTips") end,
									set = function(info, value) SetBarGroupField("combatTips", value) end,
								},
								CombatFader = {
									type = "toggle", order = 30, name = "Combat Fader", 
									desc = "If checked, the bar group's opacity will be reduced when in combat.",
									get = function(info) return GetBarGroupField("combatFader") end,
									set = function(info, value) SetBarGroupField("combatFader", value) end,
								},
								CombatAlpha = {
									type = "range", order = 40, name = "Combat Opacity", min = 0, max = 1, step = 0.05,
									desc = "Adjust opacity of bar group in combat.",
									disabled = function(info) return not GetBarGroupField("combatFader") end,
									get = function(info) return GetBarGroupField("combatAlpha") end,
									set = function(info, value) SetBarGroupField("combatAlpha", value) end,
								},
							},
						},
					},
				},
				BarTab = {
					type = "group", order = 15, name = "Custom Bars",
					hidden = function(info) return NoBarGroup() or GetBarGroupField("auto") end,
					args = {
						NewBarButton = {
							type = "execute", order = 1, name = "New", width = "half",
							desc = "Create a new bar.",
							disabled = function(info) return InMode("Bar") end,
							func = function(info) EnterNewBar("start") end,
						},
						DeleteBar = {
							type = "execute", order = 2, name = "Delete", width = "half",
							desc = "Delete the selected bar.",
							disabled = function(info) return NoBar() end,
							func = function(info) DeleteBar() end,
							confirm = function(info) return 'DELETE BAR\nAre you sure you want to delete the selected bar?' end,
						},
						-- Bars get plugged in here, with order starting at 10
					},
				},
				DetectBuffsTab = {
					type = "group", order = 20, name = "Buffs",
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() or not GetBarGroupField("auto") end,
					args = {
						EnableGroup = {
							type = "group", order = 1, name = "Enable", inline = true,
							args = {
								DetectEnable = {
									type = "toggle", order = 1, name = "Auto Buffs",
									desc = 'Enable automatically displaying bars for buffs that match these settings.',
									get = function(info) return GetBarGroupField("detectBuffs") end,
									set = function(info, value) SetBarGroupField("detectBuffs", value) end,
								},
								TrackingEnable = {
									type = "toggle", order = 10, name = "Include Tracking",
									disabled = function(info) return not GetBarGroupField("detectBuffs") end,
									desc = 'Include the active tracking type as a buff.',
									get = function(info) return GetBarGroupField("detectTracking") end,
									set = function(info, value) SetBarGroupField("detectTracking", value) end,
								},
							},
						},
						MonitorUnitGroup = {
							type = "group", order = 10, name = "Action On", inline = true,
							disabled = function(info) return not GetBarGroupField("detectBuffs") end,
							args = {
								PlayerBuff = {
									type = "toggle", order = 10, name = "Player",
									desc = "If checked, only add bars for buffs if they are on the player.",
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "player" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "player") end,
								},
								PetBuff = {
									type = "toggle", order = 15, name = "Pet",
									desc = "If checked, only add bars for buffs if they are on the player's pet.",
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "pet" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "pet") end,
								},
								TargetBuff = {
									type = "toggle", order = 20, name = "Target",
									desc = "If checked, only add bars for buffs if they are on the target.",
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "target" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "target") end,
								},
								FocusBuff = {
									type = "toggle", order = 30, name = "Focus",
									desc = "If checked, only add bars for buffs if they are on the focus.",
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "focus" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "focus") end,
								},
							},
						},
						CastUnitGroup = {
							type = "group", order = 20, name = "Cast By", inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectBuffs") end,
							args = {
								MyBuff = {
									type = "toggle", order = 10, name = "Player",
									desc = "If checked, only add bars for buffs if cast by the player.",
									get = function(info) return GetBarGroupField("detectBuffsCastBy") == "Player" end,
									set = function(info, value) SetBarGroupField("detectBuffsCastBy", "Player") end,
								},
								PetBuff = {
									type = "toggle", order = 15, name = "Pet",
									desc = "If checked, only add bars for buffs if cast by the player's pet.",
									get = function(info) return GetBarGroupField("detectBuffsCastBy") == "Pet" end,
									set = function(info, value) SetBarGroupField("detectBuffsCastBy", "Pet") end,
								},
								YourBuff = {
									type = "toggle", order = 20, name = "Other",
									desc = "If checked, only add bars for buffs if cast by anyone other than the player or pet.",
									get = function(info) return GetBarGroupField("detectBuffsCastBy") == "Other" end,
									set = function(info, value) SetBarGroupField("detectBuffsCastBy", "Other") end,
								},
								OurBuff = {
									type = "toggle", order = 30, name = "Anyone",
									desc = "If checked, add bars for buffs if cast by anyone, including player.",
									get = function(info) return GetBarGroupField("detectBuffsCastBy") == "Anyone" end,
									set = function(info, value) SetBarGroupField("detectBuffsCastBy", "Anyone") end,
								},
							},
						},
						FilterGroup = {
							type = "group", order = 40, name = "Filter", inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectBuffs") end,
							args = {
								FilterBuffs = {
									type = "toggle", order = 10, name = "Filter Buffs",
									desc = "If checked, don't display any buffs that are in the filter list.",
									get = function(info) return GetBarGroupField("filterBuff") end,
									set = function(info, value) SetBarGroupField("filterBuff", value) end,
								},
								LinkFilters = {
									type = "toggle", order = 15, name = "Link Filter",
									desc = "If checked, the buff filter list is shared with bar groups in other profiles with the same name. ",
									disabled = function(info) return not GetBarGroupField("detectBuffs") or not GetBarGroupField("filterBuff") end,
									get = function(info) return GetBarGroupField("filterBuffLink") end,
									set = function(info, value) SetBarGroupField("filterBuffLink", value) end,
								},
								Space1 = { type = "description", name = "", order = 25 },
								AddFilter = {
									type = "input", order = 30, name = " Enter Buff",
									desc = " Enter a buff to be added to the filter list.",
									disabled = function(info) return not GetBarGroupField("detectBuffs") or not GetBarGroupField("filterBuff") end,
									get = function(info) return nil end,
									set = function(info, value) ValidateSpellName(value); AddBarGroupFilter("Buff", value) end,
								},
								SelectFilter = {
									type = "select", order = 40, name = "  Filter List",
									disabled = function(info) return not GetBarGroupField("detectBuffs") or not GetBarGroupField("filterBuff") end,
									get = function(info) return GetBarGroupFilterSelection("Buff") end,
									set = function(info, value) SetBarGroupField("filterBuffSelection", value) end,
									values = function(info) return GetBarGroupFilter("Buff") end,
									style = "dropdown",
								},
								DeleteFilter = {
									type = "execute", order = 50, name = "Delete", width = "half",
									disabled = function(info) return not GetBarGroupField("detectBuffs") or not GetBarGroupField("filterBuff") end,
									desc = "Delete the selected buff from the filter list.",
									func = function(info) DeleteBarGroupFilter("Buff", GetBarGroupField("filterBuffSelection")) end,
								},
								ResetFilter = {
									type = "execute", order = 60, name = "Reset", width = "half",
									desc = "Reset the buff filter list.",
									disabled = function(info) return not GetBarGroupField("detectBuffs") or not GetBarGroupField("filterBuff") end,
									confirm = function(info) return 'RESET\nAre you sure you want to reset the buff filter list?' end,
									func = function(info) ResetBarGroupFilter("Buff") end,
								},
							},
						},
					},
				},
				DetectDebuffsTab = {
					type = "group", order = 25, name = "Debuffs",
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() or not GetBarGroupField("auto") end,
					args = {
						EnableGroup = {
							type = "group", order = 1, name = "Enable", inline = true,
							args = {
								DetectEnable = {
									type = "toggle", order = 1, name = "Auto Debuffs",
									desc = 'Enable automatically displaying bars for debuffs that match these settings.',
									get = function(info) return GetBarGroupField("detectDebuffs") end,
									set = function(info, value) SetBarGroupField("detectDebuffs", value) end,
								},
							},
						},
						MonitorUnitGroup = {
							type = "group", order = 10, name = "Action On", inline = true,
							disabled = function(info) return not GetBarGroupField("detectDebuffs") end,
							args = {
								PlayerBuff = {
									type = "toggle", order = 10, name = "Player",
									desc = "If checked, only add bars for debuffs if they are on the player.",
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "player" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "player") end,
								},
								PetBuff = {
									type = "toggle", order = 15, name = "Pet",
									desc = "If checked, only add bars for debuffs if they are on the player's pet.",
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "pet" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "pet") end,
								},
								TargetBuff = {
									type = "toggle", order = 20, name = "Target",
									desc = "If checked, only add bars for debuffs if they are on the target.",
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "target" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "target") end,
								},
								FocusBuff = {
									type = "toggle", order = 30, name = "Focus",
									desc = "If checked, only add bars for debuffs if they are on the focus.",
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "focus" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "focus") end,
								},
							},
						},
						CastUnitGroup = {
							type = "group", order = 20, name = "Cast By", inline = true,
							disabled = function(info) return not GetBarGroupField("detectDebuffs") end,
							args = {
								MyBuff = {
									type = "toggle", order = 10, name = "Player",
									desc = "If checked, only add bars for debuffs if cast by the player.",
									get = function(info) return GetBarGroupField("detectDebuffsCastBy") == "Player" end,
									set = function(info, value) SetBarGroupField("detectDebuffsCastBy", "Player") end,
								},
								PetBuff = {
									type = "toggle", order = 15, name = "Pet",
									desc = "If checked, only add bars for debuffs if cast by the player's pet.",
									get = function(info) return GetBarGroupField("detectDebuffsCastBy") == "Pet" end,
									set = function(info, value) SetBarGroupField("detectDebuffsCastBy", "Pet") end,
								},
								YourBuff = {
									type = "toggle", order = 20, name = "Other",
									desc = "If checked, only add bars for debuffs if cast by anyone other than the player or pet.",
									get = function(info) return GetBarGroupField("detectDebuffsCastBy") == "Other" end,
									set = function(info, value) SetBarGroupField("detectDebuffsCastBy", "Other") end,
								},
								OurBuff = {
									type = "toggle", order = 30, name = "Anyone",
									desc = "If checked, add bars for debuffs if cast by anyone, including player.",
									get = function(info) return GetBarGroupField("detectDebuffsCastBy") == "Anyone" end,
									set = function(info, value) SetBarGroupField("detectDebuffsCastBy", "Anyone") end,
								},
							},
						},
						FilterGroup = {
							type = "group", order = 40, name = "Filter", inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectDebuffs") end,
							args = {
								FilterBuffs = {
									type = "toggle", order = 10, name = "Filter Debuffs",
									desc = "If checked, don't display any debuffs that are in the filter list.",
									get = function(info) return GetBarGroupField("filterDebuff") end,
									set = function(info, value) SetBarGroupField("filterDebuff", value) end,
								},
								LinkFilters = {
									type = "toggle", order = 15, name = "Link Filter",
									desc = "If checked, the debuff filter list is shared with bar groups in other profiles with the same name. ",
									disabled = function(info) return not GetBarGroupField("detectDebuffs") or not GetBarGroupField("filterDebuff") end,
									get = function(info) return GetBarGroupField("filterDebuffLink") end,
									set = function(info, value) SetBarGroupField("filterDebuffLink", value) end,
								},
								Space1 = { type = "description", name = "", order = 25 },
								AddFilter = {
									type = "input", order = 30, name = " Enter Debuff",
									desc = " Enter a debuff to be added to the filter list.",
									disabled = function(info) return not GetBarGroupField("detectDebuffs") or not GetBarGroupField("filterDebuff") end,
									get = function(info) return nil end,
									set = function(info, value) ValidateSpellName(value); AddBarGroupFilter("Debuff", value) end,
								},
								SelectFilter = {
									type = "select", order = 40, name = "  Filter List",
									disabled = function(info) return not GetBarGroupField("detectDebuffs") or not GetBarGroupField("filterDebuff") end,
									get = function(info) return GetBarGroupField("filterDebuffSelection") end,
									set = function(info, value) SetBarGroupField("filterDebuffSelection", value) end,
									values = function(info) return GetBarGroupFilter("Debuff") end,
									style = "dropdown",
								},
								DeleteFilter = {
									type = "execute", order = 50, name = "Delete", width = "half",
									disabled = function(info) return not GetBarGroupField("detectDebuffs") or not GetBarGroupField("filterDebuff") end,
									desc = "Delete the selected debuff from the filter list.",
									func = function(info) DeleteBarGroupFilter("Debuff", GetBarGroupField("filterDebuffSelection")) end,
								},
								ResetFilter = {
									type = "execute", order = 60, name = "Reset", width = "half",
									desc = "Reset the debuff filter list.",
									disabled = function(info) return not GetBarGroupField("detectDebuffs") or not GetBarGroupField("filterDebuff") end,
									confirm = function(info) return 'RESET\nAre you sure you want to reset the debuff filter list?' end,
									func = function(info) ResetBarGroupFilter("Debuff") end,
								},
							},
						},
					},
				},
				DetectCooldownsTab = {
					type = "group", order = 30, name = "Cooldowns",
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() or not GetBarGroupField("auto") end,
					args = {
						EnableGroup = {
							type = "group", order = 1, name = "Enable", inline = true,
							args = {
								DetectEnable = {
									type = "toggle", order = 1, name = "Auto Cooldowns",
									desc = 'Enable automatically displaying bars for cooldowns that match these settings.',
									get = function(info) return GetBarGroupField("detectCooldowns") end,
									set = function(info, value) SetBarGroupField("detectCooldowns", value) end,
								},
							},
						},
						ActionUnitGroup = {
							type = "group", order = 20, name = "Action By", inline = true,
							disabled = function(info) return not GetBarGroupField("detectCooldowns") end,
							args = {
								MyBuff = {
									type = "toggle", order = 10, name = "Player",
									desc = "If checked, only add bars for cooldowns associated with the player.",
									get = function(info) return GetBarGroupField("detectCooldownsBy") == "Player" end,
									set = function(info, value) SetBarGroupField("detectCooldownsBy", "Player") end,
								},
								PetBuff = {
									type = "toggle", order = 20, name = "Pet",
									desc = "If checked, only add bars for cooldowns associated with the player's pet.",
									get = function(info) return GetBarGroupField("detectCooldownsBy") == "Pet" end,
									set = function(info, value) SetBarGroupField("detectCooldownsBy", "Pet") end,
								},
								AnyBuff = {
									type = "toggle", order = 30, name = "Anyone",
									desc = "If checked, add bars for cooldowns cast by either player or player's pet.",
									get = function(info) return GetBarGroupField("detectCooldownsBy") == "Anyone" end,
									set = function(info, value) SetBarGroupField("detectCooldownsBy", "Anyone") end,
								},
							},
						},
						CooldownTypeGroup = {
							type = "group", order = 30, name = "Include", inline = true,
							disabled = function(info) return not GetBarGroupField("detectCooldowns") end,
							args = {
								SpellCooldowns = {
									type = "toggle", order = 10, name = "Spells",
									desc = "If checked, include bars for spell cooldowns.",
									get = function(info) return GetBarGroupField("detectSpellCooldowns") end,
									set = function(info, value) SetBarGroupField("detectSpellCooldowns", value) end,
								},
								TrinketCooldowns = {
									type = "toggle", order = 20, name = "Trinkets",
									desc = "If checked, include bars for cooldowns on equipped trinkets.",
									get = function(info) return GetBarGroupField("detectTrinketCooldowns") end,
									set = function(info, value) SetBarGroupField("detectTrinketCooldowns", value) end,
								},
								PotionCooldowns = {
									type = "toggle", order = 30, name = "Potions/Elixirs",
									desc = "If checked, include bars for shared potion/elixir cooldowns.",
									get = function(info) return GetBarGroupField("detectPotionCooldowns") end,
									set = function(info, value) SetBarGroupField("detectPotionCooldowns", value) end,
								},
								OtherCooldowns = {
									type = "toggle", order = 40, name = "Other",
									desc = "If checked, include bars for other cooldowns.",
									get = function(info) return GetBarGroupField("detectOtherCooldowns") end,
									set = function(info, value) SetBarGroupField("detectOtherCooldowns", value) end,
								},
							},
						},
						SharedCooldownGroup = {
							type = "group", order = 30, name = "Shared Cooldowns", inline = true,
							disabled = function(info) return not GetBarGroupField("detectCooldowns") end,
							args = {
								JudgementsCooldowns = {
									type = "toggle", order = 10, name = "Judgements",
									desc = "If checked, only show one cooldown for Paladin Judgements.",
									get = function(info) return GetBarGroupField("detectSharedJudgements") end,
									set = function(info, value) SetBarGroupField("detectSharedJudgements", value) end,
								},
								FrostTrapsCooldowns = {
									type = "toggle", order = 20, name = "Frost Traps",
									desc = "If checked, only show one cooldown for Hunter Frost Traps.",
									get = function(info) return GetBarGroupField("detectSharedFrostTraps") end,
									set = function(info, value) SetBarGroupField("detectSharedFrostTraps", value) end,
								},
								ShockCooldowns = {
									type = "toggle", order = 20, name = "Shocks",
									desc = "If checked, only show one cooldown for Shaman Shocks.",
									get = function(info) return GetBarGroupField("detectSharedShocks") end,
									set = function(info, value) SetBarGroupField("detectSharedShocks", value) end,
								},
							},
						},
						FilterGroup = {
							type = "group", order = 40, name = "Filter", inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectCooldowns") end,
							args = {
								FilterBuffs = {
									type = "toggle", order = 10, name = "Filter Cooldowns",
									desc = "If checked, don't display any cooldowns that are in the filter list.",
									get = function(info) return GetBarGroupField("filterCooldown") end,
									set = function(info, value) SetBarGroupField("filterCooldown", value) end,
								},
								LinkFilters = {
									type = "toggle", order = 15, name = "Link Filter",
									desc = "If checked, the cooldown filter list is shared with bar groups in other profiles with the same name. ",
									disabled = function(info) return not GetBarGroupField("detectCooldowns") or not GetBarGroupField("filterCooldown") end,
									get = function(info) return GetBarGroupField("filterCooldownLink") end,
									set = function(info, value) SetBarGroupField("filterCooldownLink", value) end,
								},
								Space1 = { type = "description", name = "", order = 25 },
								AddFilter = {
									type = "input", order = 30, name = " Enter Cooldown",
									desc = " Enter a cooldown to be added to the filter list.",
									disabled = function(info) return not GetBarGroupField("detectCooldowns") or not GetBarGroupField("filterCooldown") end,
									get = function(info) return nil end,
									set = function(info, value) ValidateSpellName(value); AddBarGroupFilter("Cooldown", value) end,
								},
								SelectFilter = {
									type = "select", order = 40, name = "  Filter List",
									disabled = function(info) return not GetBarGroupField("detectCooldowns") or not GetBarGroupField("filterCooldown") end,
									get = function(info) return GetBarGroupField("filterCooldownSelection") end,
									set = function(info, value) SetBarGroupField("filterCooldownSelection", value) end,
									values = function(info) return GetBarGroupFilter("Cooldown") end,
									style = "dropdown",
								},
								DeleteFilter = {
									type = "execute", order = 50, name = "Delete", width = "half",
									disabled = function(info) return not GetBarGroupField("detectCooldowns") or not GetBarGroupField("filterCooldown") end,
									desc = "Delete the selected cooldown from the filter list.",
									func = function(info) DeleteBarGroupFilter("Cooldown", GetBarGroupField("filterCooldownSelection")) end,
								},
								ResetFilter = {
									type = "execute", order = 60, name = "Reset", width = "half",
									desc = "Reset the cooldown filter list.",
									disabled = function(info) return not GetBarGroupField("detectCooldowns") or not GetBarGroupField("filterCooldown") end,
									confirm = function(info) return 'RESET\nAre you sure you want to reset the cooldown filter list?' end,
									func = function(info) ResetBarGroupFilter("Cooldown") end,
								},
							},
						},
					},
				},
				LayoutTab = {
					type = "group", order = 40, name = "Layout",
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() or GetBarGroupField("merged") end,
					args = {
						OrientationGroup = {
							type = "group", order = 10, name = "Configuration", inline = true,
							args = {
								Configuration = {
									type = "select", order = 1, name = "  Options", width = "double",
									desc = "Select a configuration option for bars or icons.",
									get = function(info) return GetBarGroupField("configuration") end,
									set = function(info, value) SetBarGroupField("configuration", value) end,
									values = function(info) return GetOrientationList() end,
									style = "dropdown",
								},
								ReverseGrowthGroup = {
									type = "toggle", order = 20, name = "Direction",
									desc = "If checked, grow up or to the right from the anchor, " ..
											"otherwise grow down or to the left.",
									get = function(info) return GetBarGroupField("growDirection") end,
									set = function(info, value) SetBarGroupField("growDirection", value) end,
								},
								CopyLayoutGroup = {
									type = "select", order = 60, name = "  Copy Layout From",
									desc = "Select bar group to copy all layout settings from.",
									get = function(info) return nil end,
									set = function(info, value) CopyBarGroupConfiguration(GetBarGroupList()[value]) end,
									values = function(info) return GetBarGroupList() end,
									style = "dropdown",
								},
							},
						},
						DimensionGroup = {
							type = "group", order = 20, name = "Dimensions", inline = true,
							args = {
								UseDefaultsGroup = {
									type = "toggle", order = 1, name = "Use Defaults",
									desc = "If checked, bar dimensions are set to default values.",
									get = function(info) return GetBarGroupField("useDefaultDimensions") end,
									set = function(info, value) SetBarGroupField("useDefaultDimensions", value) end,
								},
								RestoreDefaults = {
									type = "execute", order = 5, name = "Restore Defaults",
									desc = "Reset dimensions for this bar group back to the current defaults.",
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									func = function(info) local bg = GetBarGroupEntry()
										bg.barWidth = MOD.db.global.Defaults.barWidth
										bg.barHeight = MOD.db.global.Defaults.barHeight
										bg.iconSize = MOD.db.global.Defaults.iconSize
										bg.scale = MOD.db.global.Defaults.scale				
										bg.spacingX = MOD.db.global.Defaults.spacingX
										bg.spacingY = MOD.db.global.Defaults.spacingY
										bg.iconOffsetX = MOD.db.global.Defaults.iconOffsetX
										bg.iconOffsetY = MOD.db.global.Defaults.iconOffsetY
										MOD:UpdateAllBarGroups()
									end,
								},
								Space1 = { type = "description", name = "", order = 10 },
								BarWidth = {
									type = "range", order = 20, name = "Bar Width", min = 5, max = 500, step = 1,
									desc = "Set width of bars.",
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("barWidth") end,
									set = function(info, value) SetBarGroupField("barWidth", value) end,
								},
								BarHeight = {
									type = "range", order = 25, name = "Bar Height", min = 5, max = 100, step = 1,
									desc = "Set height of bars.",
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("barHeight") end,
									set = function(info, value) SetBarGroupField("barHeight", value) end,
								},
								IconSize = {
									type = "range", order = 30, name = "Icon Size", min = 5, max = 100, step = 1,
									desc = "Set width/height for icons.",
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("iconSize") end,
									set = function(info, value) SetBarGroupField("iconSize", value) end,
								},
								Scale = {
									type = "range", order = 35, name = "Scale", min = 0.1, max = 2, step = 0.05,
									desc = "Set scale factor for bars and icons.",
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("scale") end,
									set = function(info, value) SetBarGroupField("scale", value) end,
								},
								Space2 = { type = "description", name = "", order = 40 },
								HorizontalSpacing = {
									type = "range", order = 60, name = "Horizontal Spacing", min = 0, max = 100, step = 1,
									desc = "Adjust horizontal spacing between bars.",
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("spacingX") end,
									set = function(info, value) SetBarGroupField("spacingX", value) end,
								},
								VerticalSpacing = {
									type = "range", order = 65, name = "Vertical Spacing", min = 0, max = 100, step = 1,
									desc = "Adjust vertical spacing between bars.",
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("spacingY") end,
									set = function(info, value) SetBarGroupField("spacingY", value) end,
								},
								IconOffsetX = {
									type = "range", order = 70, name = "Icon Inset", min = -100, max = 100, step = 1,
									desc = "Set icon's horizontal inset from bar.",
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("iconOffsetX") end,
									set = function(info, value) SetBarGroupField("iconOffsetX", value) end,
								},
								IconOffsetY = {
									type = "range", order = 75, name = "Icon Offset", min = -100, max = 100, step = 1,
									desc = "Set vertical offset between icon and bar.",
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("iconOffsetY") end,
									set = function(info, value) SetBarGroupField("iconOffsetY", value) end,
								},
							},
						},
						BarFormatGroup = {
							type = "group", order = 20, name = "Format", inline = true,
							args = {
								HideIconGroup = {
									type = "toggle", order = 30, name = "Icon", width = "half",
									desc = "Show icons with bars (required for tooltips).",
									get = function(info) return not GetBarGroupField("hideIcon") end,
									set = function(info, value) SetBarGroupField("hideIcon", not value) end,
								},
								HideClockGroup = {
									type = "toggle", order = 31, name = "Clock", width = "half",
									desc = "Show clock animation on icons for timer bars.",
									get = function(info) return not GetBarGroupField("hideClock") end,
									set = function(info, value) SetBarGroupField("hideClock", not value) end,
								},
								HideBarGroup = {
									type = "toggle", order = 32, name = "Bar", width = "half",
									desc = "Show colored bar and background.",
									get = function(info) return not GetBarGroupField("hideBar") end,
									set = function(info, value) SetBarGroupField("hideBar", not value) end,
								},
								HideSparkGroup = {
									type = "toggle", order = 33, name = "Spark ", width = "half",
									desc = "Show spark that moves across bars to indicate remaining time.",
									disabled = function(info) return GetBarGroupField("hideBar") end,
									get = function(info) return not GetBarGroupField("hideSpark") end,
									set = function(info, value) SetBarGroupField("hideSpark", not value) end,
								},
								HideLabelGroup = {
									type = "toggle", order = 34, name = "Label", width = "half",
									desc = "Show label text on bars.",
									get = function(info) return not GetBarGroupField("hideLabel") end,
									set = function(info, value) SetBarGroupField("hideLabel", not value) end,
								},
								HideTimerGroup = {
									type = "toggle", order = 35, name = "Time", width = "half",
									desc = "Show time left on bars that have a duration.",
									get = function(info) return not GetBarGroupField("hideValue") end,
									set = function(info, value) SetBarGroupField("hideValue", not value) end,
								},
								TooltipsGroup = {
									type = "toggle", order = 36, name = "Tooltips", width = "half",
									desc = "Show tooltips when the cursor is over a bar's icon.",
									get = function(info) return GetBarGroupField("showTooltips") end,
									set = function(info, value) SetBarGroupField("showTooltips", value) end,
								},
							},
						},
						TextSettings = {
							type = "group", order = 30, name = "Text Settings", inline = true,
							args = {
								LabelInset = {
									type = "range", order = 10, name = "Label Text Inset", min = -100, max = 100, step = 1,
									desc = "Set horizontal inset for label from edge of bar.",
									get = function(info) return GetBarGroupField("labelInset") end,
									set = function(info, value) SetBarGroupField("labelInset", value) end,
								},
								LabelOffset = {
									type = "range", order = 15, name = "Label Text Offset", min = -100, max = 100, step = 1,
									desc = "Set vertical offset for label text from center of bar.",
									get = function(info) return GetBarGroupField("labelOffset") end,
									set = function(info, value) SetBarGroupField("labelOffset", value) end,
								},
								LabelWrapGroup = {
									type = "toggle", order = 20, name = "Wrap", width = "half",
									desc = "If checked, wrap label text when it doesn't fit in the bar's width.",
									get = function(info) return GetBarGroupField("labelWrap") end,
									set = function(info, value) SetBarGroupField("labelWrap", value) end,
								},
								LabelTopGroup = {
									type = "toggle", order = 21, name = "Top", width = "half",
									desc = "If checked, set \"Top\" alignment for label text.",
									get = function(info) return GetBarGroupField("labelAlign") == "TOP" end,
									set = function(info, value) SetBarGroupField("labelAlign", "TOP") end,
								},
								LabelCenterGroup = {
									type = "toggle", order = 22, name = "Middle", width = "half",
									desc = "If checked, set \"Middle\" alignment for label text.",
									get = function(info) return GetBarGroupField("labelAlign") == "MIDDLE" end,
									set = function(info, value) SetBarGroupField("labelAlign", "MIDDLE") end,
								},
								LabelBottomGroup = {
									type = "toggle", order = 23, name = "Bottom", width = "half",
									desc = "If checked, set \"Bottom\" alignment for label text.",
									get = function(info) return GetBarGroupField("labelAlign") == "BOTTOM" end,
									set = function(info, value) SetBarGroupField("labelAlign", "BOTTOM") end,
								},
								Space1 = { type = "description", name = "", order = 30 },
								TimeTextInset = {
									type = "range", order = 40, name = "Time Text Inset", min = -100, max = 100, step = 1,
									desc = "Set horizontal inset for time text from edge of bar.",
									get = function(info) return GetBarGroupField("timeInset") end,
									set = function(info, value) SetBarGroupField("timeInset", value) end,
								},
								TimeTextOffset = {
									type = "range", order = 45, name = "Time Text Offset", min = -100, max = 100, step = 1,
									desc = "Set vertical offset for time text from center of bar.",
									get = function(info) return GetBarGroupField("timeOffset") end,
									set = function(info, value) SetBarGroupField("timeOffset", value) end,
								},
								IconTextInset = {
									type = "range", order = 50, name = "Icon Text Inset", min = -100, max = 100, step = 1,
									desc = "Set horizontal inset for icon text from middle of icon.",
									get = function(info) return GetBarGroupField("iconInset") end,
									set = function(info, value) SetBarGroupField("iconInset", value) end,
								},
								IconTextOffset = {
									type = "range", order = 55, name = "Icon Text Offset", min = -100, max = 100, step = 1,
									desc = "Set vertical offset for icon text from center of icon.",
									get = function(info) return GetBarGroupField("iconOffset") end,
									set = function(info, value) SetBarGroupField("iconOffset", value) end,
								},
							},
						},
						IconOnlyGroup = {
							type = "group", order = 35, name = "Icon Configuration Options", inline = true,
							hidden = function(info)
								local config = GetBarGroupField("configuration")
								if config then return not Nest_SupportedConfigurations[config].iconOnly else return true end
							end,
							args = {
								Wrap = {
									type = "range", order = 20, name = "Wrap", min = 0, max = 50, step = 1,
									desc = "Set how many icons to line up before wrapping to next row or column.",
									get = function(info) return GetBarGroupField("wrap") end,
									set = function(info, value) SetBarGroupField("wrap", value) end,
								},
								WrapDirection = {
									type = "toggle", order = 25, name = "Wrap Direction",
									desc = "If checked, wrap icons up when arranged in rows or to the right when arranged in columns, " ..
										"otherwise wrap down or to the left.",
									get = function(info) return GetBarGroupField("wrapDirection") end,
									set = function(info, value) SetBarGroupField("wrapDirection", value) end,
								},
								SnapCenter = {
									type = "toggle", order = 30, name = "Snap To Center",
									desc = "If checked and the bar group is locked, icons snap to center at the anchor position.",
									get = function(info) return GetBarGroupField("snapCenter") end,
									set = function(info, value) SetBarGroupField("snapCenter", value) end,
								},
							},
						},
						PositionGroup = {
							type = "group", order = 40, name = "Display Position", inline = true,
							args = {
								Horizontal = {
									type = "range", order = 10, name = "Horizontal", min = 0, max = 100, step = 0.01,
									desc = "Set horizontal position as percentage of overall width (cannot move beyond edge of display).",
									get = function(info) return GetBarGroupField("pointX") * 100 end,
									set = function(info, value) SetBarGroupField("pointX", value / 100) end,
								},
								Vertical = {
									type = "range", order = 20, name = "Vertical", min = 0, max = 100, step = 0.01,
									desc = "Set vertical position as percentage of overall height (cannot move beyond edge of display).",
									get = function(info) return GetBarGroupField("pointY") * 100 end,
									set = function(info, value) SetBarGroupField("pointY", value / 100) end,
								},
							},
						},
						AnchorGroup = {
							type = "group", order = 50, name = "Attachment", inline = true,
							args = {
								Anchor = {
									type = "select", order = 10, name = "  Bar Group",
									desc = "Select a bar group to attach to (for independent position, attach to self).",
									get = function(info) return GetBarGroupAnchor() end,
									set = function(info, value) SetBarGroupAnchor(value) end,
									values = function(info) return GetBarGroupList() end,
									style = "dropdown",
								},
								OffsetX = {
									type = "range", order = 20, name = "Offset X", min = -500, max = 500, step = 1,
									desc = "Set horizontal offset from the selected bar group.",
									disabled = function(info) return not GetBarGroupField("anchor") end,
									get = function(info) return GetBarGroupField("anchorX") end,
									set = function(info, value) SetBarGroupField("anchorX", value) end,
								},
								OffsetY = {
									type = "range", order = 30, name = "Offset Y", min = -500, max = 500, step = 1,
									desc = "Set vertical offset from the selected bar group.",
									disabled = function(info) return not GetBarGroupField("anchor") end,
									get = function(info) return GetBarGroupField("anchorY") end,
									set = function(info, value) SetBarGroupField("anchorY", value) end,
								},
								Relative = {
									type = "toggle", order = 40, name = "Last Bar", width = "half",
									desc = "If checked, position is relative to last bar in the selected bar group.",
									disabled = function(info) return not GetBarGroupField("anchor") end,
									get = function(info) return GetBarGroupField("anchorLastBar") end,
									set = function(info, value) SetBarGroupField("anchorLastBar", value) end,
								},
								ResetAnchor = {
									type = "execute", order = 60, name = "Reset", width = "half",
									desc = "Reset position to screen center, unlock anchor, and reset attachment options.",
									func = function(info) SetBarGroupAnchor(nil) end,
								},
							},
						},
					},
				},
				AppearanceTab = {
					type = "group", order = 45, name = "Appearance",
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() or GetBarGroupField("merged") end,
					args = {
						FontsGroup = {
							type = "group", order = 20, name = "Fonts and Textures", inline = true,
							args = {
								UseDefaultsGroup = {
									type = "toggle", order = 1, name = "Use Defaults",
									desc = "If checked, fonts and textures use the default values.",
									get = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									set = function(info, value) SetBarGroupField("useDefaultFontsAndTextures", value) end,
								},
								RestoreDefaults = {
									type = "execute", order = 3, name = "Restore Defaults",
									desc = "Reset standard colors for this bar group back to the current defaults.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									func = function(info) local bg = GetBarGroupEntry()
										bg.labelFont = MOD.db.global.Defaults.labelFont
										bg.labelFSize = MOD.db.global.Defaults.labelFSize
										bg.labelAlpha = MOD.db.global.Defaults.labelAlpha
										bg.labelColor = MOD.db.global.Defaults.labelColor				
										bg.labelOutline = MOD.db.global.Defaults.labelOutline
										bg.timeFont = MOD.db.global.Defaults.timeFont
										bg.timeFSize = MOD.db.global.Defaults.timeFSize
										bg.timeAlpha = MOD.db.global.Defaults.timeAlpha
										bg.timeColor = MOD.db.global.Defaults.timeColor					
										bg.timeOutline = MOD.db.global.Defaults.timeOutline
										bg.iconFont = MOD.db.global.Defaults.iconFont
										bg.iconFSize = MOD.db.global.Defaults.iconFSize
										bg.iconAlpha = MOD.db.global.Defaults.iconAlpha
										bg.iconColor = MOD.db.global.Defaults.iconColor					
										bg.iconOutline = MOD.db.global.Defaults.iconOutline
										bg.texture = MOD.db.global.Defaults.texture
										bg.alpha = MOD.db.global.Defaults.alpha
										bg.bgtexture = MOD.db.global.Defaults.bgtexture	
										bg.bgAlpha = MOD.db.global.Defaults.bgAlpha
										bg.fgSaturation = MOD.db.global.Defaults.fgSaturation	
										bg.fgBrightness = MOD.db.global.Defaults.fgBrightness
										bg.bgSaturation = MOD.db.global.Defaults.bgSaturation	
										bg.bgBrightness = MOD.db.global.Defaults.bgBrightness
										MOD:UpdateAllBarGroups()
									end,
								},
								Space1 = { type = "description", name = "", order = 5 },
								LabelFont = {
									type = "select", order = 21, name = " Label Font", 
									desc = "Select font for label text.",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("labelFont") end,
									set = function(info, value) SetBarGroupField("labelFont", value) end,
								},
								LabelFontSize = {
									type = "range", order = 23, name = "Font Size", min = 5, max = 30, step = 1,
									desc = "Set font size for labels.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("labelFSize") end,
									set = function(info, value) SetBarGroupField("labelFSize", value) end,
								},
								LabelAlpha = {
									type = "range", order = 25, name = "Opacity", min = 0, max = 1, step = 0.05,
									desc = "Set opacity for labels.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("labelAlpha") end,
									set = function(info, value) SetBarGroupField("labelAlpha", value) end,
								},
								LabelColor = {
									type = "color", order = 27, name = "Color", hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info)
										local t = GetBarGroupField("labelColor")
										if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
									end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("labelColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("labelColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								LabelOutline = {
									type = "toggle", order = 29, name = "Outline", width = "half",
									desc = "Add black outline to label.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("labelOutline") end,
									set = function(info, value) SetBarGroupField("labelOutline", value) end,
								},
								Space2 = { type = "description", name = "", order = 30 },
								TimeFont = {
									type = "select", order = 31, name = " Time Font", 
									desc = "Select font for time text.",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("timeFont") end,
									set = function(info, value) SetBarGroupField("timeFont", value) end,
								},
								TimeFontSize = {
									type = "range", order = 33, name = "Font Size", min = 5, max = 30, step = 1,
									desc = "Set font size for time text.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("timeFSize") end,
									set = function(info, value) SetBarGroupField("timeFSize", value) end,
								},
								TimeAlpha = {
									type = "range", order = 35, name = "Opacity", min = 0, max = 1, step = 0.05,
									desc = "Set opacity for time text.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("timeAlpha") end,
									set = function(info, value) SetBarGroupField("timeAlpha", value) end,
								},
								TimeColor = {
									type = "color", order = 37, name = "Color", hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info)
										local t = GetBarGroupField("timeColor")
										if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
									end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("timeColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("timeColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								TimeOutline = {
									type = "toggle", order = 39, name = "Outline", width = "half",
									desc = "Add black outline to time text.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("timeOutline") end,
									set = function(info, value) SetBarGroupField("timeOutline", value) end,
								},
								Space3 = { type = "description", name = "", order = 40 },
								IconFont = {
									type = "select", order = 41, name = " Icon Font", 
									desc = "Select font for icon text (i.e., spell stack count).",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("iconFont") end,
									set = function(info, value) SetBarGroupField("iconFont", value) end,
								},
								IconFontSize = {
									type = "range", order = 43, name = "Font Size", min = 5, max = 30, step = 1,
									desc = "Set font size for icon text.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("iconFSize") end,
									set = function(info, value) SetBarGroupField("iconFSize", value) end,
								},
								IconAlpha = {
									type = "range", order = 45, name = "Opacity", min = 0, max = 1, step = 0.05,
									desc = "Set opacity for icon text.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("iconAlpha") end,
									set = function(info, value) SetBarGroupField("iconAlpha", value) end,
								},
								IconColor = {
									type = "color", order = 47, name = "Color", hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info)
										local t = GetBarGroupField("iconColor")
										if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
									end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("iconColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("iconColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								IconOutline = {
									type = "toggle", order = 49, name = "Outline", width = "half",
									desc = "Add black outline to icon text.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("iconOutline") end,
									set = function(info, value) SetBarGroupField("iconOutline", value) end,
								},
								Space7 = { type = "description", name = "", order = 70 },
								BarTexture = {
									type = "select", order = 72, name = " Bar Texture", 
									desc = "Select texture for bars.",
									dialogControl = 'LSM30_Statusbar',
									values = AceGUIWidgetLSMlists.statusbar,
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("texture") end,
									set = function(info, value) SetBarGroupField("texture", value) end,
								},
								BarAlpha = {
									type = "range", order = 74, name = "Bar Opacity", min = 0, max = 1, step = 0.05,
									desc = "Set alpha for bars.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("alpha") end,
									set = function(info, value) SetBarGroupField("alpha", value) end,
								},
								BackgroundTexture = {
									type = "select", order = 76, name = " Background Texture", 
									desc = "Select background texture for bars.",
									dialogControl = 'LSM30_Statusbar',
									values = AceGUIWidgetLSMlists.statusbar,
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("bgtexture") end,
									set = function(info, value) SetBarGroupField("bgtexture", value) end,
								},
								Background = {
									type = "range", order = 78, name = "Background Opacity", min = 0, max = 1, step = 0.05,
									desc = "Set background opacity for bars.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("bgAlpha") end,
									set = function(info, value) SetBarGroupField("bgAlpha", value) end,
								},
								Space8 = { type = "description", name = "", order = 80 },
								ForegroundSaturation = {
									type = "range", order = 82, name = "Foreground Saturation", min = -1, max = 1, step = 0.05,
									desc = "Set saturation for foreground colors.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("fgSaturation") end,
									set = function(info, value) SetBarGroupField("fgSaturation", value) end,
								},
								ForegroundBrightness = {
									type = "range", order = 84, name = "Foreground Brightness", min = -1, max = 1, step = 0.05,
									desc = "Set brightness for foreground colors.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("fgBrightness") end,
									set = function(info, value) SetBarGroupField("fgBrightness", value) end,
								},
								BackgroundSaturation = {
									type = "range", order = 86, name = "Background Saturation", min = -1, max = 1, step = 0.05,
									desc = "Set saturation for foreground colors.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("bgSaturation") end,
									set = function(info, value) SetBarGroupField("bgSaturation", value) end,
								},
								BackgroundBrightness = {
									type = "range", order = 88, name = "Background Brightness", min = -1, max = 1, step = 0.05,
									desc = "Set brightness for foreground colors.",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("bgBrightness") end,
									set = function(info, value) SetBarGroupField("bgBrightness", value) end,
								},
							},
						},
						ColorsGroup = {
							type = "group", order = 30, name = "Standard Colors", inline = true,
							args = {
								UseDefaultsGroup = {
									type = "toggle", order = 1, name = "Use Defaults",
									desc = "If checked, colors use the default values.",
									get = function(info) return GetBarGroupField("useDefaultColors") end,
									set = function(info, value) SetBarGroupField("useDefaultColors", value) end,
								},
								RestoreDefaults = {
									type = "execute", order = 3, name = "Restore Defaults",
									desc = "Reset standard colors for this bar group back to the current defaults.",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									func = function(info) local bg = GetBarGroupEntry()
										bg.buffColor = MOD.db.global.DefaultBuffColor
										bg.debuffColor = MOD.db.global.DefaultDebuffColor
										bg.cooldownColor = MOD.db.global.DefaultCooldownColor
										bg.notificationColor = MOD.db.global.DefaultNotificationColor							
										bg.poisonColor = MOD.db.global.DefaultPoisonColor
										bg.curseColor = MOD.db.global.DefaultCurseColor
										bg.magicColor = MOD.db.global.DefaultMagicColor
										bg.diseaseColor = MOD.db.global.DefaultDiseaseColor						
										MOD:UpdateAllBarGroups()
									end,
								},
								Space0 = { type = "description", name = "", order = 5 },
								ColorText = { type = "description", name = "Bar Colors:", order = 7, width = "half" },
								NotificationColor = {
									type = "color", order = 10, name = "Notify", hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("notificationColor") or MOD.db.global.DefaultNotificationColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("notificationColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("notificationColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								BuffColor = {
									type = "color", order = 11, name = "Buff", hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("buffColor") or MOD.db.global.DefaultBuffColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("buffColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("buffColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								DebuffColor = {
									type = "color", order = 12, name = "Debuff", hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("debuffColor") or MOD.db.global.DefaultDebuffColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("debuffColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("debuffColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								CooldownColor = {
									type = "color", order = 13, name = "Cooldown", hasAlpha = false,
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("cooldownColor") or MOD.db.global.DefaultCooldownColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("cooldownColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("cooldownColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								Space1 = { type = "description", name = "", order = 14 },
								DebuffText = { type = "description", name = "Debuff Colors:", order = 15, width = "half" },
								PoisonColor = {
									type = "color", order = 15, name = "Poison", hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("poisonColor") or MOD.db.global.DefaultPoisonColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("poisonColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("poisonColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								CurseColor = {
									type = "color", order = 16, name = "Curse", hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("curseColor") or MOD.db.global.DefaultCurseColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("curseColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("curseColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								MagicColor = {
									type = "color", order = 17, name = "Magic", hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("magicColor") or MOD.db.global.DefaultMagicColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("magicColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("magicColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								DiseaseColor = {
									type = "color", order = 18, name = "Disease", hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("diseaseColor") or MOD.db.global.DefaultDiseaseColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("diseaseColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("diseaseColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
							},
						},
						BarColorGroup = {
							type = "group", order = 40, name = "Bar Color Scheme", inline = true,
							args = {
								ForegroundText = { type = "description", name = "Foreground:", order = 1, width = "half" },
								StandardColors = {
									type = "toggle", order = 10, name = "Standard Colors",
									desc = "Show bars in default colors for their type, including special debuff colors when applicable.",
									get = function(info) return GetBarGroupField("barColors") == "Standard" end,
									set = function(info, value) SetBarGroupField("barColors", "Standard") end,
								},
								ClassColors = {
									type = "toggle", order = 20, name = "Class Colors",
									desc = "Show bars in class colors of associated spells when possible, otherwise use default bar colors.",
									get = function(info) return GetBarGroupField("barColors") == "Class" end,
									set = function(info, value) SetBarGroupField("barColors", "Class") end,
								},
								SpellColors = {
									type = "toggle", order = 30, name = "Spell Colors",
									desc = "Show bars using spell colors when possible, otherwise use default bar colors.",
									get = function(info) return GetBarGroupField("barColors") == "Spell" end,
									set = function(info, value) SetBarGroupField("barColors", "Spell") end,
								},
								spacer = { type = "description", name = "", order = 40 },
								BackgroundText = { type = "description", name = "Background:", order = 41, width = "half" },
								NormalBackground = {
									type = "toggle", order = 50, name = "Same as Foreground",
									desc = "Color the background the same as the foreground.",
									get = function(info) return GetBarGroupField("bgColors") == "Normal" end,
									set = function(info, value) SetBarGroupField("bgColors", "Normal") end,
								},
								CustomBackground = {
									type = "toggle", order = 60, name = "Custom", width = "half",
									desc = "Color the background with a custom color.",
									get = function(info) return GetBarGroupField("bgColors") == "Custom" end,
									set = function(info, value) SetBarGroupField("bgColors", "Custom") end,
								},
								BackgroundColor = {
									type = "color", order = 70, name = "Color", hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("bgColors") ~= "Custom" end,
									get = function(info)
										local t = GetBarGroupField("bgColor")
										if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
									end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("bgColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("bgColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
							},
						},
					},
				},
				TimerOptionsTab = {
					type = "group", order = 50, name = "Timer Options",
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() end,
					args = {
						DurationMaxGroup = {
							type = "group", order = 10, name = "Show With Uniform Duration", inline = true,
							args = {
								DurationCheck = {
									type = "toggle", order = 3, name = "Enable", width = "half",
									desc = "Show timer bars scaled with a uniform duration (text still shows actual time left).",
									get = function(info) return GetBarGroupField("setDuration") end,
									set = function(info, value) SetBarGroupField("setDuration", value) end,
								},
								DurationMinutes = {
									type = "range", order = 5, name = "Minutes", min = 0, max = 120, step = 1,
									desc = "Enter minutes in the uniform duration.",
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("setDuration") end,
									get = function(info) local d = GetBarGroupField("uniformDuration"); return math.floor(d / 60) end,
									set = function(info, value) local d = GetBarGroupField("uniformDuration"); SetBarGroupField("uniformDuration", (value * 60) + (d % 60)) end,
								},
								DurationSeconds = {
									type = "range", order = 7, name = "Seconds", min = 0, max = 59, step = 1,
									desc = "Enter seconds in the uniform duration.",
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("setDuration") end,
									get = function(info) local d = GetBarGroupField("uniformDuration"); return d % 60 end,
									set = function(info, value) local d = GetBarGroupField("uniformDuration"); SetBarGroupField("uniformDuration", value + (60 * math.floor(d / 60))) end,
								},
								DurationRange = {
									type = "description", order = 9,
									disabled = function(info) return not GetBarGroupField("setDuration") end, width = "half",
									name = function(info) local d = GetBarGroupField("uniformDuration"); return string.format("      %0d:%02d", math.floor(d / 60), d % 60) end,
								},
							},
						},
						DurationLimitGroup = {
							type = "group", order = 20, name = "Show If Unlimited Duration", inline = true,
							args = {
								LongDurationCheck = {
									type = "toggle", order = 1, name = "Enable", width = "half",
									desc = "Show bars for actions with unlimited duration (e.g., paladin auras).",
									get = function(info) return GetBarGroupField("showNoDuration") end,
									set = function(info, value) SetBarGroupField("showNoDuration", value) end,
								},
								LongDurationLimit = {
									type = "toggle", order = 10, name = "Only Show Unlimited",
									desc = "Show bars for actions only if they have unlimited duration.",
									disabled = function(info) return not GetBarGroupField("showNoDuration") end,
									get = function(info) return GetBarGroupField("showOnlyNoDuration") end,
									set = function(info, value) SetBarGroupField("showOnlyNoDuration", value) end,
								},
								ForegroundBackground = {
									type = "toggle", order = 20, name = "Full Bars",
									desc = "If checked, bars with unlimited duration show as full bars, otherwise they show as empty bars.",
									disabled = function(info) return not GetBarGroupField("showNoDuration") end,
									get = function(info) return not GetBarGroupField("showNoDurationBackground") end,
									set = function(info, value) SetBarGroupField("showNoDurationBackground", not value) end,
								},
							},
						},
						DurationGroup = {
							type = "group", order = 30, name = "Check Overall Duration", inline = true,
							args = {
								DurationCheck = {
									type = "toggle", order = 3, name = "Enable", width = "half",
									desc = "Only include timer bars with a specified minimum (or maximum) duration.",
									get = function(info) return GetBarGroupField("checkDuration") end,
									set = function(info, value) SetBarGroupField("checkDuration", value) end,
								},
								DurationMinutes = {
									type = "range", order = 4, name = "Minutes", min = 0, max = 120, step = 1,
									desc = "Enter minutes for overall duration check.",
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkDuration") end,
									get = function(info) local d = GetBarGroupField("filterDuration"); return math.floor(d / 60) end,
									set = function(info, value) local d = GetBarGroupField("filterDuration"); SetBarGroupField("filterDuration", (value * 60) + (d % 60)) end,
								},
								DurationSeconds = {
									type = "range", order = 5, name = "Seconds", min = 0, max = 59, step = 1,
									desc = "Enter seconds for overall duration check.",
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkDuration") end,
									get = function(info) local d = GetBarGroupField("filterDuration"); return d % 60 end,
									set = function(info, value) local d = GetBarGroupField("filterDuration"); SetBarGroupField("filterDuration", value + (60 * math.floor(d / 60))) end,
								},
								DurationMinMax = {
									type = "select", order = 6, name = "  Duration",
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkDuration") end,
									get = function(info) if GetBarGroupField("minimumDuration") then return 1 else return 2 end end,
									set = function(info, value) if value == 1 then SetBarGroupField("minimumDuration", true) else SetBarGroupField("minimumDuration", false) end end,
									values = function(info)
										local d = GetBarGroupField("filterDuration")
										local ds = string.format("%0d:%02d", math.floor(d / 60), d % 60)
										return { "Show if " .. ds .. " or more", "Show if less than " .. ds }
									end,
									style = "dropdown",
								},
							},
						},
						TimeLeftGroup = {
							type = "group", order = 40, name = "Check Time Left", inline = true,
							disabled = function(info) return GetBarGroupField("showOnlyNoDuration") end,
							args = {
								TimeLeftCheck = {
									type = "toggle", order = 3, name = "Enable", width = "half",
									desc = "Only show timer bars with a specified minimum (or maximum) time left.",
									get = function(info) return GetBarGroupField("checkTimeLeft") end,
									set = function(info, value) SetBarGroupField("checkTimeLeft", value) end,
								},
								TimeLeftMinutes= {
									type = "range", order = 4, name = "Minutes", min = 0, max = 120, step = 1,
									desc = "Enter minutes for time left check.",
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkTimeLeft") end,
									get = function(info) local d = GetBarGroupField("filterTimeLeft"); return math.floor(d / 60) end,
									set = function(info, value) local d = GetBarGroupField("filterTimeLeft"); SetBarGroupField("filterTimeLeft", (value * 60) + (d % 60)) end,
								},
								TimeLeftSeconds = {
									type = "range", order = 5, name = "Seconds", min = 0, max = 59, step = 1,
									desc = "Enter seconds for time left check.",
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkTimeLeft") end,
									get = function(info) local d = GetBarGroupField("filterTimeLeft"); return d % 60 end,
									set = function(info, value) local d = GetBarGroupField("filterTimeLeft"); SetBarGroupField("filterTimeLeft", value + (60 * math.floor(d / 60))) end,
								},
								TimeLeftMinMax = {
									type = "select", order = 6, name = "  Time Left",
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkTimeLeft") end,
									get = function(info) if GetBarGroupField("minimumTimeLeft") then return 1 else return 2 end end,
									set = function(info, value) if value == 1 then SetBarGroupField("minimumTimeLeft", true) else SetBarGroupField("minimumTimeLeft", false) end end,
									values = function(info)
										local d = GetBarGroupField("filterTimeLeft")
										local ds = string.format("%0d:%02d", math.floor(d / 60), d % 60)
										return { "Show if " .. ds .. " or more", "Show if less than " .. ds }
									end,
									style = "dropdown",
								},
							},
						},
						TimeFormatGroup = {
							type = "group", order = 50, name = "Time Format",  inline = true,
							args = {
								TimeFormat = {
									type = "select", order = 10, name = "  Options", width = "double",
									desc = "Select time format for bars and icons. There are separate settings for five ranges: " ..
										"more then an hour, 2 minutes to an hour, 1 minute to 2 minutes, 10 seconds to 1 minute, less than 10 seconds. " ..
										"These are illustrated by samples in the drop-down menu for each format option.",
									get = function(info) return GetBarGroupField("timeFormat") end,
									set = function(info, value) SetBarGroupField("timeFormat", value) end,
									values = function(info)
										local bg = GetBarGroupEntry()
										local s, c = bg.timeSpaces, bg.timeCase
										return GetTimeFormatList(s, c)
									end,
									style = "dropdown",
								},
								Space1 = { type = "description", name = "", order = 15, width = "half" },
								Spaces = {
									type = "toggle", order = 20, name = "Spaces", width = "half",
									desc = "Include spaces between values in time format.",
									get = function(info) return GetBarGroupField("timeSpaces") end,
									set = function(info, value) SetBarGroupField("timeSpaces", value) end,
								},
								Capitals = {
									type = "toggle", order = 30, name = "Uppercase",
									desc = "If checked, use uppercase H, M and S in time format, otherwise use lowercase.",
									get = function(info) return GetBarGroupField("timeCase") end,
									set = function(info, value) SetBarGroupField("timeCase", value) end,
								},
							},
						},
						EffectsGroup = {
							type = "group", order = 60, name = "Special Effects",  inline = true,
							args = {
								FadeEnable = {
									type = "toggle", order = 10, name = "Fade After Start",
									desc = "Enable fade after start (i.e., show at normal opacity then fade to background opacity).",
									get = function(info) return GetBarGroupField("fade") end,
									set = function(info, value) SetBarGroupField("fade", value) end,
								},
								FadeDelay = {
									type = "range", order = 20, name = "Fade Delay", min = 1, max = 300, step = 1,
									desc = "Set number of seconds before bar will fade.",
									disabled = function(info) return not GetBarGroupField("fade") end,
									get = function(info) return GetBarGroupField("fadeTime") end,
									set = function(info, value) SetBarGroupField("fadeTime", value) end,
								},
								FlashExpiring = {
									type = "toggle", order = 30, name = "Flash When Expiring",
									desc = "Enable flashing of expiring bars (if both fade and flash are enabled then flash will restore normal opacity).",
									get = function(info) return GetBarGroupField("flashExpiring") end,
									set = function(info, value) SetBarGroupField("flashExpiring", value) end,
								},
								FlashTime = {
									type = "range", order = 40, name = "Flash Time", min = 1, max = 300, step = 1,
									desc = "Set number of seconds before expiration that bar should start flashing.",
									disabled = function(info) return not GetBarGroupField("flashExpiring") end,
									get = function(info) return GetBarGroupField("flashTime") end,
									set = function(info, value) SetBarGroupField("flashTime", value) end,
								},
							},
						},
					},
				},
			},
		},
		Conditions = {
			type = "group", order = 30, name = "Conditions", childGroups = "tab",
			disabled = function(info) return InMode("BG") end,
			args = {
				SelectCondition = {
					type = "select", order = 1, name = "  Condition",
					get = function(info) return GetSelectedCondition() end,
					set = function(info, value) SetSelectedCondition(value) end,
					disabled = function(info) return NoCondition() end,
					values = function(info) return GetConditionList() end,
					style = "dropdown",
				},
				Space1 = { type = "description", name = "", order = 2, width = "half" },
				NewConditionButton = {
					type = "execute", order = 3, name = "New Condition",
					desc = "Create a new condition.",
					hidden = function(info) return enterNewCondition end,
					func = function(info) enterNewCondition, toggleNewCondition = true, true end,
				},
				NewConditionName = {
					type = "input", order = 4, name = " Enter Condition Name",
					desc = " Enter name of new condition.",
					hidden = function(info) return not enterNewCondition end,
					validate = function(info, n) if not n or (n == "") then return "Invalid name." else return true end end,
					confirm = function(info, value) return ConfirmNewCondition(value) end,
					get = function(info)
						enterNewCondition = toggleNewCondition
						if toggleNewCondition then toggleNewCondition = false end
						if not enterNewCondition then MOD:UpdateOptions() end
						return false
					end,
					set = function(info, value) enterNewCondition = false; CreateCondition(value) end,
				},
				CancelNewCondition = {
					type = "execute", order = 5, name = "Cancel", width = "half",
					desc = "Cancel creating a new condition.",
					hidden = function(info) return not enterNewCondition end,
					func = function(info) enterNewCondition, toggleNewCondition = false, false end,
				},
				DeleteCondition = {
					type = "execute", order = 8, name = "Delete Condition",
					desc = "Delete the selected condition.",
					hidden = function(info) return enterNewCondition end,
					func = function(info) PurgeCondition(); DeleteCondition() end,
					confirm = function(info) return 'DELETE CONDITION\nAre you sure you want to delete the "' .. GetConditionField("name") .. '" condition?' end,
				},
				GeneralTab = {
					type = "group", order = 20, name = "General",
					hidden = function(info) return NoCondition() end,
					args = {
						EnableGroup = {
							type = "group", order = 1, name = "Enable", inline = true,
							args = {
								EnableCondition = {
									type = "toggle", order = 1, name = "Enable Condition",
									desc = "If checked, the condition is enabled and its value is determined by evaluating the associated tests, " ..
										"otherwise the condition is disabled and its value is always false.",
									get = function(info) return GetConditionField("enabled") end,
									set = function(info, value) SetConditionField("enabled", value) end,
								},
								EnableNotification = {
									type = "toggle", order = 10, name = "Notification",
									desc = "If checked, this condition is suitable for a notification and will show up as an option for new notify bars.",
									get = function(info) return GetConditionField("notify") end,
									set = function(info, value) SetConditionField("notify", value) end,
								},
								EnableTooltips = {
									type = "toggle", order = 20, name = "Detailed Tooltip",
									desc = "If checked, tooltip for notifications based on this condition will include a detailed description.",
									get = function(info) return GetConditionField("tooltip") end,
									set = function(info, value) SetConditionField("tooltip", value) end,
								},
							},
						},
						RenameGroup = {
							type = "group", order = 5, name = "Rename Condition", inline = true,
							args = {
								Rename = {
									type = "input", order = 10, name = " New Name",
									validate = function(info, n) if not n or (n == "") then return "Invalid name." else return true end end,
									confirm = function(info, value) return ConfirmNewCondition(value) end,
									desc = "Enter new name for the condition.",
									get = function(info) return GetConditionField("name") end,
									set = function(info, value) RenameCondition(value) end,
								},
							},
						},
						SpellGroup = {
							type = "group", order = 10, name = "Associated Spell", inline = true,
							args = {
								Rename = {
									type = "input", order = 10, name = " Spell Name",
									desc = "Enter spell name (or numeric identifier) whose color and icon can be used by notification bars based on this condition.",
									get = function(info) return GetConditionField("associatedSpell") end,
									set = function(info, value)
										local n, t = value, tonumber(value)
										if t then n = GetSpellInfo(t) end
										ValidateSpellName(n)
										SetConditionField("associatedSpell", n)
									end,
								},
							},
						},
					},
				},
				TestsTab = {
					type = "group", order = 40, name = "Tests",
					hidden = function(info) return NoCondition() end,
					args = {
						SummaryGroup = {
							type = "group", order = 10, name = "Summary", inline = true,
							args = {
								ConditionDescription = {
									type = "description", order = 1, name = function(info) return GetConditionDescription() end,
								},
							},
						},
						PlayerStatusGroup = {
							type = "group", order = 30, name = "Player Status",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Player Status", "enable") end,
											set = function(info, value) SetTestField("Player Status", "enable", value) end,
										},
									},
								},
								CombatGroup = {
									type = "group", order = 10, name = "Combat", inline = true,
									args = {
										CheckCombat = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test the player's combat status.",
											get = function(info) return IsTestFieldOn("Player Status", "inCombat") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "inCombat", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "In Combat",
											desc = "If checked, player must be in combat.",
											disabled = function(info) return IsTestFieldOff("Player Status", "inCombat") end,
											get = function(info) return GetTestField("Player Status", "inCombat") == true end,
											set = function(info, value) SetTestField("Player Status", "inCombat", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Out Of Combat",
											desc = "If checked, player must be out of combat.",
											disabled = function(info) return IsTestFieldOff("Player Status", "inCombat") end,
											get = function(info) return GetTestField("Player Status", "inCombat") == false end,
											set = function(info, value) SetTestField("Player Status", "inCombat", false) end,
										},
									},
								},
								RestingGroup = {
									type = "group", order = 20, name = "Resting", inline = true,
									args = {
										CheckResting = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test if the player is resting.",
											get = function(info) return IsTestFieldOn("Player Status", "isResting") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "isResting", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "Is Resting",
											desc = "If checked, player must be resting (e.g., in an inn).",
											disabled = function(info) return IsTestFieldOff("Player Status", "isResting") end,
											get = function(info) return GetTestField("Player Status", "isResting") == true end,
											set = function(info, value) SetTestField("Player Status", "isResting", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Not Resting",
											desc = "If checked, player must not be resting.",
											disabled = function(info) return IsTestFieldOff("Player Status", "isResting") end,
											get = function(info) return GetTestField("Player Status", "isResting") == false end,
											set = function(info, value) SetTestField("Player Status", "isResting", false) end,
										},
									},
								},
								MountedGroup = {
									type = "group", order = 25, name = "Mounted", inline = true,
									args = {
										CheckMounted = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test if the player is mounted.",
											get = function(info) return IsTestFieldOn("Player Status", "isMounted") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "isMounted", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "Is Mounted",
											desc = "If checked, player must be mounted.",
											disabled = function(info) return IsTestFieldOff("Player Status", "isMounted") end,
											get = function(info) return GetTestField("Player Status", "isMounted") == true end,
											set = function(info, value) SetTestField("Player Status", "isMounted", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Not Mounted",
											desc = "If checked, player must not be mounted.",
											disabled = function(info) return IsTestFieldOff("Player Status", "isMounted") end,
											get = function(info) return GetTestField("Player Status", "isMounted") == false end,
											set = function(info, value) SetTestField("Player Status", "isMounted", false) end,
										},
									},
								},
								StealthedGroup = {
									type = "group", order = 30, name = "Stealthed", inline = true,
									args = {
										CheckStealthed = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test if the player is stealthed.",
											get = function(info) return IsTestFieldOn("Player Status", "isStealthed") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "isStealthed", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "Is Stealthed",
											desc = "If checked, player must be stealthed.",
											disabled = function(info) return IsTestFieldOff("Player Status", "isStealthed") end,
											get = function(info) return GetTestField("Player Status", "isStealthed") == true end,
											set = function(info, value) SetTestField("Player Status", "isStealthed", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Not Stealthed",
											desc = "If checked, player must not be stealthed.",
											disabled = function(info) return IsTestFieldOff("Player Status", "isStealthed") end,
											get = function(info) return GetTestField("Player Status", "isStealthed") == false end,
											set = function(info, value) SetTestField("Player Status", "isStealthed", false) end,
										},
									},
								},
								FishingGroup = {
									type = "group", order = 32, name = "Fishing", inline = true,
									args = {
										CheckFishing = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test if the player is fishing.",
											get = function(info) return IsTestFieldOn("Player Status", "isFishing") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "isFishing", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "Is Fishing",
											desc = "If checked, player must be ready to fish (i.e., trained with fishing pole equipped).",
											disabled = function(info) return IsTestFieldOff("Player Status", "isFishing") end,
											get = function(info) return GetTestField("Player Status", "isFishing") == true end,
											set = function(info, value) SetTestField("Player Status", "isFishing", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Not Fishing",
											desc = "If checked, player must not be ready to fish.",
											disabled = function(info) return IsTestFieldOff("Player Status", "isFishing") end,
											get = function(info) return GetTestField("Player Status", "isFishing") == false end,
											set = function(info, value) SetTestField("Player Status", "isFishing", false) end,
										},
									},
								},
								PetGroup = {
									type = "group", order = 35, name = "Pet", inline = true,
									args = {
										CheckStealthed = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test if the player has a pet.",
											get = function(info) return IsTestFieldOn("Player Status", "hasPet") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "hasPet", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "Has Pet",
											desc = "If checked, player must have a pet.",
											disabled = function(info) return IsTestFieldOff("Player Status", "hasPet") end,
											get = function(info) return GetTestField("Player Status", "hasPet") == true end,
											set = function(info, value) SetTestField("Player Status", "hasPet", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "No Pet",
											desc = "If checked, player must not have a pet",
											disabled = function(info) return IsTestFieldOff("Player Status", "hasPet") end,
											get = function(info) return GetTestField("Player Status", "hasPet") == false end,
											set = function(info, value) SetTestField("Player Status", "hasPet", false) end,
										},
									},
								},
								CheckLevelGroup = {
									type = "group", order = 40, name = "Level", inline = true,
									args = {
										CheckLevelEnable = {
											type = "toggle", order = 1, name = "Enable", width = "half",
											desc = "If checked, test the player's level.",
											get = function(info) return IsTestFieldOn("Player Status", "checkLevel") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkLevel", v) end,
										},
										CheckLevel = {
											type = "toggle", order = 2, name = "Minimum",
											desc = "If checked, player must be at least at this level, otherwise must be lower.",
											disabled = function(info) return IsTestFieldOff("Player Status", "checkLevel") end,
											get = function(info) return GetTestField("Player Status", "checkLevel") == true end,
											set = function(info, value) SetTestField("Player Status", "checkLevel", value) end,
										},
										LevelRange = {
											type = "range", order = 3, name = "", min = 1, max = 80, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkLevel") end,
											get = function(info) return GetTestField("Player Status", "level") end,
											set = function(info, value) SetTestField("Player Status", "level", value) end,
										},
									},
								},
								CheckHealthGroup = {
									type = "group", order = 50, name = "Health", inline = true, 
									args = {
										CheckHealthEnable = {
											type = "toggle", order = 1, name = "Enable", width = "half",
											desc = "If checked, test the player's health.",
											get = function(info) return IsTestFieldOn("Player Status", "checkHealth") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkHealth", v) end,
										},
										CheckHealth = {
											type = "toggle", order = 2, name = "Minimum",
											desc = "If checked, player's health must be at least this percentage, otherwise must be less.",
											disabled = function(info) return IsTestFieldOff("Player Status", "checkHealth") end,
											get = function(info) return GetTestField("Player Status", "checkHealth") == true end,
											set = function(info, value) SetTestField("Player Status", "checkHealth", value) end,
										},
										HealthRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkHealth") end,
											get = function(info) return GetTestField("Player Status", "minHealth") end,
											set = function(info, value) SetTestField("Player Status", "minHealth", value) end,
										},
									},
								},
								CheckPowerGroup = {
									type = "group", order = 70, name = "Mana/Rage/Energy/Runic Power", inline = true, 
									args = {
										CheckPowerEnable = {
											type = "toggle", order = 1, name = "Enable", width = "half",
											desc = "If checked, test the player's mana, rage, energy, or runic power.",
											get = function(info) return IsTestFieldOn("Player Status", "checkPower") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkPower", v) end,
										},
										CheckPower = {
											type = "toggle", order = 2, name = "Minimum",
											desc = "If checked, player's mana/rage/energy/runic power must be at least this percentage, otherwise must be less.",
											disabled = function(info) return IsTestFieldOff("Player Status", "checkPower") end,
											get = function(info) return GetTestField("Player Status", "checkPower") == true end,
											set = function(info, value) SetTestField("Player Status", "checkPower", value) end,
										},
										PowerRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkPower") end,
											get = function(info) return GetTestField("Player Status", "minPower") end,
											set = function(info, value) SetTestField("Player Status", "minPower", value) end,
										},
									},
								},
								CheckRunesGroup = {
									type = "group", order = 75, name = "Runes", inline = true,
									args = {
										CheckRunes = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test the player's rune availability.",
											get = function(info) return IsTestFieldOn("Player Status", "checkRunes") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkRunes", v) end,
										},
										DoBlood = {
											type = "toggle", order = 2, name = "Blood", width = "half",
											desc = "If checked, player must have at least one blood rune (or enough death runes).",
											disabled = function(info) return IsTestFieldOff("Player Status", "checkRunes") end,
											get = function(info) return GetTestField("Player Status", "needBlood") == true end,
											set = function(info, value) SetTestField("Player Status", "needBlood", value) end,
										},
										DoFrost = {
											type = "toggle", order = 3, name = "Frost", width = "half",
											desc = "If checked, player must have at least one frost rune (or enough death runes).",
											disabled = function(info) return IsTestFieldOff("Player Status", "checkRunes") end,
											get = function(info) return GetTestField("Player Status", "needFrost") == true end,
											set = function(info, value) SetTestField("Player Status", "needFrost", value) end,
										},
										DoUnholy = {
											type = "toggle", order = 4, name = "Unholy", width = "half",
											desc = "If checked, player must have at least one unholy rune (or enough death runes).",
											disabled = function(info) return IsTestFieldOff("Player Status", "checkRunes") end,
											get = function(info) return GetTestField("Player Status", "needUnholy") == true end,
											set = function(info, value) SetTestField("Player Status", "needUnholy", value) end,
										},
									},
								},
								CheckComboPointsGroup = {
									type = "group", order = 80, name = "Combo Points", inline = true, 
									args = {
										CheckComboPointsEnable = {
											type = "toggle", order = 1, name = "Enable", width = "half",
											desc = "If checked, test how many combo points the player has on the target.",
											get = function(info) return IsTestFieldOn("Player Status", "checkComboPoints") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkComboPoints", v) end,
										},
										CheckComboPoints = {
											type = "toggle", order = 2, name = "Minimum",
											desc = "If checked, player must have at least this many combo points, otherwise must be fewer.",
											disabled = function(info) return IsTestFieldOff("Player Status", "checkComboPoints") end,
											get = function(info) return GetTestField("Player Status", "checkComboPoints") == true end,
											set = function(info, value) SetTestField("Player Status", "checkComboPoints", value) end,
										},
										ComboPointsRange = {
											type = "range", order = 3, name = "", min = 1, max = 5, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkComboPoints") end,
											get = function(info) return GetTestField("Player Status", "minComboPoints") end,
											set = function(info, value) SetTestField("Player Status", "minComboPoints", value) end,
										},
									},
								},
								CheckWeaponGroup = {
									type = "group", order = 70, name = "Weapons", inline = true, 
									args = {
										CheckMainHandEnable = {
											type = "toggle", order = 1, name = "Mainhand", width = "half",
											desc = "If checked, test if the player has a mainhand weapon equipped with at least the specified item level.",
											get = function(info) return IsTestFieldOn("Player Status", "hasMainHand") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "hasMainHand", v) end,
										},
										MainHandLevel = {
											type = "range", order = 2, name = "", min = 1, max = 400, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "hasMainHand") end,
											get = function(info)
												local level = GetTestField("Player Status", "levelMainHand")
												if not level then level = 1; SetTestField("Player Status", "levelMainHand", level) end
												return level
											end,
											set = function(info, value) SetTestField("Player Status", "levelMainHand", value) end,
										},
										CheckOffHandEnable = {
											type = "toggle", order = 3, name = "Offhand", width = "half",
											desc = "If checked, test if the player has an offhand weapon equipped with at least the specified item level.",
											get = function(info) return IsTestFieldOn("Player Status", "hasOffHand") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "hasOffHand", v) end,
										},
										OffMainHandLevel = {
											type = "range", order = 4, name = "", min = 1, max = 400, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "hasOffHand") end,
											get = function(info)
												local level = GetTestField("Player Status", "levelOffHand")
												if not level then level = 1; SetTestField("Player Status", "levelOffHand", level) end
												return level
											end,
											set = function(info, value) SetTestField("Player Status", "levelOffHand", value) end,
										},
									},
								},
								CheckStanceGroup = {
									type = "group", order = 90, name = "Stance", inline = true, 
									args = {
										CheckStanceEnable = {
											type = "toggle", order = 10, name = "Enable", width = "half",
											desc = "If checked, test if the player is in a stance.",
											get = function(info) return IsTestFieldOn("Player Status", "checkStance") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkStance", v) end,
										},
										StanceEntry = {
											type = "input", order = 20, name = " Stance",
											desc = ' Enter the stance to check for (enter "none" to check for no stance).',
											get = function(info) return GetTestField("Player Status", "stance") end,
											set = function(info, value) SetTestFieldString("Player Status", "stance", value) end,
										},
									},
								},
								CheckTalentGroup = {
									type = "group", order = 95, name = "Talent", inline = true, 
									args = {
										CheckTalentsEnable = {
											type = "toggle", order = 10, name = "Enable", width = "half",
											desc = "If checked, test if points have been spent in one of the player's talents.",
											get = function(info) return IsTestFieldOn("Player Status", "checkTalent") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkTalent", v) end,
										},
										SelectTalent = {
											type = "select", order = 20, name = "  Talent",
											get = function(info) local v = GetTestField("Player Status", "talent"); if v then local t = MOD.talents[v]; if t then return t.select end end return nil end,
											set = function(info, value) SetTestField("Player Status", "talent", MOD.talentList[value]) end,
											values = function(info) return MOD.talentList end,
											style = "dropdown",
										},
									},
								},
							},
						},
						TargetStatusGroup = {
							type = "group", order = 35, name = "Target Status",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Target Status", "enable") end,
											set = function(info, value) SetTestField("Target Status", "enable", value) end,
										},
									},
								},
								EnemyGroup = {
									type = "group", order = 10, name = "Enemy", inline = true,
									args = {
										CheckEnemy = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test if the target is an enemy.",
											get = function(info) return IsTestFieldOn("Target Status", "isEnemy") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "isEnemy", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "Is Enemy",
											desc = "If checked, target must be an enemy.",
											disabled = function(info) return IsTestFieldOff("Target Status", "isEnemy") end,
											get = function(info) return GetTestField("Target Status", "isEnemy") == true end,
											set = function(info, value) SetTestField("Target Status", "isEnemy", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Not Enemy",
											desc = "If checked, target must not be an enemy.",
											disabled = function(info) return IsTestFieldOff("Target Status", "isEnemy") end,
											get = function(info) return GetTestField("Target Status", "isEnemy") == false end,
											set = function(info, value) SetTestField("Target Status", "isEnemy", false) end,
										},
									},
								},
								FriendGroup = {
									type = "group", order = 20, name = "Friendly", inline = true,
									args = {
										CheckFriend = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test if the target is friendly.",
											get = function(info) return IsTestFieldOn("Target Status", "isFriend") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "isFriend", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "Is Friendly",
											desc = "If checked, target must be friendly.",
											disabled = function(info) return IsTestFieldOff("Target Status", "isFriend") end,
											get = function(info) return GetTestField("Target Status", "isFriend") == true end,
											set = function(info, value) SetTestField("Target Status", "isFriend", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Not Friendly",
											desc = "If checked, target must not be friendly.",
											disabled = function(info) return IsTestFieldOff("Target Status", "isFriend") end,
											get = function(info) return GetTestField("Target Status", "isFriend") == false end,
											set = function(info, value) SetTestField("Target Status", "isFriend", false) end,
										},
									},
								},
								RangeGroup = {
									type = "group", order = 30, name = "Range", inline = true,
									args = {
										CheckRange = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test if the target is in range.",
											get = function(info) return IsTestFieldOn("Target Status", "inRange") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "inRange", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "In Range",
											desc = "If checked, target must be in range.",
											disabled = function(info) return IsTestFieldOff("Target Status", "inRange") end,
											get = function(info) return GetTestField("Target Status", "inRange") == true end,
											set = function(info, value) SetTestField("Target Status", "inRange", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Out Of Range",
											desc = "If checked, target must not be in range.",
											disabled = function(info) return IsTestFieldOff("Target Status", "inRange") end,
											get = function(info) return GetTestField("Target Status", "inRange") == false end,
											set = function(info, value) SetTestField("Target Status", "inRange", false) end,
										},
									},
								},
								CheckHealthGroup = {
									type = "group", order = 50, name = "Health", inline = true, 
									args = {
										CheckHealthEnable = {
											type = "toggle", order = 1, name = "Enable", width = "half",
											desc = "If checked, test the target's health.",
											get = function(info) return IsTestFieldOn("Target Status", "checkHealth") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "checkHealth", v) end,
										},
										CheckHealth = {
											type = "toggle", order = 2, name = "Minimum",
											desc = "If checked, target's health must be at least this percentage, otherwise must be less.",
											disabled = function(info) return IsTestFieldOff("Target Status", "checkHealth") end,
											get = function(info) return GetTestField("Target Status", "checkHealth") == true end,
											set = function(info, value) SetTestField("Target Status", "checkHealth", value) end,
										},
										HealthRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Target Status", "checkHealth") end,
											get = function(info) return GetTestField("Target Status", "minHealth") end,
											set = function(info, value) SetTestField("Target Status", "minHealth", value) end,
										},
									},
								},
								CheckPowerGroup = {
									type = "group", order = 70, name = "Mana/Rage/Energy/Runic Power", inline = true, 
									args = {
										CheckPowerEnable = {
											type = "toggle", order = 1, name = "Enable", width = "half",
											desc = "If checked, test the target's mana, rage, energy, or runic power.",
											get = function(info) return IsTestFieldOn("Target Status", "checkPower") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "checkPower", v) end,
										},
										CheckPower = {
											type = "toggle", order = 2, name = "Minimum",
											desc = "If checked, target's mana/rage/energy/runic power must be at least this percentage, otherwise must be less.",
											disabled = function(info) return IsTestFieldOff("Target Status", "checkPower") end,
											get = function(info) return GetTestField("Target Status", "checkPower") == true end,
											set = function(info, value) SetTestField("Target Status", "checkPower", value) end,
										},
										PowerRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Target Status", "checkPower") end,
											get = function(info) return GetTestField("Target Status", "minPower") end,
											set = function(info, value) SetTestField("Target Status", "minPower", value) end,
										},
									},
								},
							},
						},
						FocusStatusGroup = {
							type = "group", order = 40, name = "Focus Status",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Focus Status", "enable") end,
											set = function(info, value) SetTestField("Focus Status", "enable", value) end,
										},
									},
								},
								EnemyGroup = {
									type = "group", order = 10, name = "Enemy", inline = true,
									args = {
										CheckEnemy = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test if the focus is an enemy.",
											get = function(info) return IsTestFieldOn("Focus Status", "isEnemy") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "isEnemy", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "Is Enemy",
											desc = "If checked, focus must be an enemy.",
											disabled = function(info) return IsTestFieldOff("Focus Status", "isEnemy") end,
											get = function(info) return GetTestField("Focus Status", "isEnemy") == true end,
											set = function(info, value) SetTestField("Focus Status", "isEnemy", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Not Enemy",
											desc = "If checked, focus must not be an enemy.",
											disabled = function(info) return IsTestFieldOff("Focus Status", "isEnemy") end,
											get = function(info) return GetTestField("Focus Status", "isEnemy") == false end,
											set = function(info, value) SetTestField("Focus Status", "isEnemy", false) end,
										},
									},
								},
								FriendGroup = {
									type = "group", order = 20, name = "Friendly", inline = true,
									args = {
										CheckFriend = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test if the focus is friendly.",
											get = function(info) return IsTestFieldOn("Focus Status", "isFriend") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "isFriend", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "Is Friendly",
											desc = "If checked, focus must be friendly.",
											disabled = function(info) return IsTestFieldOff("Focus Status", "isFriend") end,
											get = function(info) return GetTestField("Focus Status", "isFriend") == true end,
											set = function(info, value) SetTestField("Focus Status", "isFriend", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Not Friendly",
											desc = "If checked, focus must not be friendly.",
											disabled = function(info) return IsTestFieldOff("Focus Status", "isFriend") end,
											get = function(info) return GetTestField("Focus Status", "isFriend") == false end,
											set = function(info, value) SetTestField("Focus Status", "isFriend", false) end,
										},
									},
								},
								RangeGroup = {
									type = "group", order = 30, name = "Range", inline = true,
									args = {
										CheckRange = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, test if the focus is in range.",
											get = function(info) return IsTestFieldOn("Focus Status", "inRange") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "inRange", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "In Range",
											desc = "If checked, focus must be in range.",
											disabled = function(info) return IsTestFieldOff("Focus Status", "inRange") end,
											get = function(info) return GetTestField("Focus Status", "inRange") == true end,
											set = function(info, value) SetTestField("Focus Status", "inRange", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Out Of Range",
											desc = "If checked, focus must not be in range.",
											disabled = function(info) return IsTestFieldOff("Focus Status", "inRange") end,
											get = function(info) return GetTestField("Focus Status", "inRange") == false end,
											set = function(info, value) SetTestField("Focus Status", "inRange", false) end,
										},
									},
								},
								CheckHealthGroup = {
									type = "group", order = 50, name = "Health", inline = true, 
									args = {
										CheckHealthEnable = {
											type = "toggle", order = 1, name = "Enable", width = "half",
											desc = "If checked, test the focus's health.",
											get = function(info) return IsTestFieldOn("Focus Status", "checkHealth") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "checkHealth", v) end,
										},
										CheckHealth = {
											type = "toggle", order = 2, name = "Minimum",
											desc = "If checked, focus's health must be at least this percentage, otherwise must be less.",
											disabled = function(info) return IsTestFieldOff("Focus Status", "checkHealth") end,
											get = function(info) return GetTestField("Focus Status", "checkHealth") == true end,
											set = function(info, value) SetTestField("Focus Status", "checkHealth", value) end,
										},
										HealthRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Focus Status", "checkHealth") end,
											get = function(info) return GetTestField("Focus Status", "minHealth") end,
											set = function(info, value) SetTestField("Focus Status", "minHealth", value) end,
										},
									},
								},
								CheckPowerGroup = {
									type = "group", order = 70, name = "Mana/Rage/Energy/Runic Power", inline = true, 
									args = {
										CheckPowerEnable = {
											type = "toggle", order = 1, name = "Enable", width = "half",
											desc = "If checked, test the focus's mana, rage, energy, or runic power.",
											get = function(info) return IsTestFieldOn("Focus Status", "checkPower") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "checkPower", v) end,
										},
										CheckPower = {
											type = "toggle", order = 2, name = "Minimum",
											desc = "If checked, focus's mana/rage/energy/runic power must be at least this percentage, otherwise must be less.",
											disabled = function(info) return IsTestFieldOff("Focus Status", "checkPower") end,
											get = function(info) return GetTestField("Focus Status", "checkPower") == true end,
											set = function(info, value) SetTestField("Focus Status", "checkPower", value) end,
										},
										PowerRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Focus Status", "checkPower") end,
											get = function(info) return GetTestField("Focus Status", "minPower") end,
											set = function(info, value) SetTestField("Focus Status", "minPower", value) end,
										},
									},
								},
							},
						},
						AllBuffsGroup = {
							type = "group", order = 45, name = "All Buffs",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("All Buffs", "enable") end,
											set = function(info, value) SetTestField("All Buffs", "enable", value) end,
										},
									},
								},
								AurasGroup = {
									type = "group", order = 2, name = "Buff List Entry", inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = " Buffs", width = "full",
											desc = " Enter comma-separated list of buffs.",
											get = function(info) return GetTestFieldSpellList("All Buffs", "auras") end,
											set = function(info, value) SetTestFieldSpellList("All Buffs", "auras", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = "Test Buffs", inline = true,
									args = {
										AllActive = {
											type = "toggle", order = 10, name = "All Active",
											desc = "If checked, test if all the buffs are active.",
											get = function(info) return GetTestField("All Buffs", "toggle") ~= true end,
											set = function(info, value) SetTestField("All Buffs", "toggle", Off)  end,
										},
										NotAllActive = {
											type = "toggle", order = 20, name = "Not All Active",
											desc = "If checked, test if any of the buffs are not active.",
											get = function(info) return GetTestField("All Buffs", "toggle") == true end,
											set = function(info, value) SetTestField("All Buffs", "toggle", true)  end,
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = "Who Has Buffs", inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = "On Player",
											desc = "If checked, test if player buffs.",
											get = function(info) return GetTestField("All Buffs", "unit") == "player" end,
											set = function(info, value) SetTestField("All Buffs", "unit", "player")  end,
										},
										TargetBuff = {
											type = "toggle", order = 20, name = "On Target",
											desc = "If checked, test if target buffs.",
											get = function(info) return GetTestField("All Buffs", "unit") == "target" end,
											set = function(info, value) SetTestField("All Buffs", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 30, name = "On Focus",
											desc = "If checked, test if focus buffs.",
											get = function(info) return GetTestField("All Buffs", "unit") == "focus" end,
											set = function(info, value) SetTestField("All Buffs", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = "Who Cast Buffs", inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = "Cast By Player",
											desc = "If checked, test if the buffs were cast by the player.",
											get = function(info) return GetTestField("All Buffs", "isMine") == true end,
											set = function(info, value) SetTestField("All Buffs", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = "Cast By Other",
											desc = "If checked, test if the buffs were cast by anyone other than the player.",
											get = function(info) return GetTestField("All Buffs", "isMine") == false end,
											set = function(info, value) SetTestField("All Buffs", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = "Cast By Anyone",
											desc = "If checked, buffs may be cast by anyone, including the player.",
											get = function(info) return IsOff(GetTestField("All Buffs", "isMine")) end,
											set = function(info, value) SetTestField("All Buffs", "isMine", Off) end,
										},
									},
								},
							},
						},
						AnyBuffsGroup = {
							type = "group", order = 46, name = "Any Buffs",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Any Buffs", "enable") end,
											set = function(info, value) SetTestField("Any Buffs", "enable", value) end,
										},
									},
								},
								AurasGroup = {
									type = "group", order = 5, name = "Buff List Entry", inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = " Buffs", width = "full",
											desc = " Enter comma-separated list of buffs.",
											get = function(info) return GetTestFieldSpellList("Any Buffs", "auras") end,
											set = function(info, value) SetTestFieldSpellList("Any Buffs", "auras", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = "Test Buffs", inline = true,
									args = {
										AllActive = {
											type = "toggle", order = 10, name = "Any Active",
											desc = "If checked, test if any of the buffs are active.",
											get = function(info) return GetTestField("Any Buffs", "toggle") ~= true end,
											set = function(info, value) SetTestField("Any Buffs", "toggle", Off)  end,
										},
										NotAllActive = {
											type = "toggle", order = 20, name = "None Active",
											desc = "If checked, test if none of the buffs are active.",
											get = function(info) return GetTestField("Any Buffs", "toggle") == true end,
											set = function(info, value) SetTestField("Any Buffs", "toggle", true)  end,
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = "Who Has Buffs", inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = "On Player",
											desc = "If checked, test if player buffs.",
											get = function(info) return GetTestField("Any Buffs", "unit") == "player" end,
											set = function(info, value) SetTestField("Any Buffs", "unit", "player")  end,
										},
										TargetBuff = {
											type = "toggle", order = 20, name = "On Target",
											desc = "If checked, test if target buffs.",
											get = function(info) return GetTestField("Any Buffs", "unit") == "target" end,
											set = function(info, value) SetTestField("Any Buffs", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 30, name = "On Focus",
											desc = "If checked, test if focus buffs.",
											get = function(info) return GetTestField("Any Buffs", "unit") == "focus" end,
											set = function(info, value) SetTestField("Any Buffs", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = "Who Cast Buffs", inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = "Cast By Player",
											desc = "If checked, test if the buffs were cast by the player.",
											get = function(info) return GetTestField("Any Buffs", "isMine") == true end,
											set = function(info, value) SetTestField("Any Buffs", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = "Cast By Other",
											desc = "If checked, test if the buffs were cast by anyone other than the player.",
											get = function(info) return GetTestField("Any Buffs", "isMine") == false end,
											set = function(info, value) SetTestField("Any Buffs", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = "Cast By Anyone",
											desc = "If checked, buffs may be cast by anyone, including the player.",
											get = function(info) return IsOff(GetTestField("Any Buffs", "isMine")) end,
											set = function(info, value) SetTestField("Any Buffs", "isMine", Off) end,
										},
									},
								},
							},
						},
						BuffTimeLeftGroup = {
							type = "group", order = 47, name = "Buff Time Left",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Buff Time Left", "enable") end,
											set = function(info, value) SetTestField("Buff Time Left", "enable", value) end,
										},
									},
								},
								AuraGroup = {
									type = "group", order = 5, name = "Buff Entry", inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = " Buff", width = "full",
											desc = " Enter the buff to be tested.",
											get = function(info) return GetTestField("Buff Time Left", "aura") end,
											set = function(info, value) SetTestFieldString("Buff Time Left", "aura", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 10, name = "Time Left", inline = true,
									args = {
										TimeLeftMinutes= {
											type = "range", order = 1, name = "Minutes", min = 0, max = 120, step = 1,
											desc = "Enter minutes for time left check.",
											get = function(info) local d = GetTestField("Buff Time Left", "timeLeft"); if d then return math.floor(d / 60) else return 0 end end,
											set = function(info, value) local d = GetTestField("Buff Time Left", "timeLeft"); if not d then d = 0 end; SetTestField("Buff Time Left", "timeLeft", (value * 60) + (d % 60)) end,
										},
										TimeLeftSeconds = {
											type = "range", order = 5, name = "Seconds", min = 0, max = 59, step = 1,
											desc = "Enter seconds for time left check.",
											get = function(info) local d = GetTestField("Buff Time Left", "timeLeft"); if d then return d % 60 else return 0 end end,
											set = function(info, value) local d = GetTestField("Buff Time Left", "timeLeft"); if not d then d = 0 end; SetTestField("Buff Time Left", "timeLeft", value + (60 * math.floor(d / 60))) end,
										},
										TimeLeftMinMax = {
											type = "select", order = 6, name = "  Time Left",
											get = function(info) if GetTestField("Buff Time Left", "toggle") == true then return 1 else return 2 end end,
											set = function(info, value) if value == 1 then SetTestField("Buff Time Left", "toggle", true) else SetTestField("Buff Time Left", "toggle", Off) end end,
											values = function(info)
												local d = GetTestField("Buff Time Left", "timeLeft")
												if not d then d = 0 end
												local ds = string.format("%0d:%02d", math.floor(d / 60), d % 60)
												return { "Less Than " .. ds, ds .. " Or More" }
											end,
											style = "dropdown",
										},
									},
								},
								UnitGroup = {
									type = "group", order = 20, name = "Who Has Buff", inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = "On Player",
											desc = "If checked, test if player buff.",
											get = function(info) return GetTestField("Buff Time Left", "unit") == "player" end,
											set = function(info, value) SetTestField("Buff Time Left", "unit", "player")  end,
										},
										TargetBuff = {
											type = "toggle", order = 20, name = "On Target",
											desc = "If checked, test if target buff.",
											get = function(info) return GetTestField("Buff Time Left", "unit") == "target" end,
											set = function(info, value) SetTestField("Buff Time Left", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 30, name = "On Focus",
											desc = "If checked, test if focus buff.",
											get = function(info) return GetTestField("Buff Time Left", "unit") == "focus" end,
											set = function(info, value) SetTestField("Buff Time Left", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 25, name = "Who Cast Buff", inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = "Cast By Player",
											desc = "If checked, test if the buff was cast by the player.",
											get = function(info) return GetTestField("Buff Time Left", "isMine") == true end,
											set = function(info, value) SetTestField("Buff Time Left", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = "Cast By Other",
											desc = "If checked, test if the buff was cast by anyone other than the player.",
											get = function(info) return GetTestField("Buff Time Left", "isMine") == false end,
											set = function(info, value) SetTestField("Buff Time Left", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = "Cast By Anyone",
											desc = "If checked, buff may be cast by anyone, including the player.",
											get = function(info) return IsOff(GetTestField("Buff Time Left", "isMine")) end,
											set = function(info, value) SetTestField("Buff Time Left", "isMine", Off) end,
										},
									},
								},
							},
						},
						BuffCountGroup = {
							type = "group", order = 48, name = "Buff Count",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Buff Count", "enable") end,
											set = function(info, value) SetTestField("Buff Count", "enable", value) end,
										},
									},
								},
								AuraGroup = {
									type = "group", order = 5, name = "Buff Entry", inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = " Buff", width = "full",
											desc = " Enter buff to test for stack count (if buff doesn't stack then its count is 0).",
											get = function(info) return GetTestField("Buff Count", "aura") end,
											set = function(info, value) SetTestFieldString("Buff Count", "aura", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 10, name = "Stack Count", inline = true,
									args = {
										CountValue = {
											type = "range", order = 1, name = "Count", min = 1, max = 10, step = 1,
											desc = "Enter value to compare with the buff stack count.",
											get = function(info) local d = GetTestField("Buff Count", "count"); if d then return d else return 1 end end,
											set = function(info, value) local d = GetTestField("Buff Count", "count"); if not d then d = 1 end; SetTestField("Buff Count", "count", value) end,
										},
										CountMinMax = {
											type = "select", order = 6, name = "  Comparison",
											get = function(info) if GetTestField("Buff Count", "toggle") == true then return 1 else return 2 end end,
											set = function(info, value) if value == 1 then SetTestField("Buff Count", "toggle", true) else SetTestField("Buff Count", "toggle", Off) end end,
											values = function(info)
												local d = GetTestField("Buff Count", "count")
												if not d then d = 1 end
												return { "Less Than " .. d, d .. " Or More" }
											end,
											style = "dropdown",
										},
									},
								},
								UnitGroup = {
									type = "group", order = 15, name = "Who Has Buff", inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = "On Player",
											desc = "If checked, test if player buff.",
											get = function(info) return GetTestField("Buff Count", "unit") == "player" end,
											set = function(info, value) SetTestField("Buff Count", "unit", "player")  end,
										},
										TargetBuff = {
											type = "toggle", order = 20, name = "On Target",
											desc = "If checked, test if target buff.",
											get = function(info) return GetTestField("Buff Count", "unit") == "target" end,
											set = function(info, value) SetTestField("Buff Count", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 30, name = "On Focus",
											desc = "If checked, test if focus buff.",
											get = function(info) return GetTestField("Buff Count", "unit") == "focus" end,
											set = function(info, value) SetTestField("Buff Count", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 20, name = "Who Cast Buff", inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = "Cast By Player",
											desc = "If checked, test if the buff was cast by the player.",
											get = function(info) return GetTestField("Buff Count", "isMine") == true end,
											set = function(info, value) SetTestField("Buff Count", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = "Cast By Other",
											desc = "If checked, test if the buff was cast by anyone other than the player.",
											get = function(info) return GetTestField("Buff Count", "isMine") == false end,
											set = function(info, value) SetTestField("Buff Count", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = "Cast By Anyone",
											desc = "If checked, buff may be cast by anyone, including the player.",
											get = function(info) return IsOff(GetTestField("Buff Count", "isMine")) end,
											set = function(info, value) SetTestField("Buff Count", "isMine", Off) end,
										},
									},
								},
							},
						},
						BuffTypeGroup = {
							type = "group", order = 49, name = "Buff Type",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Buff Type", "enable") end,
											set = function(info, value) SetTestField("Buff Type", "enable", value) end,
										},
									},
								},
								CheckBuffGroup = {
									type = "group", order = 10, name = "Player Has Buff", inline = true, args = {
										CheckMainhand = {
											type = "toggle", order = 2, name = "Mainhand",
											desc = "If checked, player must have a mainhand buff.",
											disabled = function(info) return IsTestFieldOff("Buff Type", "hasBuff") end,
											get = function(info)
												if not GetTestField("Buff Type", "hasBuff") then SetTestField("Buff Type", "hasBuff", "Mainhand") end
												return GetTestField("Buff Type", "hasBuff") == "Mainhand"
											end,
											set = function(info, value) if value then SetTestField("Buff Type", "hasBuff", "Mainhand") end end,
										},
										CheckOffhand = {
											type = "toggle", order = 3, name = "Offhand",
											desc = "If checked, player must have an offhand buff.",
											disabled = function(info) return IsTestFieldOff("Buff Type", "hasBuff") end,
											get = function(info) return GetTestField("Buff Type", "hasBuff") == "Offhand" end,
											set = function(info, value) if value then SetTestField("Buff Type", "hasBuff", "Offhand") end end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 20, name = "Result", inline = true,
									args = {
										CheckPresent = {
											type = "toggle", order = 2, name = "Present",
											desc = "If checked, true if player has a buff of the specified type.",
											get = function(info) return GetTestField("Buff Type", "toggle") ~= true end,
											set = function(info, value) SetTestField("Buff Type", "toggle", false) end,
										},
										CheckMissing = {
											type = "toggle", order = 3, name = "Missing",
											desc = "If checked, true if player does not have a buff of the specified type.",
											get = function(info) return GetTestField("Buff Type", "toggle") == true end,
											set = function(info, value) SetTestField("Buff Type", "toggle", true) end,
										},
									},
								},
							},
						},
						AllDebuffsGroup = {
							type = "group", order = 50, name = "All Debuffs",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("All Debuffs", "enable") end,
											set = function(info, value) SetTestField("All Debuffs", "enable", value) end,
										},
									},
								},
								AurasGroup = {
									type = "group", order = 2, name = "Debuff List Entry", inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = " Debuffs", width = "full",
											desc = " Enter comma-separated list of debuffs.",
											get = function(info) return GetTestFieldSpellList("All Debuffs", "auras") end,
											set = function(info, value) SetTestFieldSpellList("All Debuffs", "auras", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = "Test Debuffs", inline = true,
									args = {
										AllActive = {
											type = "toggle", order = 10, name = "All Active",
											desc = "If checked, test if all the debuffs are active.",
											get = function(info) return GetTestField("All Debuffs", "toggle") ~= true end,
											set = function(info, value) SetTestField("All Debuffs", "toggle", Off)  end,
										},
										NotAllActive = {
											type = "toggle", order = 20, name = "Not All Active",
											desc = "If checked, test if any of the debuffs are not active.",
											get = function(info) return GetTestField("All Debuffs", "toggle") == true end,
											set = function(info, value) SetTestField("All Debuffs", "toggle", true)  end,
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = "Who Has Debuffs", inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = "On Player",
											desc = "If checked, test if player debuffs.",
											get = function(info) return GetTestField("All Debuffs", "unit") == "player" end,
											set = function(info, value) SetTestField("All Debuffs", "unit", "player")  end,
										},
										TargetBuff = {
											type = "toggle", order = 20, name = "On Target",
											desc = "If checked, test if target debuffs.",
											get = function(info) return GetTestField("All Debuffs", "unit") == "target" end,
											set = function(info, value) SetTestField("All Debuffs", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 30, name = "On Focus",
											desc = "If checked, test if focus debuffs.",
											get = function(info) return GetTestField("All Debuffs", "unit") == "focus" end,
											set = function(info, value) SetTestField("All Debuffs", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = "Who Cast Debuffs", inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = "Cast By Player",
											desc = "If checked, test if the debuffs were cast by the player.",
											get = function(info) return GetTestField("All Debuffs", "isMine") == true end,
											set = function(info, value) SetTestField("All Debuffs", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = "Cast By Other",
											desc = "If checked, test if the debuffs were cast by anyone other than the player.",
											get = function(info) return GetTestField("All Debuffs", "isMine") == false end,
											set = function(info, value) SetTestField("All Debuffs", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = "Cast By Anyone",
											desc = "If checked, debuffs may be cast by anyone, including the player.",
											get = function(info) return IsOff(GetTestField("All Debuffs", "isMine")) end,
											set = function(info, value) SetTestField("All Debuffs", "isMine", Off) end,
										},
									},
								},
							},
						},
						AnyDebuffsGroup = {
							type = "group", order = 51, name = "Any Debuffs",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Any Debuffs", "enable") end,
											set = function(info, value) SetTestField("Any Debuffs", "enable", value) end,
										},
									},
								},
								AurasGroup = {
									type = "group", order = 2, name = "Debuff List Entry", inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = " Debuffs", width = "full",
											desc = " Enter comma-separated list of debuffs.",
											get = function(info) return GetTestFieldSpellList("Any Debuffs", "auras") end,
											set = function(info, value) SetTestFieldSpellList("Any Debuffs", "auras", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = "Test Debuff", inline = true,
									args = {
										AllActive = {
											type = "toggle", order = 10, name = "Any Active",
											desc = "If checked, test if any of the debuffs are active.",
											get = function(info) return GetTestField("Any Debuffs", "toggle") ~= true end,
											set = function(info, value) SetTestField("Any Debuffs", "toggle", Off)  end,
										},
										NotAllActive = {
											type = "toggle", order = 20, name = "None Active",
											desc = "If checked, test if none of the debuffs are active.",
											get = function(info) return GetTestField("Any Debuffs", "toggle") == true end,
											set = function(info, value) SetTestField("Any Debuffs", "toggle", true)  end,
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = "Who Has Debuff", inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = "On Player",
											desc = "If checked, test if player debuffs.",
											get = function(info) return GetTestField("Any Debuffs", "unit") == "player" end,
											set = function(info, value) SetTestField("Any Debuffs", "unit", "player")  end,
										},
										TargetBuff = {
											type = "toggle", order = 20, name = "On Target",
											desc = "If checked, test if target debuffs.",
											get = function(info) return GetTestField("Any Debuffs", "unit") == "target" end,
											set = function(info, value) SetTestField("Any Debuffs", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 30, name = "On Focus",
											desc = "If checked, test if focus debuffs.",
											get = function(info) return GetTestField("Any Debuffs", "unit") == "focus" end,
											set = function(info, value) SetTestField("Any Debuffs", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = "Who Cast Debuff", inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = "Cast By Player",
											desc = "If checked, test if the debuffs were cast by the player.",
											get = function(info) return GetTestField("Any Debuffs", "isMine") == true end,
											set = function(info, value) SetTestField("Any Debuffs", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = "Cast By Other",
											desc = "If checked, test if the debuffs were cast by anyone other than the player.",
											get = function(info) return GetTestField("Any Debuffs", "isMine") == false end,
											set = function(info, value) SetTestField("Any Debuffs", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = "Cast By Anyone",
											desc = "If checked, debuffs may be cast by anyone, including the player.",
											get = function(info) return IsOff(GetTestField("Any Debuffs", "isMine")) end,
											set = function(info, value) SetTestField("Any Debuffs", "isMine", Off) end,
										},
									},
								},
							},
						},
						DebuffTimeLeftGroup = {
							type = "group", order = 52, name = "Debuff Time Left",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Debuff Time Left", "enable") end,
											set = function(info, value) SetTestField("Debuff Time Left", "enable", value) end,
										},
									},
								},
								AuraGroup = {
									type = "group", order = 2, name = "Debuff Entry", inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = " Debuff", width = "full",
											desc = " Enter the debuff to be tested.",
											get = function(info) return GetTestField("Debuff Time Left", "aura") end,
											set = function(info, value) SetTestFieldString("Debuff Time Left", "aura", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = "Time Left", inline = true,
									args = {
										TimeLeftMinutes= {
											type = "range", order = 1, name = "Minutes", min = 0, max = 120, step = 1,
											desc = "Enter minutes for time left check.",
											get = function(info) local d = GetTestField("Debuff Time Left", "timeLeft"); if d then return math.floor(d / 60) else return 0 end end,
											set = function(info, value) local d = GetTestField("Debuff Time Left", "timeLeft"); if not d then d = 0 end; SetTestField("Debuff Time Left", "timeLeft", (value * 60) + (d % 60)) end,
										},
										TimeLeftSeconds = {
											type = "range", order = 5, name = "Seconds", min = 0, max = 59, step = 1,
											desc = "Enter seconds for time left check.",
											get = function(info) local d = GetTestField("Debuff Time Left", "timeLeft"); if d then return d % 60 else return 0 end end,
											set = function(info, value) local d = GetTestField("Debuff Time Left", "timeLeft"); if not d then d = 0 end; SetTestField("Debuff Time Left", "timeLeft", value + (60 * math.floor(d / 60))) end,
										},
										TimeLeftMinMax = {
											type = "select", order = 6, name = "  Time Left",
											get = function(info) if GetTestField("Debuff Time Left", "toggle") == true then return 1 else return 2 end end,
											set = function(info, value) if value == 1 then SetTestField("Debuff Time Left", "toggle", true) else SetTestField("Debuff Time Left", "toggle", Off) end end,
											values = function(info)
												local d = GetTestField("Debuff Time Left", "timeLeft")
												if not d then d = 0 end
												local ds = string.format("%0d:%02d", math.floor(d / 60), d % 60)
												return { "Less Than " .. ds, ds .. " Or More" }
											end,
											style = "dropdown",
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = "Who Has Debuff", inline = true,
									args = {
										PlayerDebuff = {
											type = "toggle", order = 10, name = "On Player",
											desc = "If checked, test if player debuff.",
											get = function(info) return GetTestField("Debuff Time Left", "unit") == "player" end,
											set = function(info, value) SetTestField("Debuff Time Left", "unit", "player")  end,
										},
										TargetDebuff = {
											type = "toggle", order = 20, name = "On Target",
											desc = "If checked, test if target debuff.",
											get = function(info) return GetTestField("Debuff Time Left", "unit") == "target" end,
											set = function(info, value) SetTestField("Debuff Time Left", "unit", "target")  end,
										},
										FocusDebuff = {
											type = "toggle", order = 30, name = "On Focus",
											desc = "If checked, test if focus debuff.",
											get = function(info) return GetTestField("Debuff Time Left", "unit") == "focus" end,
											set = function(info, value) SetTestField("Debuff Time Left", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = "Who Cast Debuff", inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = "Cast By Player",
											desc = "If checked, test if the debuff was cast by the player.",
											get = function(info) return GetTestField("Debuff Time Left", "isMine") == true end,
											set = function(info, value) SetTestField("Debuff Time Left", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = "Cast By Other",
											desc = "If checked, test if the debuff was cast by anyone other than the player.",
											get = function(info) return GetTestField("Debuff Time Left", "isMine") == false end,
											set = function(info, value) SetTestField("Debuff Time Left", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = "Cast By Anyone",
											desc = "If checked, debuff may be cast by anyone, including the player.",
											get = function(info) return IsOff(GetTestField("Debuff Time Left", "isMine")) end,
											set = function(info, value) SetTestField("Debuff Time Left", "isMine", Off) end,
										},
									},
								},
							},
						},
						DebuffCountGroup = {
							type = "group", order = 53, name = "Debuff Count",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Debuff Count", "enable") end,
											set = function(info, value) SetTestField("Debuff Count", "enable", value) end,
										},
									},
								},
								AuraGroup = {
									type = "group", order = 2, name = "Debuff Entry", inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = " Debuff", width = "full",
											desc = " Enter debuff to test for stack count (if debuff doesn't stack then its count is 0).",
											get = function(info) return GetTestField("Debuff Count", "aura") end,
											set = function(info, value) SetTestFieldString("Debuff Count", "aura", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = "Stack Count", inline = true,
									args = {
										CountValue = {
											type = "range", order = 1, name = "Count", min = 1, max = 10, step = 1,
											desc = "Enter value to compare with the debuff stack count.",
											get = function(info) local d = GetTestField("Debuff Count", "count"); if d then return d else return 1 end end,
											set = function(info, value) local d = GetTestField("Debuff Count", "count"); if not d then d = 1 end; SetTestField("Debuff Count", "count", value) end,
										},
										CountMinMax = {
											type = "select", order = 6, name = "  Comparison",
											get = function(info) if GetTestField("Debuff Count", "toggle") == true then return 1 else return 2 end end,
											set = function(info, value) if value == 1 then SetTestField("Debuff Count", "toggle", true) else SetTestField("Debuff Count", "toggle", Off) end end,
											values = function(info)
												local d = GetTestField("Debuff Count", "count")
												if not d then d = 1 end
												return { "Less Than " .. d, d .. " Or More" }
											end,
											style = "dropdown",
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = "Who Has Debuff", inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = "On Player",
											desc = "If checked, test if player debuff.",
											get = function(info) return GetTestField("Debuff Count", "unit") == "player" end,
											set = function(info, value) SetTestField("Debuff Count", "unit", "player")  end,
										},
										TargetBuff = {
											type = "toggle", order = 20, name = "On Target",
											desc = "If checked, test if target debuff.",
											get = function(info) return GetTestField("Debuff Count", "unit") == "target" end,
											set = function(info, value) SetTestField("Debuff Count", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 30, name = "On Focus",
											desc = "If checked, test if focus debuff.",
											get = function(info) return GetTestField("Debuff Count", "unit") == "focus" end,
											set = function(info, value) SetTestField("Debuff Count", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = "Who Cast Debuff", inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = "Cast By Player",
											desc = "If checked, test if the debuff was cast by the player.",
											get = function(info) return GetTestField("Debuff Count", "isMine") == true end,
											set = function(info, value) SetTestField("Debuff Count", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = "Cast By Other",
											desc = "If checked, test if the debuff was cast by anyone other than the player.",
											get = function(info) return GetTestField("Debuff Count", "isMine") == false end,
											set = function(info, value) SetTestField("Debuff Count", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = "Cast By Anyone",
											desc = "If checked, debuff may be cast by anyone, including the player.",
											get = function(info) return IsOff(GetTestField("Debuff Count", "isMine")) end,
											set = function(info, value) SetTestField("Debuff Count", "isMine", Off) end,
										},
									},
								},
							},
						},
						DebuffTypeGroup = {
							type = "group", order = 54, name = "Debuff Type",
							args = {
								EnableTestGroup = {
									type = "group", order = 10, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Debuff Type", "enable") end,
											set = function(info, value) SetTestField("Debuff Type", "enable", value) end,
										},
									},
								},
								CheckDebuffGroup = {
									type = "group", order = 20, name = "Player Has Debuff", inline = true, args = {
										CheckPoisonDebuff = {
											type = "toggle", order = 2, name = "Poison", width = "half",
											desc = "If checked, player must have a poison debuff.",
											get = function(info)
												if not GetTestField("Debuff Type", "hasDebuff") then SetTestField("Debuff Type", "hasDebuff", "Poison") end
												return GetTestField("Debuff Type", "hasDebuff") == "Poison"
												end,
											set = function(info, value) if value then SetTestField("Debuff Type", "hasDebuff", "Poison") end end,
										},
										CheckDiseaseDebuff = {
											type = "toggle", order = 3, name = "Disease", width = "half",
											desc = "If checked, player must have a disease debuff.",
											get = function(info) return GetTestField("Debuff Type", "hasDebuff") == "Disease" end,
											set = function(info, value) if value then SetTestField("Debuff Type", "hasDebuff", "Disease") end end,
										},
										CheckCurseDebuff = {
											type = "toggle", order = 4, name = "Curse", width = "half",
											desc = "If checked, player must have a curse debuff.",
											get = function(info) return GetTestField("Debuff Type", "hasDebuff") == "Curse" end,
											set = function(info, value) if value then SetTestField("Debuff Type", "hasDebuff", "Curse") end end,
										},
										CheckMagicDebuff = {
											type = "toggle", order = 5, name = "Magic", width = "half",
											desc = "If checked, player must have a magic debuff.",
											get = function(info) return GetTestField("Debuff Type", "hasDebuff") == "Magic" end,
											set = function(info, value) if value then SetTestField("Debuff Type", "hasDebuff", "Magic") end end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 30, name = "Result", inline = true,
									args = {
										CheckPresent = {
											type = "toggle", order = 2, name = "Present", 
											desc = "If checked, true if player has a debuff of the specified type.",
											get = function(info) return GetTestField("Debuff Type", "toggle") ~= true end,
											set = function(info, value) SetTestField("Debuff Type", "toggle", false) end,
										},
										CheckMissing = {
											type = "toggle", order = 3, name = "Missing", 
											desc = "If checked, true if player does not have a debuff of the specified type.",
											get = function(info) return GetTestField("Debuff Type", "toggle") == true end,
											set = function(info, value) SetTestField("Debuff Type", "toggle", true) end,
										},
									},
								},
							},
						},
						AllCooldownsGroup = {
							type = "group", order = 56, name = "All Cooldowns",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("All Cooldowns", "enable") end,
											set = function(info, value) SetTestField("All Cooldowns", "enable", value) end,
										},
									},
								},
								SpellsGroup = {
									type = "group", order = 2, name = "Spell List Entry", inline = true,
									args = {
										SpellList = {
											type = "input", order = 1, name = " Spells", width = "full",
											desc = " Enter comma-separated list of spells.",
											get = function(info) return GetTestFieldSpellList("All Cooldowns", "spells") end,
											set = function(info, value) SetTestFieldSpellList("All Cooldowns", "spells", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 10, name = "Time Left", inline = true,
									args = {
										TimeLeftSeconds = {
											type = "range", order = 5, name = "Seconds", min = 0, max = 5, step = 0.1,
											desc = "Enter seconds for cooldown time left check.",
											get = function(info) local d = GetTestField("All Cooldowns", "timeLeft"); if d then return d else return 0 end end,
											set = function(info, value) local d = GetTestField("All Cooldowns", "timeLeft"); if not d then d = 0 end; SetTestField("All Cooldowns", "timeLeft", value) end,
										},
										TimeLeftMinMax = {
											type = "select", order = 6, name = "  Time Left",
											get = function(info) if GetTestField("All Cooldowns", "toggle") == true then return 1 else return 2 end end,
											set = function(info, value) if value == 1 then SetTestField("All Cooldowns", "toggle", true) else SetTestField("All Cooldowns", "toggle", Off) end end,
											values = function(info)
												local d = GetTestField("All Cooldowns", "timeLeft")
												if not d then d = 0 end
												local ds = string.format("%0.1f", d)
												return { "Less Than " .. ds .. " Seconds", ds .. " Seconds Or More" }
											end,
											style = "dropdown",
										},
									},
								},
							},
						},
						SpellReadyGroup = {
							type = "group", order = 60, name = "Spell Ready",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Spell Ready", "enable") end,
											set = function(info, value) SetTestField("Spell Ready", "enable", value) end,
										},
										EnableGCD = {
											type = "toggle", order = 2, name = "Global Cooldown",
											desc = "If checked, test if spell is on global cooldown, otherwise ignore global cooldown.",
											get = function(info) return GetTestField("Spell Ready", "gcd") end,
											set = function(info, value) SetTestField("Spell Ready", "gcd", value) end,
										},
										IsUsable = {
											type = "toggle", order = 2, name = "Is Usable",
											desc = "If checked, test if spell is usable (i.e., enough mana, reagents, etc.).",
											get = function(info) return not GetTestField("Spell Ready", "notUsable") end,
											set = function(info, value) SetTestField("Spell Ready", "notUsable", not value) end,
										},
									},
								},
								SpellGroup = {
									type = "group", order = 5, name = "Spell Entry", inline = true,
									args = {
										SpellName = {
											type = "input", order = 1, name = " Spell", width = "full",
											desc = " Enter spell to test if ready to be cast.",
											get = function(info) return GetTestField("Spell Ready", "spell") end,
											set = function(info, value) SetTestFieldString("Spell Ready", "spell", value) end,
										},
									},
								},
								RangeGroup = {
									type = "group", order = 10, name = "Range Check", inline = true,
									args = {
										CheckCombat = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, also test if the target is in range of the spell.",
											get = function(info) return IsTestFieldOn("Spell Ready", "inRange") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Spell Ready", "inRange", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = "Target In Range",
											desc = "If checked, target must be in range.",
											disabled = function(info) return IsTestFieldOff("Spell Ready", "inRange") end,
											get = function(info) return GetTestField("Spell Ready", "inRange") == true end,
											set = function(info, value) SetTestField("Spell Ready", "inRange", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Target Out Of Range",
											desc = "If checked, target must be out of range.",
											disabled = function(info) return IsTestFieldOff("Spell Ready", "inRange") end,
											get = function(info) return GetTestField("Spell Ready", "inRange") == false end,
											set = function(info, value) SetTestField("Spell Ready", "inRange", false) end,
										},
									},
								},
							},
						},
						SpellCastingGroup = {
							type = "group", order = 65, name = "Spell Casting",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Spell Casting", "enable") end,
											set = function(info, value) SetTestField("Spell Casting", "enable", value) end,
										},
									},
								},
								SpellGroup = {
									type = "group", order = 5, name = "Spell Entry", inline = true,
									args = {
										SpellName = {
											type = "input", order = 1, name = " Spell", width = "full",
											desc = " Enter spell to test if being cast or channeled.",
											get = function(info) return GetTestField("Spell Casting", "spell") end,
											set = function(info, value) SetTestFieldString("Spell Casting", "spell", value) end,
										},
									},
								},
								UnitGroup = {
									type = "group", order = 20, name = "Who's Casting", inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = "Cast By Player",
											desc = "If checked, test player casting.",
											get = function(info) return GetTestField("Spell Casting", "unit") == "player" end,
											set = function(info, value) SetTestField("Spell Casting", "unit", "player")  end,
										},
										TargetBuff = {
											type = "toggle", order = 20, name = "Cast By Target",
											desc = "If checked, test target casting.",
											get = function(info) return GetTestField("Spell Casting", "unit") == "target" end,
											set = function(info, value) SetTestField("Spell Casting", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 30, name = "Cast By Focus",
											desc = "If checked, test focus casting.",
											get = function(info) return GetTestField("Spell Casting", "unit") == "focus" end,
											set = function(info, value) SetTestField("Spell Casting", "unit", "focus")  end,
										},
									},
								},
							},
						},
						ItemReadyGroup = {
							type = "group", order = 70, name = "Item Ready",
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = "Enable Test", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, enable this test.",
											get = function(info) return GetTestField("Item Ready", "enable") end,
											set = function(info, value) SetTestField("Item Ready", "enable", value) end,
										},
									},
								},
								ItemGroup = {
									type = "group", order = 2, name = "Item Entry", inline = true,
									args = {
										ItemName = {
											type = "input", order = 1, name = " Item", width = "full",
											desc = " Enter item to test if ready to be used.",
											get = function(info) return GetTestField("Item Ready", "item") end,
											set = function(info, value) SetTestFieldString("Item Ready", "item", value) end,
										},
									},
								},
								ReadyOrNot = {
									type = "group", order = 10, name = "Ready Check", inline = true,
									args = {
										DoTrue = {
											type = "toggle", order = 2, name = "Item Is Ready",
											desc = "If checked, item must be ready.",
											get = function(info) return GetTestField("Item Ready", "toggle") ~= true end,
											set = function(info, value) SetTestField("Item Ready", "toggle", Off) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = "Item Is Not Ready",
											desc = "If checked, item must not be ready.",
											get = function(info) return GetTestField("Item Ready", "toggle") == true end,
											set = function(info, value) SetTestField("Item Ready", "toggle", true) end,
										},
									},
								},
							},
						},
						DependenciesGroup = {
							type = "group", order = 95, name = "Result",
							args = {
								SelectDependencies = {
									type = "group", order = 10, name = "Dependencies", inline = true,
									args = {
										intro = {
											type = "description", order = 1,
											name = "Dependencies are other conditions that must evaluate to specific results for this condition's " ..
												"result to be true. " ..
												"For example, spell rotations might depend on conditions that test target status and if you are in combat. " ..
												"Note all other tests are evaluated for all conditions before checking dependencies.\n",
										},
										SelectCondition = {
											type = "select", order = 10, name = "  Condition",
											get = function(info) return GetSelectedDependency() end,
											set = function(info, value) selectedDependency = value end,
											values = function(info) return GetDependenciesList() end,
											style = "dropdown",
										},
										TrueConditionButton = {
											type = "execute", order = 20, name = "True", width = "half",
											desc = "Condition is true only if the selected condition evaluates to true.",
											func = function(info) SetDependency(GetDependenciesList()[selectedDependency], true) end,
										},
										FalseConditionButton = {
											type = "execute", order = 30, name = "False", width = "half",
											desc = "Condition is true only if the selected condition evaluates to false.",
											func = function(info) SetDependency(GetDependenciesList()[selectedDependency], false) end,
										},
										NoConditionButton = {
											type = "execute", order = 40, name = "Delete", width = "half",
											hidden = function(info) return not CheckDependency(GetDependenciesList()[selectedDependency]) end,
											desc = "Delete this condition from the dependencies list.",
											func = function(info) SetDependency(GetDependenciesList()[selectedDependency], nil) end,
										},
									},
								},
								ToggleResult = {
									type = "group", order = 20, name = "Toggle Result", inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = "Enable",
											desc = "If checked, the result for this condition, after evaluating all tests and checking dependencies, is toggled.",
											get = function(info) return GetConditionField("toggleResult") end,
											set = function(info, value) SetConditionField("toggleResult", value) end,
										},
									},
								},
							},
						},
					},
				},
			},
		},
		Highlights = {
			type = "group", order = 40, name = "Highlights", 
			disabled = function(info) return InMode() end,
			args = {
				intro = {
					type = "description", order = 1,
					name = "If you are using Bartender4, Dominos or Macaroon then you can enable button highlights " ..
					"to indicate active buffs and debuffs that you cast " ..
					"(with flashing to indicate they are expiring). " ..
					"One highlight can be shown at a time and the priority is those on target first, focus second, and yourself last. " ..
					"You can also show cooldown counters on buttons for spells and items (disabled by default if OmniCC is loaded).\n",
				},
				PlayerGroup = {
					type = "group", order = 20, name = "Player", inline = true, 
					args = {
						PlayerBuffHighlightsEnable = {
							type = "toggle", order = 4, name = "Highlight Buffs",
							desc = "Enable player buff highlights on buttons.",
							get = function(info) return MOD.db.global.PlayerBuffHighlights end,
							set = function(info, value) MOD.db.global.PlayerBuffHighlights = value end,
						},
						PlayerBuffColor = {
							type = "color", order = 5, name = "Buff Color", hasAlpha = false,
							desc = "Set player buff highlight color.",
							get = function(info)
								local t = MOD.db.global.PlayerBuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.PlayerBuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
						PlayerDebuffHighlightsEnable = {
							type = "toggle", order = 23, name = "Highlight Debuffs",
							desc = "Enable player debuff highlights on buttons.",
							get = function(info) return MOD.db.global.PlayerDebuffHighlights end,
							set = function(info, value) MOD.db.global.PlayerDebuffHighlights = value end,
						},
						PlayerDebuffColor = {
							type = "color", order = 24, name = "Debuff Color", hasAlpha = false,
							desc = "Set player debuff highlight color.",
							get = function(info)
								local t = MOD.db.global.PlayerDebuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.PlayerDebuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
					},
				},
				TargetBuffs = {
					type = "group", order = 30, name = "Target", inline = true,
					args = {
						TargetBuffHighlightsEnable = {
							type = "toggle", order = 3, name = "Highlight Buffs",
							desc = "Enable target buff highlights on buttons.",
							get = function(info) return MOD.db.global.TargetBuffHighlights end,
							set = function(info, value) MOD.db.global.TargetBuffHighlights = value end,
						},
						TargetBuffColor = {
							type = "color", order = 4, name = "Buff Color", hasAlpha = false,
							desc = "Set target buff highlight color.",
							get = function(info)
								local t = MOD.db.global.TargetBuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.TargetBuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
						TargetDebuffHighlightsEnable = {
							type = "toggle", order = 23, name = "Highlight Debuffs",
							desc = "Enable target debuff highlights on buttons.",
							get = function(info) return MOD.db.global.TargetDebuffHighlights end,
							set = function(info, value) MOD.db.global.TargetDebuffHighlights = value end,
						},
						TargetDebuffColor = {
							type = "color", order = 24, name = "Debuff Color", hasAlpha = false,
							desc = "Set target debuff highlight color.",
							get = function(info)
								local t = MOD.db.global.TargetDebuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.TargetDebuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
					},
				},
				FocusBuffs = {
					type = "group", order = 40, name = "Focus", inline = true,
					args = {
						FocusBuffHighlightsEnable = {
							type = "toggle", order = 3, name = "Highlight Buffs",
							desc = "Enable focus buff highlights on buttons.",
							get = function(info) return MOD.db.global.FocusBuffHighlights end,
							set = function(info, value) MOD.db.global.FocusBuffHighlights = value end,
						},
						FocusBuffColor = {
							type = "color", order = 4, name = "Buff Color", hasAlpha = false,
							desc = "Set focus buff highlight color.",
							get = function(info)
								local t = MOD.db.global.FocusBuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.FocusBuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
						FocusDebuffHighlightsEnable = {
							type = "toggle", order = 23, name = "Highlight Debuffs",
							desc = "Enable focus debuff highlights on buttons.",
							get = function(info) return MOD.db.global.FocusDebuffHighlights end,
							set = function(info, value) MOD.db.global.FocusDebuffHighlights = value end,
						},
						FocusDebuffColor = {
							type = "color", order = 24, name = "Debuff Color", hasAlpha = false,
							desc = "Set focus debuff highlight color.",
							get = function(info)
								local t = MOD.db.global.FocusDebuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.FocusDebuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
					},
				},
				Flashing = {
					type = "group", order = 80, name = "Options",  inline = true,
					args = {
						ButtonFacade = {
							type = "toggle", order = 1, name = "Use ButtonFacade",
							desc = "Enable setting highlights through ButtonFacade if it is loaded. This is not ideal for every BF skin which is why it can be disabled. Try both enabled and disabled settings and see which works best.",
							disabled = function(info) return not LibStub("LibButtonFacade", true) end,
							get = function(info) return MOD.db.global.ButtonFacade end,
							set = function(info, value) MOD.db.global.ButtonFacade = value end,
						},
						FlashExpiring = {
							type = "toggle", order = 2, name = "Flash When Expiring",
							desc = "Enable highlight flashing for expiring buffs and debuffs.",
							get = function(info) return MOD.db.global.FlashExpiring end,
							set = function(info, value) MOD.db.global.FlashExpiring = value end,
						},
						FlashTime = {
							type = "range", order = 3, name = "Seconds", min = 1, max = 30, step = 1,
							desc = "Set number of seconds to flash before buffs and debuffs expire.",
							disabled = function(info) return not MOD.db.global.FlashExpiring end,
							get = function(info) return MOD.db.global.FlashTime end,
							set = function(info, value) MOD.db.global.FlashTime = value end,
						},
					},
				},
				Cooldowns = {
					type = "group", order = 90, name = "Cooldowns",  inline = true,
					args = {
						CooldownText = {
							type = "toggle", order = 10, name = "Cooldown Time",
							desc = "Show cooldown time remaining on buttons.",
							get = function(info) return MOD.db.global.CooldownText end,
							set = function(info, value) MOD.db.global.CooldownText = value end,
						},
						CooldownFontSize = {
							type = "range", order = 20, name = "Font Size", min = 5, max = 30, step = 1,
							desc = "Set font size for cooldown time text overlay.",
							disabled = function(info) return not MOD.db.global.CooldownText end,
							get = function(info) return MOD.db.global.CooldownFsize end,
							set = function(info, value) MOD.db.global.CooldownFsize = value end,
						},
						CooldownFont = {
							type = "select", order = 30, name = "  Font", 
							desc = "Select font for cooldown time text overlay.",
							disabled = function(info) return not MOD.db.global.CooldownText end,
							dialogControl = 'LSM30_Font',
							values = AceGUIWidgetLSMlists.font,
							get = function(info) return MOD.db.global.CooldownFont end,
							set = function(info, value) MOD.db.global.CooldownFont = value end,
						},
						TimeFormatGroup = {
							type = "group", order = 35, name = "Time Format",  inline = true,
							args = {
								TimeFormat = {
									type = "select", order = 10, name = "  Options", width = "double",
									desc = "Select time format for bars and icons. There are separate settings for five ranges: " ..
										"more then an hour, 2 minutes to an hour, 1 minute to 2 minutes, 10 seconds to 1 minute, less than 10 seconds. " ..
										"These are illustrated by samples in the drop-down menu for each format option.",
									get = function(info) return MOD.db.global.CooldownTimeFormat end,
									set = function(info, value) MOD.db.global.CooldownTimeFormat = value end,
									values = function(info)
										local s = MOD.db.global.CooldownTimeSpaces
										local c = MOD.db.global.CooldownTimeCase
										return GetTimeFormatList(s, c)
									end,
									style = "dropdown",
								},
								Space1 = { type = "description", name = "", order = 15, width = "half" },
								Spaces = {
									type = "toggle", order = 20, name = "Spaces", width = "half",
									desc = "Include spaces between values in time format.",
									get = function(info) return MOD.db.global.CooldownTimeSpaces end,
									set = function(info, value) MOD.db.global.CooldownTimeSpaces = value end,
								},
								Capitals = {
									type = "toggle", order = 30, name = "Uppercase",
									desc = "If checked, use uppercase H, M and S in time format, otherwise use lowercase.",
									get = function(info) return MOD.db.global.CooldownTimeCase end,
									set = function(info, value) MOD.db.global.CooldownTimeCase = value end,
								},
							},
						},
						CooldownAlignment = {
							type = "group", order = 40, name = "Alignment", inline = true,
							args = {
								VText = { type = "description", name = "Vertical:", order = 1, width = "half" },
								VLeftAlign = {
									type = "toggle", order = 10, name = "Top",
									desc = "If checked, vertically align cooldown text to top.",
									get = function(info) return MOD.db.global.CooldownVertical == "Top" end,
									set = function(info, value) MOD.db.global.CooldownVertical = "Top" end,
								},
								VCenterAlign = {
									type = "toggle", order = 20, name = "Middle",
									desc = "If checked, vertically align cooldown text in middle.",
									get = function(info) return not MOD.db.global.CooldownVertical end,
									set = function(info, value) MOD.db.global.CooldownVertical = false end,
								},
								VRightAlign = {
									type = "toggle", order = 30, name = "Bottom",
									desc = "If checked, vertically align cooldown text to bottom.",
									get = function(info) return MOD.db.global.CooldownVertical == "Bottom" end,
									set = function(info, value) MOD.db.global.CooldownVertical = "Bottom" end,
								},
								spacer = { type = "description", name = "", order = 40 },
								HText = { type = "description", name = "Horizontal:", order = 41, width = "half" },
								HLeftAlign = {
									type = "toggle", order = 50, name = "Left",
									desc = "If checked, horizontally align cooldown text to left.",
									get = function(info) return MOD.db.global.CooldownHorizontal == "Left" end,
									set = function(info, value) MOD.db.global.CooldownHorizontal = "Left" end,
								},
								HCenterAlign = {
									type = "toggle", order = 60, name = "Center",
									desc = "If checked, horizontally align cooldown text in center.",
									get = function(info) return not MOD.db.global.CooldownHorizontal end,
									set = function(info, value) MOD.db.global.CooldownHorizontal = false end,
								},
								HRightAlign = {
									type = "toggle", order = 70, name = "Right",
									desc = "If checked, horizontally align cooldown text to right.",
									get = function(info) return MOD.db.global.CooldownHorizontal == "Right" end,
									set = function(info, value) MOD.db.global.CooldownHorizontal = "Right" end,
								},
							},
						},
					},
				},
			},
		},
	},
}

-- This table gets inserted into the bar settings tab for each active bar as the args field for a group.
-- Bar fields: name, enableBar, barLabel, barType, barData, monitor, isMine, color
MOD.barOptions = {
	SummaryGroup = {
		type = "group", order = 1, name = "Bar Information", inline = true,
		hidden = function(info) return NoBar() end,
		args = {
			BarDescription = {
				type = "description", order = 1, name = function(info) return GetBarDescription(info) end,
			},
		},
	},
	SortingGroup = {
		type = "group", order = 5, name = "Custom Sort Order", inline = true,
		hidden = function(info) return NoBar() or GetBarGroupField("sor") ~= "X" end,
		args = {
			PromoteBar = {
				type = "execute", order = 60, name = "Move Up",
				desc = "Move the bar up, overriding sort order.",
				func = function(info) MoveBarInList(info, "up") end,
			},
			DemoteBar = {
				type = "execute", order = 70, name = "Move Down",
				desc = "Move the bar down, overriding sort order.",
				func = function(info) MoveBarInList(info, "down") end,
			},
		},
	},
	OptionsGroup = {
		type = "group", order = 10, name = "Settings", inline = true,
		hidden = function(info) return NoBar() end,
		args = {
			EnableBar = {
				type = "toggle", order = 10, name = "Enable", width = "half",
				desc = "If checked, enable showing the bar.",
				get = function(info) return GetBarField(info, "enableBar") end,
				set = function(info, value) SetBarField(info, "enableBar", value); MOD:UpdateAllBarGroups() end,
			},
			BarLabel = {
				type = "input", order = 20, name = " Label",
				desc = "Enter label text for the bar.",
				get = function(info) return GetBarField(info, "barLabel") end,
				set = function(info, value) SetBarLabel(info, value, not GetBarField(info, "labelLink")); MOD:UpdateAllBarGroups() end,
			},
			LabelLink = {
				type = "toggle", order = 30, name = "Label Link",
				desc = "If checked, label is linked to the associated spell and changing it here will change it for all bars linked to the same spell.",
				hidden = function(info) return GetBarField(info, "barType") == "Notification" end,
				get = function(info) return not GetBarField(info, "labelLink") end,
				set = function(info, value)
					if value then
						local bar = GetBarEntry(info)
						local label = MOD:GetLabel(bar.action)
						value = SetBarLabel(info, label, false) -- no need to update the cached label, also don't link if fails validation test
					end
					SetBarField(info, "labelLink", not value)
					MOD:UpdateAllBarGroups()
				end,
			},
			spacer = { type = "description", name = "", order = 50, hidden = function(info) return GetBarField(info, "barType") ~= "Cooldown" end, },
			EnableReadyBar = {
				type = "toggle", order = 60, name = "Show When Ready",
				desc = "If checked, show bar when action is not on cooldown. You must enable showing bars with no duration to see ready bars.",
				hidden = function(info) return GetBarField(info, "barType") ~= "Cooldown" end,
				get = function(info) return GetBarField(info, "enableReady") end,
				set = function(info, value) SetBarField(info, "enableReady", value); MOD:UpdateAllBarGroups() end,
			},
			EnableUsableTest = {
				type = "toggle", order = 70, name = "Is Usable",
				desc = "If checked, show ready bar only if spell is usable (i.e., enough mana, reagents, etc.).",
				hidden = function(info) return GetBarField(info, "barType") ~= "Cooldown" end,
				disabled = function(info) return not GetBarField(info, "enableReady") end,
				get = function(info) return not GetBarField(info, "readyNotUsable") end,
				set = function(info, value) SetBarField(info, "readyNotUsable", not value); MOD:UpdateAllBarGroups() end,
			},
		},
	},
	ColorGroup = {
		type = "group", order = 20, name = "Color and Icon", inline = true,
		hidden = function(info) return NoBar() end,
		args = {
			UseColor = {
				type = "toggle", order = 10, name = "Use Spell Color", 
				desc = "If checked, use the color from the associated spell.",
				hidden = function(info) return GetBarField(info, "barType") ~= "Notification" end,
				disabled = function(info)
					local bar = GetBarEntry(info)
					return bar.barType == "Notification" and not MOD:GetAssociatedSpellForBar(bar)
				end,
				get = function(info) return not GetBarField(info, "notColor") end,
				set = function(info, value)
					SetBarField(info, "notColor", not value)
					if value then SetBarField(info, "color", nil) end
					MOD:UpdateAllBarGroups()
				end,
			},
			UseIcon = {
				type = "toggle", order = 10, name = "Use Spell Icon", 
				desc = "If checked, use the icon from the associated spell.",
				hidden = function(info) return GetBarField(info, "barType") ~= "Notification" end,
				disabled = function(info)
					local bar = GetBarEntry(info)
					return bar.barType == "Notification" and not MOD:GetAssociatedSpellForBar(bar)
				end,
				get = function(info) return not GetBarField(info, "notIcon") end,
				set = function(info, value) SetBarField(info, "notIcon", not value); MOD:UpdateAllBarGroups() end,
			},
			spacer = { type = "description", name = "", order = 20, hidden = function(info) return GetBarField(info, "barType") ~= "Notification" end, },
			StandardColors = {
				type = "select", order = 30, name = "  Color", width = "half",
				desc = "Select a standard color or click to set a custom color.",
				get = function(info) return nil end,
				set = function(info, value)
					local r, g, b, a = GetStandardColor(value)
					local bar = GetBarEntry(info)
					MOD:SetBarColor(bar, r, g, b, a)
					MOD:UpdateAllBarGroups()
				end,
				values = function(info) return GetStandardColorList() end,
				style = "dropdown",
			},
			BarColor = {
				type = "color", order = 40, name = "", hasAlpha = false, width = "half",
				get = function(info)
					local bar = GetBarEntry(info)
					local t = MOD:GetBarColor(bar, true)
					return t.r, t.g, t.b, t.a
				end,
				set = function(info, r, g, b, a)
					local bar = GetBarEntry(info)
					MOD:SetBarColor(bar, r, g, b, a)
					MOD:UpdateAllBarGroups()
				end,
			},
			ColorLink = {
				type = "toggle", order = 50, name = "Color Link",
				desc = "If checked, the color is linked to the associated spell and changing it here will change it for all bars linked to the same spell.",
				disabled = function(info)
					local bar = GetBarEntry(info)
					return bar.barType == "Notification" and not MOD:GetAssociatedSpellForBar(bar)
				end,
				get = function(info) return not GetBarField(info, "colorLink") end,
				set = function(info, value)
					local bar = GetBarEntry(info)
					bar.colorLink = not value
					MOD:LinkBarColor(bar)
					MOD:UpdateAllBarGroups()
				end,
			},
			BarIcon = {
				type = "description", order = 60, name = "", width = "half", 
				hidden = function(info) return not GetBarIcon(info) end,
				image = function(info) return GetBarIcon(info) end,
				imageWidth = 24, imageHeight = 24,
			},
		},
	},
	ConditionGroup = {
		type = "group", order = 30, name = "Conditional Effects", inline = true,
		hidden = function(info) return NoBar() end,
		args = {
			HideBar = {
				type = "toggle", order = 10, name = "Hide", width = "half",
				desc = "If checked, bar is conditionally hidden.",
				get = function(info) return IsOn(GetBarField(info, "hideBar")) end,
				set = function(info, value)
					local v = GetBarField(info, "hideBar")
					SetBarField(info, "hideBar", IsOn(v) and Off or true)
					MOD:UpdateAllBarGroups()
				end,
			},
			HideBarTrue = {
				type = "toggle", order = 11, name = "True", width = "half",
				desc = "If checked, bar is hidden if the condition is true.",
				disabled = function(info) return IsOff(GetBarField(info, "hideBar")) end,
				get = function(info) return GetBarField(info, "hideBar") == true end,
				set = function(info, v) SetBarField(info, "hideBar", true); MOD:UpdateAllBarGroups() end,
			},
			HideBarFalse = {
				type = "toggle", order = 12, name = "False", width = "half",
				desc = "If checked, bar is hidden if the condition is false.",
				disabled = function(info) return IsOff(GetBarField(info, "hideBar")) end,
				get = function(info) return GetBarField(info, "hideBar") == false end,
				set = function(info, v) SetBarField(info, "hideBar", false); MOD:UpdateAllBarGroups() end,
			},
			SelectCondition1 = {
				type = "select", order = 15, name = "  Hide Condition",
				disabled = function(info) return IsOff(GetBarField(info, "hideBar")) end,
				get = function(info) return GetBarSelectedCondition(GetSelectConditionList(), GetBarField(info, "hideCondition")) end,
				set = function(info, value) SetBarField(info, "hideCondition", GetSelectConditionList()[value]) end,
				values = function(info) return GetSelectConditionList() end,
				style = "dropdown",
			},
			spacer1 = { type = "description", name = "", order = 18 },
			FlashBar = {
				type = "toggle", order = 20, name = "Flash", width = "half",
				desc = "If checked, bar will conditionally flash.",
				get = function(info) return IsOn(GetBarField(info, "flashBar")) end,
				set = function(info, value)
					local v = GetBarField(info, "flashBar")
					SetBarField(info, "flashBar", IsOn(v) and Off or true)
					MOD:UpdateAllBarGroups()
				end,
			},
			FlashBarTrue = {
				type = "toggle", order = 21, name = "True", width = "half",
				desc = "If checked, bar will flash if the condition is true.",
				disabled = function(info) return IsOff(GetBarField(info, "flashBar")) end,
				get = function(info) return GetBarField(info, "flashBar") == true end,
				set = function(info, v) SetBarField(info, "flashBar", true); MOD:UpdateAllBarGroups() end,
			},
			FlashBarFalse = {
				type = "toggle", order = 22, name = "False", width = "half",
				desc = "If checked, bar will flash if the condition is false.",
				disabled = function(info) return IsOff(GetBarField(info, "flashBar")) end,
				get = function(info) return GetBarField(info, "flashBar") == false end,
				set = function(info, v) SetBarField(info, "flashBar", false); MOD:UpdateAllBarGroups() end,
			},
			SelectCondition2 = {
				type = "select", order = 25, name = "  Flash Condition",
				disabled = function(info) return IsOff(GetBarField(info, "flashBar")) end,
				get = function(info) return GetBarSelectedCondition(GetSelectConditionList(), GetBarField(info, "flashCondition")) end,
				set = function(info, value) SetBarField(info, "flashCondition", GetSelectConditionList()[value]) end,
				values = function(info) return GetSelectConditionList() end,
				style = "dropdown",
			},
			spacer2 = { type = "description", name = "", order = 28 },
			FadeBar = {
				type = "toggle", order = 30, name = "Fade", width = "half",
				desc = "If checked, bar will conditionally change from normal opacity to fade opacity.",
				get = function(info) return IsOn(GetBarField(info, "fadeBar")) end,
				set = function(info, value)
					local v = GetBarField(info, "fadeBar")
					SetBarField(info, "fadeBar", IsOn(v) and Off or true)
					MOD:UpdateAllBarGroups()
				end,
			},
			FadeBarTrue = {
				type = "toggle", order = 31, name = "True", width = "half",
				desc = "If checked, bar will fade if the condition is true.",
				disabled = function(info) return IsOff(GetBarField(info, "fadeBar")) end,
				get = function(info) return GetBarField(info, "fadeBar") == true end,
				set = function(info, v) SetBarField(info, "fadeBar", true); MOD:UpdateAllBarGroups() end,
			},
			FadeBarFalse = {
				type = "toggle", order = 32, name = "False", width = "half",
				desc = "If checked, bar will fade if the condition is false.",
				disabled = function(info) return IsOff(GetBarField(info, "fadeBar")) end,
				get = function(info) return GetBarField(info, "fadeBar") == false end,
				set = function(info, v) SetBarField(info, "fadeBar", false); MOD:UpdateAllBarGroups() end,
			},
			SelectCondition3 = {
				type = "select", order = 35, name = "  Fade Condition",
				disabled = function(info) return IsOff(GetBarField(info, "fadeBar")) end,
				get = function(info) return GetBarSelectedCondition(GetSelectConditionList(), GetBarField(info, "fadeCondition")) end,
				set = function(info, value) SetBarField(info, "fadeCondition", GetSelectConditionList()[value]) end,
				values = function(info) return GetSelectConditionList() end,
				style = "dropdown",
			},
			spacer3 = { type = "description", name = "", order = 40 },
			spacer4 = { type = "description", name = "", order = 41, width = "half" },
			BarNormalAlpha = {
				type = "range", order = 44, name = "Normal Opacity", min = 0, max = 1, step = 0.05,
				desc = "Set alpha for normal bars.",
				disabled = function(info) return IsOff(GetBarField(info, "fadeBar")) end,
				get = function(info) return GetBarField(info, "normalAlpha") end,
				set = function(info, value) SetBarField(info, "normalAlpha", value) end,
			},
			BarFadeAlpha = {
				type = "range", order = 45, name = "Fade Opacity", min = 0, max = 1, step = 0.05,
				desc = "Set alpha for faded bars.",
				disabled = function(info) return IsOff(GetBarField(info, "fadeBar")) end,
				get = function(info) return GetBarField(info, "fadeAlpha") end,
				set = function(info, value) SetBarField(info, "fadeAlpha", value) end,
			},
		},
	},
	BarTypeGroup = {
		type = "group", order = 40, name = "", inline = true,
		hidden = function(info) return not InMode("Bar") end,
		args = {
			SelectBarType = {
				type = "group", order = 1, name = "Type", inline = true,
				args = {
					NotificationBar = {
						type = "toggle", order = 10, name = "Notify", width = "half",
						desc = "If checked, this is a notify bar.",
						get = function(info) return newbar.barType == "Notification" end,
						set = function(info, value) SetSelectedBarType("Notification") end,
					},
					BuffBar = {
						type = "toggle", order = 20, name = "Buff", width = "half",
						desc = "If checked, this is a buff bar.",
						get = function(info) return newbar.barType == "Buff" end,
						set = function(info, value) SetSelectedBarType("Buff") end,
					},
					DebuffBar = {
						type = "toggle", order = 30, name = "Debuff", width = "half",
						desc = "If checked, this is a debuff bar.",
						get = function(info) return newbar.barType == "Debuff" end,
						set = function(info, value) SetSelectedBarType("Debuff") end,
					},
					CooldownBar = {
						type = "toggle", order = 40, name = "Cooldown",
						desc = "If checked, this is a cooldown bar.",
						get = function(info) return newbar.barType == "Cooldown" end,
						set = function(info, value) SetSelectedBarType("Cooldown") end,
					},
				},
			},
			SelectBarSource = {
				type = "group", order = 10, name = "Source", inline = true,
				hidden = function(info) return (newbar.barType == "Notification") end,
				args = {
					Class = {
						type = "toggle", order = 10, name = "Class", width = "half",
						desc = "If checked, select actions associated with a class.",
						get = function(info) return newbar.barSource <= 10 end,
						set = function(info, value) if newbar.barSource > 10 then SetSelectedBarSource(classIndex[MOD.myClass]) end end,
					},
					Racial = {
						type = "toggle", order = 20, name = "Race", width = "half",
						desc = "If checked, select actions associated with a race.",
						get = function(info) return newbar.barSource == 11 end,
						set = function(info, value) SetSelectedBarSource(11) end,
					},
					Spell = {
						type = "toggle", order = 30, name = "Other Spell",
						desc = "If checked, enter spell name for an associated action.",
						get = function(info) return newbar.barSource == 12 end,
						set = function(info, value) SetSelectedBarSource(12) end,
					},
				},
			},
			SelectClass = {
				type = "group", order = 15, name = "Class", inline = true,
				hidden = function(info) return (newbar.barType == "Notification") or (newbar.barSource > 10) end,
				args = {
					Druid = {
						type = "toggle", order = 10, name = "Druid", width = "half",
						desc = "If checked, select druid actions.",
						get = function(info) return newbar.barSource == 2 end,
						set = function(info, value) SetSelectedBarSource(2) end,
					},
					Hunter = {
						type = "toggle", order = 15, name = "Hunter", width = "half",
						desc = "If checked, select hunter actions.",
						get = function(info) return newbar.barSource == 3 end,
						set = function(info, value) SetSelectedBarSource(3) end,
					},
					Mage = {
						type = "toggle", order = 20, name = "Mage", width = "half",
						desc = "If checked, select mage actions.",
						get = function(info) return newbar.barSource == 4 end,
						set = function(info, value) SetSelectedBarSource(4) end,
					},
					Paladin = {
						type = "toggle", order = 25, name = "Paladin", width = "half",
						desc = "If checked, select paladin actions.",
						get = function(info) return newbar.barSource == 5 end,
						set = function(info, value) SetSelectedBarSource(5) end,
					},
					Priest = {
						type = "toggle", order = 30, name = "Priest", width = "half",
						desc = "If checked, select priest actions.",
						get = function(info) return newbar.barSource == 6 end,
						set = function(info, value) SetSelectedBarSource(6) end,
					},
					Rogue = {
						type = "toggle", order = 35, name = "Rogue", width = "half",
						desc = "If checked, select rogue actions.",
						get = function(info) return newbar.barSource == 7 end,
						set = function(info, value) SetSelectedBarSource(7) end,
					},
					Shaman = {
						type = "toggle", order = 40, name = "Shaman", width = "half",
						desc = "If checked, select shaman actions.",
						get = function(info) return newbar.barSource == 8 end,
						set = function(info, value) SetSelectedBarSource(8) end,
					},
					Warlock = {
						type = "toggle", order = 45, name = "Warlock", width = "half",
						desc = "If checked, select warlock actions.",
						get = function(info) return newbar.barSource == 9 end,
						set = function(info, value) SetSelectedBarSource(9) end,
					},
					Warrior = {
						type = "toggle", order = 50, name = "Warrior", width = "half",
						desc = "If checked, select warrior actions.",
						get = function(info) return newbar.barSource == 10 end,
						set = function(info, value) SetSelectedBarSource(10) end,
					},
					DeathKnight = {
						type = "toggle", order = 55, name = "Death Knight",
						desc = "If checked, select death knight actions.",
						get = function(info) return newbar.barSource == 1 end,
						set = function(info, value) SetSelectedBarSource(1) end,
					},
				},
			},
			SelectConditionGroup = {
				type = "group", order = 20, name = "Conditions To Monitor", inline = true,
				hidden = function(info) return (newbar.barType ~= "Notification") or (GetBarConditionListCount() == 0) end,
				args = {
					AllOnBar = {
						type = "execute", order = 10, name = "All On", width = "half",
						desc = "Select all the conditions.",
						func = function(info) SetAllBarConditions(true) end,
					},
					AllOffBar = {
						type = "execute", order = 20, name = "All Off", width = "half",
						desc = "Deselect all the conditions.",
						func = function(info) SetAllBarConditions(false) end,
					},
					SelectCondition = {
						type = "multiselect", order = 30, name = "",
						get = function(info, key) return GetSelectedBarCondition(key) end,
						set = function(info, key, value) SetSelectedBarCondition(key, value) end,
						values = function(info) return GetBarConditionList() end,
					},
				},
			},
			EnterActionGroup = {
				type = "group", order = 40, name = "Other Spell", inline = true,
				hidden = function(info) return (newbar.barType == "Notification") or (newbar.barSource ~= 12) end,
				args = {
					SpellName = {
						type = "input", order = 10, name = " Spell Name",
						desc = "Enter spell name (or numeric identifier) to associate with a bar.",
						get = function(info) return selectedSpellName end,
						set = function(info, value)
							local n, t = value, tonumber(value)
							if t then n = GetSpellInfo(t) end
							ValidateSpellName(n)
							selectedSpellName = n
						end,
					},
				},
			},
			BarActionGroup = {
				type = "group", order = 50, name = "Actions To Monitor", inline = true,
				hidden = function(info) return (newbar.barSource == 12) or (GetBarActionListCount() == 0) end,
				args = {
					AllOnBar = {
						type = "execute", order = 10, name = "All On", width = "half",
						desc = "Select all the actions.",
						func = function(info) SetAllBarActions(true) end,
					},
					AllOffBar = {
						type = "execute", order = 20, name = "All Off", width = "half",
						desc = "Deselect all the actions.",
						func = function(info) SetAllBarActions(false) end,
					},
					SelectBarAction = {
						type = "multiselect", order = 30, name = "",
						get = function(info, key) return GetSelectedBarAction(key) end,
						set = function(info, key, value) SetSelectedBarAction(key, value) end,
						values = function(info) return GetBarActionList() end,
					},
				},
			},
		},
	},
	MonitorUnitGroup = {
		type = "group", order = 50, name = "Action On", inline = true,
		hidden = function(info) return not InMode("Bar") or ((newbar.barType ~= "Buff") and (newbar.barType ~= "Debuff")) end,
		args = {
			PlayerBuff = {
				type = "toggle", order = 10, name = "Player",
				desc = "If checked, this is a player buff or debuff.",
				get = function(info) return newbar.monitor == "player" end,
				set = function(info, value) newbar.monitor = "player" end,
			},
			TargetBuff = {
				type = "toggle", order = 20, name = "Target",
				desc = "If checked, this is a target buff or debuff.",
				get = function(info) return newbar.monitor == "target" end,
				set = function(info, value) newbar.monitor = "target" end,
			},
			FocusBuff = {
				type = "toggle", order = 30, name = "Focus",
				desc = "If checked, this is a focus buff or debuff.",
				get = function(info) return newbar.monitor == "focus" end,
				set = function(info, value) newbar.monitor = "focus" end,
			},
		},
	},
	CastUnitGroup = {
		type = "group", order = 60, name = "Cast By", inline = true,
		hidden = function(info) return not InMode("Bar") or ((newbar.barType ~= "Buff") and (newbar.barType ~= "Debuff")) end,
		args = {
			MyBuff = {
				type = "toggle", order = 10, name = "Player",
				desc = "If checked, only track if cast by the player.",
				get = function(info) return newbar.isMine and not newbar.isOther end,
				set = function(info, value) newbar.isMine = true; newbar.isOther = false end,
			},
			YourBuff = {
				type = "toggle", order = 20, name = "Other",
				desc = "If checked, only track if cast by anyone other than the player.",
				get = function(info) return not newbar.isMine and newbar.isOther end,
				set = function(info, value) newbar.isMine = false; newbar.isOther = true end,
			},
			OurBuff = {
				type = "toggle", order = 30, name = "Anyone",
				desc = "If checked, track if cast by anyone, including player.",
				get = function(info) return newbar.isMine and newbar.isOther end,
				set = function(info, value) newbar.isMine = true; newbar.isOther = true end,
			},
		},
	},
	OKNewBar = {
		type = "execute", order = 93, name = "OK", width = "half",
		desc = "Cancel creating a new bar.",
		hidden = function(info) return not InMode("Bar") end,
		func = function(info) EnterNewBar("ok") end,
	},
	CancelNewBar = {
		type = "execute", order = 94, name = "Cancel", width = "half",
		desc = "Cancel creating a new bar.",
		hidden = function(info) return not InMode("Bar") end,
		func = function(info) EnterNewBar("cancel") end,
	},
}
